#!/usr/bin/perl
#
# Build a hash table describing every alignment in input SAM file,
# (in which all alignments have at least one gap), and calculate
# entropy for every unique intron.
#
# Input: gapped alignment SAM file
#
# Ouput is 10 column format:
#  chr  b  e strand  n_alns  nl  nr  entropy  m  err
# where:
#
# chr    : sequence (chromosome)
# b, e   : nt coord of exons flanking gap (b < e)
# strand : + or -
# n_alns : number of alignments to this gap
# nl, nr : max number of nt aligned to left and right of gap
# entropy : entropy (see below and/or paper for calculation)
# m      : sum(multi)/n_alns, where multi = 9th column of input file, for all alns to this gap
# err    : max alignment error to this gap (among all alignments to this gap)
#
# Written by Yang Li.
# Description by Paul Burns, Aug 2012.
#

## inputs
$sam_file = shift@ARGV;
$outfile = shift@ARGV;

#print "sam_file:       $sam_file\n";
#print "outfile:        $outfile\n";

open($in_sam,"<",$sam_file);    # sam file
open($out,">$outfile");

$ct = 0;

while(<$in_sam>)
{
	$lll = $_;
	@m = split(/\t/);
	@k = split(/M|N/,$m[5]);
	$s = $m[3];## need change after gapped reconpile
	$err = $m[7];
	$multi = $m[8];
	if($multi == 0){
		$multi = 1;
	}
	
	for $i (1 .. ($#k)/2){
		$s+=$k[$i*2-2];
		$e = $s + $k[$i*2-1];
		$t = "$m[2]:$s:$e";
		
		if (exists $hash{$t}) {
			$bExists = 1;
		}
		else {
			$bExists = 0;
		}

		$hash{$t}[0]++;
		@p = split(/:/,$m[12]);
		if($p[2] == 1){
			$hash{$t}[3] = "+";
		}
		else{
			$hash{$t}[3] = "-";
		}
		$min_over = $k[$i*2-2];
		if ($k[$i*2] < $min_over) { $min_over = $k[$i*2]; }
		if ($bExists) {
			if ($hash{$t}[1] < $min_over) {
				$hash{$t}[1] = $min_over;
			}
		}
		else {
			$hash{$t}[1] = $min_over;
		}
		##if($hash{$t}[1] eq ""){$hash{$t}[1] = $k[$i*2-2];}
		##else{if($hash{$t}[1] < $k[$i*2-2]){$hash{$t}[1] = $k[$i*2-2];}}
		##if($hash{$t}[2] eq ""){$hash{$t}[2] = $k[$i*2];}
		##else{if($hash{$t}[2] < $k[$i*2]){$hash{$t}[2] = $k[$i*2];}}
		$s+=$k[$i*2-1];
		$hash{$t}[4] += $multi;
		if($hash{$t}[5] eq ""){
			$hash{$t}[5] = $err;
		}
		elsif($hash{$t}[5] > $err){
			$hash{$t}[5] = $err;
		}
		$hash{$t}[6]{$m[3]}++;
	}
}

foreach $key (keys %hash){
#	print "$key\t$hash{$key}[0]\t$hash{$key}[1]\t$hash{$key}[2]\t$hashjunction{$key}\n";
#	if($hash{$key}[0] > 5){
	@j = split(/:/,$key);
	$m = $hash{$key}[4]/$hash{$key}[0];
#		if($j[1] == )
#		if($hash{$key}[3] eq "+"){
	$total = 0;
	foreach $pos (keys %{$hash{$key}[6]}){
		$total += $hash{$key}[6]{$pos};
	}
	$entropy = 0;
	foreach $pos (keys %{$hash{$key}[6]}){
		$entropy -= log($hash{$key}[6]{$pos}/$total)/log(10);
	}
	print $out "$j[0]\t$j[1]\t$j[2]\t$hash{$key}[3]\t$hash{$key}[0]\t$hash{$key}[1]\t$entropy\t$m\t$hash{$key}[5]\n";
}
