#!/usr/bin/perl -w
######
#
#	Li, Yang	yangli9@illinois.edu
#   Burns, Paul pb128@gatech.edu
#
#
######
use Getopt::Long ;
use Getopt::Long qw(:config no_ignore_case);
use POSIX;
use FindBin qw($Bin);


#################################
$VERSION = 0.04;
#################################

##if(scalar @Fq == 0 || $bwt eq ""){
$Usage = "re-alignment pipeline\nUsage:\n
	-a/--gap-aligns         input gapped alignments (SAM)
	-e/--paired-end <0|1>   1 if paired-end alignment, otherwise 0
	-p/--thread             number of cores
	-r/--bowtie-index       Bowtie reference dir/name, thus bowtie index should be dir/name.*.ebwt, reference genome should also be located here, e.g. dir/name.fa; chromosome line for dir/name.fa should be clean and there is no space within it, such as \">XXXXX\\n\"
	-l/--read-length        read length
	-v/--mismatch           mismatches (0-3)
	-s/--segment-length     segment length (18-25)
\n\n";
	
if ( $#ARGV < 1 ) { print $Usage; exit 1; }

GetOptions(
	'a|gap-aligns=s' => \$input_sam,
	'e|paired-end=i' => \$pair,
	'r|bowtie-index=s' => \$bwt,
	'p|thread=i' => \$P,
	'l|read-length=i' => \$read_length,
	's|segment-length=i' => \$segment_length,
	'v|mismatch=i' => \$V);
	

$now_string = localtime;
print "Realignment beginning at time " . "$now_string\n";

$segment_length = $segment_length || 25 ; ## length of segments 18-25
$read_length = $read_length || 75;
$V = $V || 2;
$M = $M || 5;
$P = $P || 1;
$pair = $pair || 0;


# combine all scored gapped alignments into one file (Gapped.sam_4)
if ($pair) {
	system("cat $input_sam | $Bin/linkPair_Report_memfix.pl 1  > alignments");
}
else { system("cp $input_sam alignments"); }
system("$Bin/junctionEva.pl alignments all_junctions");

system("$Bin/presentJunc.pl all_junctions alignments > PJ_junctions");

system("$Bin/getCanoFlag.pl PJ_junctions $bwt.fa > PJ_canoflag");

system("$Bin/reMapFromJunction_new PJ_canoflag $bwt.fa $read_length > REMAP.fa");

system("$Bin/bowtie-build --quiet REMAP.fa REMAP; $Bin/bowtie --quiet -k 20 -v $V -p $P REMAP temp.un.fq > REMAP.bwt");
system("$Bin/translateRemapBWT.pl $read_length PJ_canoflag REMAP.bwt > REMAP.sam");

system("cat alignments REMAP.sam | sort -t '/' -k 1,1n > remap_sorted.sam; rm alignments");
system("cat remap_sorted.sam | $Bin/mergeRemapSam.pl PJ_canoflag 1 $bwt.fa | sort -t '/' -k 1,1 > alignments");

$now_string = localtime;
print "Realignment finished at time " . "$now_string\n";

#mySystem("cat Gapped.sam_4 REMAP.sam | sort -t '/' -k 1,1n | $Bin/mergeRemapSAM_memfix.pl SJS_canoflag 1 $bwt.fa | sort -t '/' -k 1,1n | $Bin/rmUnPairGapAli_new_19.pl | $Bin/linkPair_Report_memfix_new.pl 0 > final.sam_2","Present sam and junctions");
