#include <iostream>
#include <string>
#include <vector>
#include <map>
#include <iomanip>
#include <fstream>
#include <cctype>
#include <set>
#include <iterator>
#include <algorithm>
#include <sstream>
#include <cmath>
#include <cstdlib>
#include "targetver.h"
#include "options.h"
#include "structs.h"
#include "enum.h"

#include <limits>

#define _DEBUG_MODE

#define _USE_MATH_DEFINES
using namespace std;


//////////////////////////////////////////////////
// load all model files found in given directory (call at beginning of your program)
bool initSeqScore(string& dir);
bool get_splice_scores( double& don_score, double& acc_score, interval tron, int jcnType, string chr );
bool intronScore(double& logScore, int order, string& chr, interval& tron);
double lengthScore(string& chr, interval& tron);

void modCalc( double *pMat, vector<string>& dataVect, int order, int flankLeft, int flankRight );

bool getStrand(interval&, int& iJcnType, string chr, int donor, int acceptor);
bool isValidChr(string chr);

//////////////////////////////////////////////////


// program options
cOptions opts;

ofstream Out_junction;
ofstream Out_junction_multi;
ofstream Out_junction_noncano;

ofstream Out_junction_test;
ofstream Out_test_set;

int critical_intron = 10000;
string BIN;
vector<double> param_lr;
double intron_eva, intron_var;

ofstream Out_train;
ofstream Out_predict;
ofstream Out_predict_multi;

ofstream Out_training_debug;

int orderSite = 0;

map<string, map<interval, intronInfo> > positive_intron, negative_intron, unlabel_intron, all_junction;
map<string, map<pair<int,int>, num_info> > exist_Junction;
map<string, map<int, int> > exist_Donor, exist_Acceptor;
map< string, map<string, double> > donor_p, acceptor_p, cod_p, noncod_p;

map<string, string> ref;
vector< potJunc > candidate;

map<string, map<int, double> > full_cov_first, full_cov_second;

string m_delim = " \t";

int order = 5;
int donor_length = 23;
int donor_offset = 3;
int acceptor_length = 23;
int acceptor_offset = 20;

int ctTrain;
int ctTrainLimit = 5000;
int ctTestLimit = 10000;		// test set size (used for parameter selection)

double *pDonMod(NULL);
double *pAccMod(NULL);

double log0 = -100.;

char alphabet[256];

FILE *pfDonScore;
FILE *pfAccScore;
FILE *pfDebug;


bool operator<(interval const& a, interval const& b)
{
	if(a.start == b.start)
		return a.end < b.end;
	return a.start < b.start;
}

inline int samBeg(map<int,vector<vector<int> > >::iterator it, int t) { return it->second[t][1]+it->second[t][0]; }
inline int samEnd(map<int,vector<vector<int> > >::iterator it, int t) { return it->second[t][1]+it->second[t][2]+it->second[t][0]; }

void clean()
{
	if (pDonMod != NULL) delete []pDonMod;
	if (pAccMod != NULL) delete []pAccMod;
}

string revString(string Forward){
        string Reverse;
        for(int i=(int)Forward.length()-1;i>=0;i--){
                if(Forward[i]=='A')
                        Reverse.push_back('T');
                else if(Forward[i]=='T')
                        Reverse.push_back('A');
                else if(Forward[i]=='G')
                        Reverse.push_back('C');
                else if(Forward[i]=='C')
                        Reverse.push_back('G');
                else
                        Reverse.push_back(Forward[i]);// incase of NNN
        }
        return (Reverse);
}

string getSeq(string chr, int point, int offset, int length, string ori){
	if(ori == "+"){
		if(point-offset < 0 || point-offset+length > (int)ref[chr].length()){
			return "";
		}
		return(ref[chr].substr(point-offset, length));
	}
	else{
		if(point-(length-offset) < 0 || point+offset > (int)ref[chr].length()){
			return "";
                }
		return(revString(ref[chr].substr(point-(length-offset), length)));
	}
}

void getSeq(string& seq, string chr, int ori, int b, int e)
{
	seq.clear();

	if (b<0 || e>=(int)ref[chr].length())
		return;

	if(ori == -1) {		// negative strand
		seq = revString( ref[chr].substr(b, e-b+1) );
	}
	else {				// positive strand
		seq = ref[chr].substr(b, e-b+1);
	}
}

void getSeq(string& seq, string chr, int ori, interval tron)
{
	seq.clear();

	int b, e;
	if (ori > 0) { b = tron.start; e = tron.end; }
	else { b = tron.end; e = tron.start; }

	getSeq(seq, chr, ori, b, e);
}

void sendOutTrain(ofstream& ofs, intronSample& samp)
{
	bool bFirst(false);

	intronScores is = samp.is;

	if (opts.m_bDebug) {
		Out_training_debug << samp.chr << "\t" << samp.start << "\t" << samp.end;
		if (opts.m_uiFeatureMask & _FEAT_SPLICE_DON)	{ Out_training_debug << "\t" << is.spliceDon /*+ is.spliceAcc*/; }
		if (opts.m_uiFeatureMask & _FEAT_SPLICE_ACC)	{ Out_training_debug << "\t" << is.spliceAcc; }
		if (opts.m_uiFeatureMask & _FEAT_CODING)	{ Out_training_debug << "\t" << is.coding; }
		if (opts.m_uiFeatureMask & _FEAT_DIFF_COV)	{ Out_training_debug << "\t" << is.covDiffLeft << "\t" << is.covDiffRight; }
		if (opts.m_uiFeatureMask & _FEAT_LENGTH)	{ Out_training_debug << "\t" << is.length; }
		if (opts.m_uiFeatureMask & _FEAT_DEPTH)		{ Out_training_debug << "\t" << is.depthScore; }
		if (opts.m_uiFeatureMask & _FEAT_OVERHANG)	{ Out_training_debug << "\t" << is.minMapLen; }
		if (opts.m_uiFeatureMask & _FEAT_ENTROPY)	{ Out_training_debug << "\t" << is.entropy; }
		if (opts.m_uiFeatureMask & _FEAT_AMBIGUITY)	{ Out_training_debug << "\t" << is.invNumCand; }
		if (opts.m_uiFeatureMask & _FEAT_MISMATCH)	{ Out_training_debug << "\t" << is.mismatches; }
		if (opts.m_uiFeatureMask & _FEAT_FRAMESHIFT_ERROR)	{ Out_training_debug << "\t" << is.iframeErr; }
		if (opts.m_uiFeatureMask & _FEAT_STRAND_ERROR)	{ Out_training_debug << "\t" << is.istrandErr; }
		Out_training_debug << "\t" << is.train << "\n";
	}

	//if (opts.m_bDebug) {
	//	Out_training_debug << samp.chr << "\t" << samp.start << "\t" << samp.end << "\t" << is.spliceDon << "\t"
	//		<< is.spliceAcc << "\t" << is.coding << "\t" << is.covDiff << "\t" << is.length << "\t"
	//		<< is.depthScore << "\t" << is.minMapLen << "\t" << is.entropy << "\t"
	//		<< is.invNumCand << "\t" << is.mismatches << "\t" << is.iframeErr << "\t" << is.istrandErr << "\t" << is.train << "\n";
	//}

	int ix(0);
	int itrain = is.train==0 ? -1 : 1;
	ofs << itrain;
	if (opts.m_uiFeatureMask & _FEAT_SPLICE_DON)	{ ofs << "  " << ++ix << ":" << is.spliceDon /*+ is.spliceAcc*/; }
	if (opts.m_uiFeatureMask & _FEAT_SPLICE_ACC)	{ofs << "  " << ++ix << ":" << is.spliceAcc; }
	if (opts.m_uiFeatureMask & _FEAT_CODING)	{ ofs << "  " << ++ix << ":" << is.coding; }
	if (opts.m_uiFeatureMask & _FEAT_DIFF_COV)	{ ofs << "  " << ++ix << ":" << is.covDiffLeft;
												  ofs << "  " << ++ix << ":" << is.covDiffRight; }
	if (opts.m_uiFeatureMask & _FEAT_LENGTH)	{ ofs << "  " << ++ix << ":" << is.length; }
	if (opts.m_uiFeatureMask & _FEAT_DEPTH)		{ ofs << "  " << ++ix << ":" << is.depthScore; }
	if (opts.m_uiFeatureMask & _FEAT_OVERHANG)	{ ofs << "  " << ++ix << ":" << is.minMapLen; }
	if (opts.m_uiFeatureMask & _FEAT_ENTROPY)	{ ofs << "  " << ++ix << ":" << is.entropy; }
	if (opts.m_uiFeatureMask & _FEAT_AMBIGUITY)	{ ofs << "  " << ++ix << ":" << is.invNumCand; }
	if (opts.m_uiFeatureMask & _FEAT_MISMATCH)	{ ofs << "  " << ++ix << ":" << is.mismatches; }
	if (opts.m_uiFeatureMask & _FEAT_FRAMESHIFT_ERROR)	{ ofs << "  " << ++ix << ":" << is.iframeErr; }
	if (opts.m_uiFeatureMask & _FEAT_STRAND_ERROR)	{ ofs << "  " << ++ix << ":" << is.istrandErr; }
	ofs << "\n";

}

void sendOutFeatures(ofstream& ofs, intronScores& is)
{
	bool bFirst(false);

	int ix(0);
	ofs << "0";
	if (opts.m_uiFeatureMask & _FEAT_SPLICE_DON)	{ ofs << "  " << ++ix << ":" << is.spliceDon /*+ is.spliceAcc*/; }
	if (opts.m_uiFeatureMask & _FEAT_SPLICE_ACC)	{ofs << "  " << ++ix << ":" << is.spliceAcc; }
	if (opts.m_uiFeatureMask & _FEAT_CODING)	{ ofs << "  " << ++ix << ":" << is.coding; }
	if (opts.m_uiFeatureMask & _FEAT_DIFF_COV)	{ ofs << "  " << ++ix << ":" << is.covDiffLeft;
												  ofs << "  " << ++ix << ":" << is.covDiffRight; }
	if (opts.m_uiFeatureMask & _FEAT_LENGTH)	{ ofs << "  " << ++ix << ":" << is.length; }
	if (opts.m_uiFeatureMask & _FEAT_DEPTH)		{ ofs << "  " << ++ix << ":" << is.depthScore; }
	if (opts.m_uiFeatureMask & _FEAT_OVERHANG)	{ ofs << "  " << ++ix << ":" << is.minMapLen; }
	if (opts.m_uiFeatureMask & _FEAT_ENTROPY)	{ ofs << "  " << ++ix << ":" << is.entropy; }
	if (opts.m_uiFeatureMask & _FEAT_AMBIGUITY)	{ ofs << "  " << ++ix << ":" << is.invNumCand; }
	if (opts.m_uiFeatureMask & _FEAT_MISMATCH)	{ ofs << "  " << ++ix << ":" << is.mismatches; }
	if (opts.m_uiFeatureMask & _FEAT_FRAMESHIFT_ERROR)	{ ofs << "  " << ++ix << ":" << is.iframeErr; }
	if (opts.m_uiFeatureMask & _FEAT_STRAND_ERROR)	{ ofs << "  " << ++ix << ":" << is.istrandErr; }
	ofs << "\n";

}

void sendOutFeatures(ofstream& ofs, vector<double>& feat)
{
	ofs << "0";
	for (int i=0; i<(int)feat.size(); ++i)
	{
		ofs << "  " << i+1 << ":" << feat[i];
	}
	ofs << "\n";
}


//void writeCandLine( FILE *pfData, int iAln, int iCand, string& chr, char chOri, interval& tron, intronScores& ts )
//{
//	fprintf(pfData,"%d\t%d\t%s\t%c\t%d\t%d\t%f\t%f\t%f\t%f\t%d\t%d\t%d\t%d\t%f\t%f\t%d\n", iAln, iCand, chr.c_str(), chOri,
//		tron.start, tron.end, ts.spliceDon, ts.spliceAcc, ts.coding, ts.covDiff, ts.length, ts.flag, ts.depthScore,
//		ts.minMapLen, ts.entropy, ts.invNumCand, ts.mismatches );
//}

void writeCandLine( FILE *pfData, int iAln, int iCand, string& chr, char chOri, interval& tron, vector<double>& SCORE )
{
	fprintf(pfData,"%d\t%d\t%s\t%c\t%d\t%d", iAln, iCand, chr.c_str(), chOri,
		tron.start, tron.end );
	for (int i=0; i<(int)SCORE.size(); ++i) fprintf(pfData,"\t%f",SCORE[i]);
	fprintf(pfData,"\n");
}

void writeSeqs( string fname, vector<string>& seqs )
{
	FILE *pf = fopen( fname.c_str(), "w" );
	for (int i=0; i<(int)seqs.size(); ++i)
		fprintf(pf,"%s\n", seqs[i].c_str());
	fclose(pf);
}

void is2dv( vector<double>& v_score, intronScores& is )
{
	v_score.clear();

	if (opts.m_uiFeatureMask & _FEAT_SPLICE_DON)	v_score.push_back( is.spliceDon );
	if (opts.m_uiFeatureMask & _FEAT_SPLICE_ACC)	v_score.push_back( is.spliceAcc );
	if (opts.m_uiFeatureMask & _FEAT_CODING)	v_score.push_back( is.coding );
	if (opts.m_uiFeatureMask & _FEAT_DIFF_COV)	{ v_score.push_back( is.covDiffLeft ); v_score.push_back( is.covDiffRight ); }
	if (opts.m_uiFeatureMask & _FEAT_LENGTH)	v_score.push_back( is.length );
	if (opts.m_uiFeatureMask & _FEAT_DEPTH)		v_score.push_back( is.depthScore );
	if (opts.m_uiFeatureMask & _FEAT_OVERHANG)	v_score.push_back( is.minMapLen );
	if (opts.m_uiFeatureMask & _FEAT_ENTROPY)	v_score.push_back( is.entropy );
	if (opts.m_uiFeatureMask & _FEAT_AMBIGUITY)	v_score.push_back( is.invNumCand );
	if (opts.m_uiFeatureMask & _FEAT_MISMATCH)	v_score.push_back( is.mismatches );
	if (opts.m_uiFeatureMask & _FEAT_FRAMESHIFT_ERROR)	v_score.push_back( is.iframeErr );
	if (opts.m_uiFeatureMask & _FEAT_STRAND_ERROR)	v_score.push_back( is.istrandErr );
}

void dv2is( intronScores& is, vector<double>& v_score )
{
	int ix(0);

	if (opts.m_uiFeatureMask & _FEAT_SPLICE_DON)	is.spliceDon = v_score[ix++];
	if (opts.m_uiFeatureMask & _FEAT_SPLICE_ACC)	is.spliceAcc = v_score[ix++];
	if (opts.m_uiFeatureMask & _FEAT_CODING)	is.coding = v_score[ix++];
	if (opts.m_uiFeatureMask & _FEAT_DIFF_COV)	{ is.covDiffLeft = v_score[ix]; is.covDiffRight = v_score[ix+1]; ix+=2; }
	if (opts.m_uiFeatureMask & _FEAT_LENGTH)	is.length = v_score[ix++];
	if (opts.m_uiFeatureMask & _FEAT_DEPTH)		is.depthScore = v_score[ix++];
	if (opts.m_uiFeatureMask & _FEAT_OVERHANG)	is.minMapLen = (int)v_score[ix++];
	if (opts.m_uiFeatureMask & _FEAT_ENTROPY)	is.entropy = v_score[ix++];
	if (opts.m_uiFeatureMask & _FEAT_AMBIGUITY)	is.invNumCand = v_score[ix++];
	if (opts.m_uiFeatureMask & _FEAT_MISMATCH)	is.mismatches = (int)v_score[ix++];
	if (opts.m_uiFeatureMask & _FEAT_FRAMESHIFT_ERROR)	is.iframeErr = (int)v_score[ix++];
	if (opts.m_uiFeatureMask & _FEAT_STRAND_ERROR)	is.istrandErr = (int)v_score[ix++];
}


void writeStrings(vector<string>& vs, string fname)
{
	FILE *pf = fopen(fname.c_str(),"w");

	for (int i=0; i<(int)vs.size(); ++i)
	{
		fprintf(pf, "%s\n", vs[i].c_str());
	}
	fclose( pf );
}

int siteModIdx(string s)
{
	int adr = 0;
	char ch;

	for (int i=0; i<(int)s.length(); ++i) {
		ch = s.at(i);
		adr = 4 * adr + alphabet[ ch ];
	}

	return adr;
}

void writeSiteModel(double *pMod, int width, int order, string fname)
{
	FILE *pf = fopen(fname.c_str(),"w");
	int size = 4 << 2*order;
	int i, j;

	for (j=0; j<size; ++j)
	{
		if (j==0) {
			for (i=0; i<width; ++i) { fprintf(pf,"\t%d\t",i); }
			fprintf(pf, "\n");
		}

		for (i=0; i<width; ++i)
			fprintf(pf, "\t%f", pMod[size*i+j]);

		fprintf(pf, "\n");
	}

	fclose( pf );
}


void getRef(ifstream& input_ref)
{
	string fa_line,chr;

	while(!input_ref.eof())
	{
        getline(input_ref, fa_line);

        if(fa_line.length() < 1)
                continue;
        if(fa_line.substr(0,1) == ">")
		{
                istringstream ss(fa_line);
                string fa;
                ss >> fa;
				chr = fa.substr(1);
				continue;
		}
		ref[chr]+=fa_line;
	}
	for(map<string, string> ::iterator it = ref.begin(); it != ref.end(); it++){
		transform(it->second.begin(),it->second.end(),it->second.begin(),::toupper);
	}
}

double scoreDonFlanks(string& site)
{	// score model excluding canonical "GT"

	int size = 4 << 2*orderSite;
	int len, adr, i;
	double score(0.);
	string s;

	if (site.length() != opts.m_nDonWidth)
		return -100.;

	for (i = 0; i < opts.m_nDonOffset; i++){
		len = min( i, orderSite );
		s = site.substr(i-len,len+1);
		if (s.find('N',0) != string::npos)
			return -100.;
		adr = size*i + siteModIdx(s);
		if (!(adr>=0 && adr<size*opts.m_nDonWidth))
		{
			cout << "DON: adr=" << adr << ", i=" << i << ", size=" << size << ", len=" << len << endl;
			cout << "site=" << site << endl;
			return -99.;
		}
		score += pDonMod[adr];
	}
	for (i = opts.m_nDonOffset; i < opts.m_nDonWidth; i++){
		len = min( i, orderSite );
		s = site.substr(i-len,len+1);
		if (s.find('N',0) != string::npos)
			return -100.;
		adr = size*i + siteModIdx(s);
		if (!(adr>=0 && adr<size*opts.m_nDonWidth))
		{
			cout << "DON: adr=" << adr << ", i=" << i << ", size=" << size << ", len=" << len << endl;
			cout << "site=" << site << endl;
			return -99.;
		}
		score += pDonMod[adr];
	}

	return score;
}

double scoreAccFlanks(string& site)
{	// score model excluding canonical "AG"

	int size = 4 << 2*orderSite;
	int len, adr, i;
	double score(0.);
	string s;

	if (site.length() != opts.m_nAccWidth)
		return -100.;

	for (i = 0; i < opts.m_nAccOffset; i++){
		len = min( i, orderSite );
		s = site.substr(i-len,len+1);
		if (s.find('N',0) != string::npos)
			return -100.;
		adr = size*i + siteModIdx(s);
		if (!(adr>=0 && adr<size*opts.m_nAccWidth))
		{
			cout << "ACC: adr=" << adr << ", i=" << i << ", size=" << size << ", len=" << len << endl;
			cout << "site=" << site << endl;
			return -99.;
		}
		score += pAccMod[adr];
	}
	for (i = opts.m_nAccOffset; i < opts.m_nAccWidth; i++){
		len = min( i, orderSite );
		s = site.substr(i-len,len+1);
		if (s.find('N',0) != string::npos)
			return -100.;
		adr = size*i + siteModIdx(s);
		if (!(adr>=0 && adr<size*opts.m_nAccWidth))
		{
			cout << "ACC: adr=" << adr << ", i=" << i << ", size=" << size << ", len=" << len << endl;
			cout << "site=" << site << endl;
			return -99.;
		}
		score += pAccMod[adr];
	}

	return score;
}


void logSiteModel(double *pMod, int width, int order)
{
	int size = 4 << 2*orderSite;

	for(int i = 0; i < width*size; i++)
	{
		if (pMod[i] > 0.)
			pMod[i] = log( pMod[i] );
		else
			pMod[i] = log0;
	}
}

bool isNegativeTrain(intronScores& is)
{
	double cov_diff = is.covDiffLeft + is.covDiffRight;

	bool bCov = is.covDiffLeft < -1 || is.covDiffRight < -1 || cov_diff < -1;
	bool bMap = is.minMapLen <= 2;// && is.entropy < 5;

	return bCov || bMap;
}


bool isPositiveTrain(intronScores& is)
{
	int L = opts.m_nReadLength/2;
	int D = (L - 9) * .5;
	D = D < 20 ? D : 20;
	bool bCheck1 = is.minMapLen > D && is.entropy > 20;

	return bCheck1;
}

bool isSoftPositiveTrain(intronScores& is)
{
	int L = opts.m_nReadLength/2;
	int D = (L - 9) * .5;
	D = D < 20 ? D : 20;
	bool bCheck1 = is.minMapLen > D && is.depthScore > 2;	// gets thos with depth>=7

	return bCheck1;
}

int loadJunctions(string& fname)
{
	all_junction.clear();
	exist_Junction.clear();
	exist_Donor.clear();
	exist_Acceptor.clear();

	std::cout << "Loading junctions from: " << fname << std::endl;

	ifstream input_train( fname.c_str() );

	vector<string> donor_input, acceptor_input;// store all training sequence
	vector<string> cod_input, noncod_input;
	map<string, set<string> > donor_index, acceptor_index;
	string chr, b ,e, f, train_line, num, min_over, err, entropy, bad_frame, bad_strand, genic_state, dc;
	string seqSite;
	double total = 0;
	interval tron(0,0);
	int donor, acceptor, flag;

	double cod_potential(0.), don_score(0.), acc_score(0.), length_score(0.);
	intronScores is;
	int ctNegatives(0), ctPositives(0), ctTotal(0);
	int jcnType(0);
	int ctGtAg(0), ctGcAg(0), ctAtAc(0), ctBad(0);

	while (!input_train.eof())
	{
		getline(input_train, train_line);
		istringstream ss(train_line);
		ss >> chr >> b >> e >> f >> num >> min_over >> entropy >> dc >> err >> bad_frame >> bad_strand;

		if (!isValidChr(chr))		// is this chr in our reference sequence?
			continue;

		int first = atoi(b.c_str());
		int second = atoi(e.c_str());

		if (second-first <= 0)
			continue;

		if(f == "+"){
			donor = first;
			acceptor = second;
			flag = 1;
		}
		if(f == "-"){
			donor = second;
			acceptor = first;
			flag = -1;
		}
		tron = interval(donor, acceptor);
		tron.ori = flag;

		// verify canonical type, and possibly correct tron
		getStrand(tron, jcnType, chr, donor, acceptor);
		switch (jcnType) {
			case INTRON_GT_AG:
				++ctGtAg; break;
			case INTRON_GC_AG:
				++ctGcAg; break;
			case INTRON_AT_AC:
				++ctAtAc; break;
			default:
				++ctBad;
		};

		//if (jcnType != INTRON_GT_AG)
		//	continue;
		all_junction[chr][interval(first,second)].jcnType = jcnType;

		all_junction[chr][interval(first,second)].ori = f;
		is.depthScore = log( atof( num.c_str() ) );
		is.minMapLen = atoi( min_over.c_str() );
		is.minMapLen = abs( 9 - is.minMapLen );
		is.entropy = atof( entropy.c_str() );

		exist_Junction[chr][make_pair(first, second)].count = atoi(num.c_str());
		exist_Junction[chr][make_pair(first, second)].entro = atof(entropy.c_str());

		if(f == "+"){
			exist_Donor[chr][atoi(b.c_str())] += atoi(num.c_str());
			exist_Acceptor[chr][atoi(e.c_str())] += atoi(num.c_str());//change
		}
		if(f == "-"){
			exist_Donor[chr][atoi(e.c_str())] += atoi(num.c_str());
			exist_Acceptor[chr][atoi(b.c_str())] += atoi(num.c_str());
		}

		if (!opts.m_bHeuristic) {
			get_splice_scores( don_score, acc_score, tron, jcnType, chr );
			length_score = lengthScore(chr, tron);
		}

		//if ( length_score==std::numeric_limits<double>::infinity() ) {
		//	std::cout << "infinite length found: " << chr << "\t" << tron.start << "\t" << tron.end << "\t" << tron.ori << std::endl;
		//	continue;
		//}

		//double cov_score = getDiffCov(chr, first, 1) + getDiffCov(chr, second, 2);
		double cov_score = full_cov_first[chr][first] + full_cov_second[chr][second];

		if (tron.ori == -1) {
			is.covDiffLeft = full_cov_second[chr][second];
			is.covDiffRight = full_cov_first[chr][first];
		}
		else {
			is.covDiffLeft = full_cov_first[chr][first];
			is.covDiffRight = full_cov_second[chr][second];
		}

		is.coding = cod_potential;
		is.spliceDon = don_score;
		is.spliceAcc = acc_score;
		is.length = length_score;
		//is.covDiff = cov_score;
		is.invNumCand = 1;
		is.mismatches = atoi( err.c_str() );
		is.iframeErr = atoi( bad_frame.c_str() );
		is.istrandErr = atoi( bad_strand.c_str() );
		//is.igenicState = atoi( genic_state.c_str() );
		is.extScore = 0;

		 //pre-classify based on a subset of feature values
		all_junction[chr][interval(first,second)].stable_partition = -1;
		if ( isNegativeTrain(is) ) {
			is.train = 0;
			is.extScore = 0;
			all_junction[chr][interval(first,second)].stable_partition = 0;
			++ctNegatives;
		}
		else if ( !opts.m_bNoncano && isPositiveTrain(is) ) {
			is.train = 1;
			is.extScore = 1;
			all_junction[chr][interval(first,second)].stable_partition = 1;
			++ctPositives;
		}
		else if ( !opts.m_bNoncano && opts.m_bSoft && isSoftPositiveTrain(is) ) {
			is.train = 1;
			is.extScore = 1;
			all_junction[chr][interval(first,second)].stable_partition = 1;
			++ctPositives;
		}

		is2dv( all_junction[chr][interval(first,second)].score, is );

		++ctTotal;

		all_junction[chr][interval(first,second)].is = is;
	}

	std::cout << ctTotal << " junctions loaded from input file." << std::endl;

	// select training set
	std::cout << ctPositives << " positives and " << ctNegatives << " negatives found for training." << std::endl;
	ctTrain = ctPositives < ctNegatives ? ctPositives : ctNegatives;

	if (ctTrain > ctTrainLimit) ctTrain = ctTrainLimit;

	std::cout << "Intron count by type:" << std::endl;
	std::cout << ctGtAg << " GT_AG introns" << std::endl;
	std::cout << ctGcAg << " GC_AG introns" << std::endl;
	std::cout << ctAtAc << " AT_AC introns" << std::endl;
	std::cout << ctBad << "  unknown introns" << std::endl;

	return ctTrain;
}


void getCandidate(ifstream& input_can)
{
	string can_line, sam_line;
	double don_score, acc_score;

	sInfoScore ifs;

	while(!input_can.eof())
	{
		getline(input_can, can_line);
		if(can_line.length() < 1)
			continue;
		if(can_line == ">")
		{
			int flag = 0;
			getline(input_can, sam_line);
			//			cout << sam_line << endl;
			potJunc junc;
			junc.SAM = sam_line;

			while(!input_can.eof())
			{
				getline(input_can, can_line);
				if(can_line == ">")
					break;

				istringstream ss(can_line);
				string temp;
				ss >> temp;
				int part = atoi(temp.c_str());
				vector<int> in;
				for(int i=1;i<=8;i++){///
					ss >> temp;
					in.push_back(atoi(temp.c_str()));
//					junc.SAM_info[part].push_back(atoi(temp));
				}
				if(atoi(temp.c_str()) > 1){//////////////
					flag = 2;
					break;

				}

				ifs.SAM_info = in;
				junc.blocks[part].push_back( ifs );
				in.clear();
			}		// if !EOF

			if(flag == 2)
			{
				while(!input_can.eof()){
                    getline(input_can, can_line);
                    if(can_line == ">") break;
                }
				continue;
			}
			candidate.push_back(junc);
		}	// if(can_line == ">")
	}	// while(!input_can.eof())

	interval tron(0,0);
	int donor, acceptor;
	intronScores is;

	FILE *pf = fopen("cand","w");

	for (int i=0; i<(int)candidate.size(); i++)
	{
//parrallel here
		//cout << candidate[i].SAM << endl;
		istringstream ss(candidate[i].SAM);
		vector<string> info;
		string temp;
		for(int k = 1; k<= 14; k++){
			ss >> temp;
			info.push_back(temp);
		}
		map<int, vector<int> >local_cigar;
		local_cigar.clear();
		
		// loop over each block of candidate[i]
		for(map<int,vector<sInfoScore> >::iterator ib = candidate[i].blocks.begin(); ib != candidate[i].blocks.end(); ib++)
		{
			double max = 0, splice_score =0, cod_potential = 0;

			map<int, vector<double> >local_score;
			local_score.clear();

			int local_flag = 0;

			// loop over one block of alignments
			for (int t=0; t<(int)ib->second.size(); t++)		// ib->second is vector<sInfoScore>
			{
				vector<double> SCORE;

				int count_junction = 1 , count_donor = 1, count_acceptor = 1;
				double entropy = 0;

				if(ib->second[t].SAM_info[7] == 0){
					tron.start = ib->second[t].SAM_info[0] + ib->second[t].SAM_info[1] + ib->second[t].SAM_info[2];
					tron.end = ib->second[t].SAM_info[0] + ib->second[t].SAM_info[1];
					tron.ori = -1;
				}
				else{
					tron.start = ib->second[t].SAM_info[0] + ib->second[t].SAM_info[1];
					tron.end =  ib->second[t].SAM_info[1] + ib->second[t].SAM_info[2] + ib->second[t].SAM_info[0];
					tron.ori = 1;
				}

				donor = tron.ori==1 ? tron.start : tron.end;
				acceptor = tron.ori==1 ? tron.end : tron.start;

				get_splice_scores( don_score, acc_score, tron, INTRON_GT_AG, info[2] );
				//splice_score = don_score + acc_score;

				double length_score = lengthScore(info[2], tron);

				if(exist_Donor[info[2]].find(tron.start) != exist_Donor[info[2]].end())
					count_donor = exist_Donor[info[2]][tron.start];
				if(exist_Donor[info[2]].find(tron.end) != exist_Acceptor[info[2]].end())
					count_acceptor = exist_Acceptor[info[2]][tron.end];
//coverage input

				int b1 = ib->second[t].SAM_info[1]+ib->second[t].SAM_info[0];
				int b2 = ib->second[t].SAM_info[1]+ib->second[t].SAM_info[0]+ib->second[t].SAM_info[2];

				double first_covdiff = full_cov_first[info[2]][b1];	//getDiffCov(info[2], b1, 1);
				double second_covdiff = full_cov_second[info[2]][b2];	//getDiffCov(info[2], b2, 2);
				
				if (tron.ori == -1) {
					is.covDiffLeft = second_covdiff;
					is.covDiffRight = first_covdiff;
				}
				else {
					is.covDiffLeft = first_covdiff;
					is.covDiffRight = second_covdiff;
				}
		
				if(exist_Junction[info[2]].find(make_pair(b1, b2)) != exist_Junction[info[2]].end()){
					count_junction = exist_Junction[info[2]][make_pair(b1,b2)].count;
					entropy = exist_Junction[info[2]][make_pair(b1,b2)].entro;
				}

				double intron_size = ib->second[t].SAM_info[2] > critical_intron ? log( double(ib->second[t].SAM_info[2] - critical_intron) ) : 0;
				double intron_flag = ib->second[t].SAM_info[2] > critical_intron ? 1 : 0;

				is.coding = cod_potential;
				is.spliceDon = don_score;
				is.spliceAcc = acc_score;
				is.length = length_score;
				//is.covDiff = first_covdiff + second_covdiff;
				//if (fabs(is.covDiff)<1e-9) is.covDiff = 0.;
				is.depthScore = count_junction;
				is.minMapLen = ib->second[t].SAM_info[1] < ib->second[t].SAM_info[3] ? ib->second[t].SAM_info[1] : ib->second[t].SAM_info[3];
				is.entropy = entropy;
				is.invNumCand = double(1.0/double(ib->second.size()));
				is.mismatches = ib->second[t].SAM_info[4] + ib->second[t].SAM_info[5];

				char chOri = tron.ori > 0 ? '+' : '-';

				is.train = 0;
				is2dv( SCORE, is );

				ib->second[t].score_info.score = SCORE;
				ib->second[t].score_info.is = is;

				ib->second[t].score_info.stable_partition = -1;

				//candidate[i].blocks[it->first].push_back(SCORE);

			}	// for (int t=0; t< (int)it->second.size(); t++)
		}	// for(map<int, vector<vector<int> > >::iterator it = candidate[i].SAM_info.begin(); it != candidate[i].SAM_info.end(); it++)
	}	// for (int i=0; i<(int)candidate.size(); i++)
//	Out_train.close();

	fclose( pf );
}


void createPredictionSet()
{
	interval tron(0,0);

	if (opts.m_extTrainFile.empty())
		loadJunctions( opts.m_solidPredFile );

	for(map<string, map<interval, intronInfo> >::iterator it = all_junction.begin(); it != all_junction.end(); it++)
	{
		for(map<interval, intronInfo> :: iterator it2 = it->second.begin(); it2 != it->second.end(); it2++)
		{

			//if (it2->second.is.depthScore < opts.m_low_depth_thresh) {
			//	continue;
			//}
			//else
			if (it2->second.stable_partition == -1)
				sendOutFeatures(Out_predict, it2->second.score);
		}
	}

	Out_predict.close();
}

void createPredictionSetMulti()
{
	Out_predict_multi.open("candidates_multi");

	int nFeat = (int) opts.m_featureSet.length();
	int ct_multi(0);
	double don_score, acc_score;

	for (int i=0; i<(int)candidate.size(); i++)
	{
		for(map<int, vector<sInfoScore> >::iterator ib = candidate[i].blocks.begin(); ib != candidate[i].blocks.end(); ib++)
		{
			for (int j=0; j<(int)ib->second.size(); ++j) {
				//if (nFeat != (int)ib->second[j].score_info.size())
				//	std::cout << "nFeat=" << nFeat << ", yet found " << ib->second[j].score_info.size() << " : " << candidate[i].SAM << std::endl;

				don_score = ib->second[j].score_info.is.spliceDon;
				acc_score = ib->second[j].score_info.is.spliceAcc;

				//if (opts.m_bCoarseGate) {
				//	if ( (isNegative(don_score, acc_score) && ctJcn==1) || isPositive(don_score, acc_score) ) {
				//		continue;
				//	}
				//}

				if (ib->second[j].score_info.stable_partition == -1) {
					++ct_multi;
					sendOutFeatures(Out_predict_multi, ib->second[j].score_info.score);
				}

			}
		}
	}

	Out_predict_multi.close();

	std::cout << "ct_multi = " << ct_multi << std::endl;
}

void loadPredictions(vector<double>& viPredictions, string& fname)
{
	viPredictions.clear();

	FILE *pf = fopen(fname.c_str(),"r");
	double val;
	int idc;
	char pchLab[64], pchDca[64], pchDcb[64];

	fscanf(pf,"%s %s %s\n", pchLab, pchDca, pchDcb);
	std::cout << "lab: " << pchLab << ", dca: " << pchDca << ", dcb: " << pchDcb << std::endl;

	std::cout << "preparing to read prediction file." << std::endl;

	int ct(0);
	while ( fscanf(pf,"%d %s %s\n",&idc, pchDca, pchDcb)==3 )
	{
		val = atof( pchDcb );
		if (ct<20) {
			std::cout << "string: " << pchDcb << ", val:" << val << std::endl;
		}
		++ct;
		viPredictions.push_back( val );
	}

	fclose( pf );

	std::cout << viPredictions.size() << " predictions loaded from file " << fname << std::endl;
}

void getPredictions(vector<double>& viPredictions)
{
	double don_score, acc_score;
	double dcovL, dcovR;
	int ix(0), ct(0);
	double score;

std::cout << "Getting predictions..." << std::endl;

	for(map<string, map<interval, intronInfo> >::iterator it = all_junction.begin(); it != all_junction.end(); it++)
	{
		for(map<interval, intronInfo> :: iterator it2 = it->second.begin(); it2 != it->second.end(); it2++)
		{

			don_score = it2->second.is.spliceDon;
			acc_score = it2->second.is.spliceAcc;
			dcovL = it2->second.is.covDiffLeft;
			dcovR = it2->second.is.covDiffRight;

			score = viPredictions[ix];
			if (it2->second.stable_partition==-1) {
				++ix;
				//if (dcov < -0.25 && score > 0.5) {
				//	it2->second.final_score = score / 2.;
				//}
				//else {
					it2->second.final_score = score;
				//}
			}
			else {
				it2->second.final_score = it2->second.stable_partition==0 ? 0 : 1;
			}
		}
	}

	if (ix != (int)viPredictions.size())
		std::cout << "WARNING: number of predictions does not match number of junctions!!" << std::endl;
}

void getPredictionsMulti(vector<double>& viPredictions)
{
	double don_score, acc_score;
	interval tron(0,0);

	int ix(0);
	double depth, score;

	for (int i=0; i<(int)candidate.size(); i++)
	{
		// loop over alignment blocks (sequence segments)
		for(map<int,vector<sInfoScore> >::iterator ib = candidate[i].blocks.begin(); ib != candidate[i].blocks.end(); ib++)
		{
			// loop over one block of alignments
			for (int t=0; t<(int)ib->second.size(); t++)		// ib->second is vector<sInfoScore>
			{
				if (ix >= (int)viPredictions.size()) { ++ix; break; }

				don_score = ib->second[t].score_info.is.spliceDon;
				acc_score = ib->second[t].score_info.is.spliceAcc;
				depth = ib->second[t].score_info.is.depthScore;

				score = viPredictions[ix];
				if (ib->second[t].score_info.stable_partition==-1) {
					ib->second[t].score_info.final_score = score;
					++ix;
				}
				else {
					ib->second[t].score_info.final_score = ib->second[t].score_info.stable_partition==0 ? 0 : 1;
				}

			}
		}
	}

	if (ix != (int)viPredictions.size()) {
		std::cout << "WARNING (multi): number of predictions does not match number of junctions!!" << std::endl;
		std::cout << "ix = " << ix << ", viPredictions.size() = " << viPredictions.size() << std::endl;
	}
}

/*
void createMulti()
{
	map<int, vector<int> > local_cigar;
	ofstream Out("multi.solid");

	for(int i=0; i<(int)candidate.size(); i++)
	{
		istringstream ss(candidate[i].SAM);
		vector<string> info;
        string temp;

		info.clear();

        for(int k = 1; k<= 14; k++){
                ss >> temp;
                info.push_back(temp);
        }

		double m = 1.;
		double largest_multi = 0;
		local_cigar.clear();

		// loop over blocks (segments)
		for(map<int, vector<sInfoScore> >::iterator ib = candidate[i].blocks.begin(); ib != candidate[i].blocks.end(); ++ib)
		{
			largest_multi = largest_multi > ib->second.size() ? largest_multi : ib->second.size();

			double max = -1, temp;
			int max_index;

			// loop over alignment candidates in a segment
			for (int t=0; t<(int)ib->second.size(); t++)
			{
				temp = ib->second[t].score_info.final_score;
				temp = temp < 0 ? 0 : temp;
				if(temp > max){
					max = temp;
					max_index = t;
				}
			}
			m *= max;
			local_cigar[ib->first] = ib->second[max_index].SAM_info;
		}
		int ali=0;
		ostringstream CIGAR;

        int SAM_start;
        int signalOri;
        int flag = 0;
		int total_mis = 0;

		int cano = -1;

		if (m < 0.5)
			continue;
		
		if (atoi(info[1].c_str()) == 0)
		{//0 or 16
			if (local_cigar.find(1) != local_cigar.end()) {
				SAM_start=local_cigar[1][0];
				signalOri = local_cigar[1][7];

			}

			for (int t=0;t<opts.m_nFragments;t++) {

				if(local_cigar.find(t+1) != local_cigar.end()){
					cano = local_cigar[t+1][6];
					total_mis += local_cigar[1+t][4]+local_cigar[1+t][5];
					if(flag == 0){
						flag = 1;
						SAM_start = local_cigar[t+1][0] - t*opts.m_nSegmentLength;
					}
					signalOri = local_cigar[t+1][7];
					ali+=local_cigar[t+1][1];
					CIGAR << ali <<"M"<<local_cigar[t+1][2]<<"N";

					ali=local_cigar[t+1][3];

				}
				else {
					if (t == opts.m_nFragments-1)
						ali += opts.m_nReadLength-opts.m_nSegmentLength*(opts.m_nFragments-1);
					else
						ali+=opts.m_nSegmentLength;
				}
			}
			CIGAR << ali <<"M";
		}
		else {
			if(local_cigar.find(opts.m_nFragments) != local_cigar.end()) {
				SAM_start=local_cigar[opts.m_nFragments][0];
				signalOri = local_cigar[opts.m_nFragments][7];
				flag = 1;
			}

			for (int t=opts.m_nFragments-1;t>=0;t--) {
				if (local_cigar.find(t+1) != local_cigar.end()) {
					cano = local_cigar[t+1][6];
					total_mis += local_cigar[1+t][4]+local_cigar[1+t][5];
					if (flag == 0 && t != opts.m_nFragments-1) {
						flag = 1;
						SAM_start = local_cigar[t+1][0] - (opts.m_nFragments-2-t)*opts.m_nSegmentLength - (opts.m_nReadLength-opts.m_nSegmentLength*(opts.m_nFragments-1));
					}
					signalOri = local_cigar[t+1][7];
					ali+=local_cigar[t+1][1];
					CIGAR << ali <<"M"<<local_cigar[t+1][2]<<"N";

					ali=local_cigar[t+1][3];

				}
				else {
					if (t == opts.m_nFragments-1)
						ali += opts.m_nReadLength-opts.m_nSegmentLength*(opts.m_nFragments-1);
					else
						ali+=opts.m_nSegmentLength;
				}
			}
			CIGAR << ali <<"M";
		}	// if/else atoi(info[1].c_str()) == 0


		Out << info[0] <<"\t"<<info[1]<<"\t"<<info[2]<<"\t"<<SAM_start<<"\t255\t"<<CIGAR.str()<<"\t"<<cano<<"\t"<<total_mis<<"\t"<<double(1.0/largest_multi)<<"\t"<<info[9]<<"\t"<<info[10]<<"\tNM:i:1\tXX:A:"<<signalOri<<"\tNH:i:"<<m<<"\n";
		//
	}
	Out.close();
}
*/

void auto_train(int ctTrain)
{
	intronScores is;
	intronSample samp;
	int ctP(0), ctN(0);
	int ctLongN(0);
	int irc(0);
	vector<intronSample> visPos, visNeg;

	for(map<string, map<interval, intronInfo> >::iterator it = all_junction.begin(); it != all_junction.end(); it++)
	{
		for(map<interval, intronInfo> :: iterator it2 = it->second.begin(); it2 != it->second.end(); it2++)
		{
			//dv2is( is, it2->second.score );
			is = it2->second.is;
			samp.is = is;
			samp.chr = it->first;
			samp.start = it2->first.start;
			samp.end = it2->first.end;
		
			if ( isNegativeTrain(is) ) {
				//it2->second.stable_partition = 0;
				samp.is.train = 0;
			}
			else if ( isPositiveTrain(is) ) {				
				//it2->second.stable_partition = 1;
				samp.is.train = 1;
			}
			else if ( opts.m_bSoft && isSoftPositiveTrain(is) ) {
				samp.is.train = 1;
			}
			else {
				it2->second.stable_partition = -1;
				samp.is.train = -1;
			}

			if (it2->second.stable_partition != -1) {
				if ( samp.is.train==0 && ctN < ctTrain) {	// || visNegShort.size() < ctTrain/2) ) {
					++ctN;
					visNeg.push_back( samp );
				}
				else if (samp.is.train==1 && ctP < ctTrain) {
					++ctP;
					visPos.push_back( samp );
				}
			}

		}
	}	// all_junction

	std::cout << visPos.size() << " positives and " << visNeg.size() << " negatives found for training." << std::endl;
	ctTrain = visPos.size() < visNeg.size() ? (int)visPos.size() : (int)visNeg.size();

	if (ctTrain > ctTrainLimit) ctTrain = ctTrainLimit;

	for (int i=0; i<ctTrain; ++i)
	{
		sendOutTrain( Out_train, visNeg[i] );
		sendOutTrain( Out_train, visPos[i] );
	}

}

bool parseExtTrainLine(intronSample& samp, string& line)
{
	std::basic_string <char>::size_type L, R;
	
	L = line.find_first_not_of( m_delim );
	
	// skip white space and comments
	if (( L == -1 )||( line[L] == '#' ))
		return false;

	vector< string > cols;
	cols.reserve(16);
	int i;

	R = line.find_first_of( m_delim, L );

	for( i = 0 ; L != -1; ++i )
	{
		cols.push_back( line.substr( L, R - L ) );

		L = line.find_first_not_of( m_delim, R );
		R = line.find_first_of( m_delim, L );
	}

	samp.chr = cols[0];
	samp.start = atoi( cols[1].c_str() );
	samp.end = atoi( cols[2].c_str() );

	samp.is.spliceDon = atof( cols[3].c_str() );
	samp.is.spliceAcc = atof( cols[4].c_str() );
	samp.is.coding = atof( cols[5].c_str() );
	samp.is.covDiffLeft = atof( cols[6].c_str() );
	samp.is.covDiffRight = atof( cols[7].c_str() );
	samp.is.length = atof( cols[8].c_str() );
	samp.is.depthScore = atof( cols[9].c_str() );
	samp.is.minMapLen = atoi( cols[10].c_str() );
	samp.is.entropy = atof( cols[11].c_str() );
	samp.is.mismatches = atoi( cols[12].c_str() );
	samp.is.iframeErr = atoi( cols[13].c_str() );
	samp.is.istrandErr = atoi( cols[14].c_str() );

	double score = atof( cols[15].c_str() );
	samp.is.extScore = score;

	if (score > 1-opts.m_p_delta)
		samp.is.train = 1;
	else if (score < opts.m_p_delta)
		samp.is.train = 0;
	else
		samp.is.train = -1;

	return false;
}

bool ext_train()
{
	intronSample samp;
	int ctP(0), ctN(0);
	int irc(0);
	vector<intronSample> visPos, visNeg;

	std::ifstream data_file( opts.m_extTrainFile.c_str() );

	if ( !data_file.is_open() )
	{
		std::cout << PROG_NAME << " : error, opening file " << opts.m_extTrainFile << "\n"; 
		return true;
	}

	string line;

	while( !data_file.eof() )
	{
		getline( data_file, line );

		if (data_file.fail())
			continue;

		if ( parseExtTrainLine( samp, line ) )
			std::cout << PROG_NAME << " : warning, file " << opts.m_extTrainFile << " line ignored : " << line << "\n";

		if ( samp.is.train==0 && ctN < 7500) {	// || visNegShort.size() < ctTrain/2) ) {
			++ctN;
			visNeg.push_back( samp );
		}
		else if (samp.is.train==1 && ctP < 7500) {
			++ctP;
			visPos.push_back( samp );
		}
	}

	data_file.close();

	int ctNegatives = (int)visNeg.size();
	int ctPositives = (int)visPos.size();

	// select training set
	std::cout << ctPositives << " positives and " << ctNegatives << " negatives found for training." << std::endl;
	ctTrain = ctPositives < ctNegatives ? ctPositives : ctNegatives;

	if (ctTrain > ctTrainLimit) ctTrain = ctTrainLimit;

	for (int i=0; i<ctTrain; ++i)
	{
		sendOutTrain( Out_train, visNeg[i] );
		sendOutTrain( Out_train, visPos[i] );
	}

	return false;
}

void getCov(ifstream& input_fullBWT)
{
	string line, chr, temp1, temp2, begin;
	int b, e;
	string chr_temp = "";
	int b_temp = -1;
	while(!input_fullBWT.eof()){
		getline(input_fullBWT, line);
		if(line.length()==0)  //in case of a blank line
			break;
		istringstream ss(line);
		vector<string> T;
		while (ss)
		{
			string s;
			if (!getline( ss, s, '\t' )) break;
			T.push_back(s);
		}

		chr = T[0];
		b = atoi(T[1].c_str());
		e = atoi(T[2].c_str());

		full_cov_first[chr][b] = atof( T[3].c_str() );
		full_cov_second[chr][b] = atof( T[4].c_str() );
	}
}

void doSvmTraining()
{
	// type 1 training
	string strTrain = "svm-train -b 1 -g ";
	strTrain.append( opts.m_svmGamma );
	strTrain.append( " -c " );
	strTrain.append( opts.m_svmC );

	if (opts.m_bVerbose) std::cout << "SVM training..." << std::endl;
	system("svm-scale -s training.range training_set > training.scaled");
	string cmd = strTrain;
	cmd.append(" training.scaled svm.model > svm_train_log");
	system( cmd.c_str() );					// train
}

void doSvmPrediction()
{
	if (opts.m_bVerbose) std::cout << "forming prediction set..." << std::endl;
	createPredictionSet();
	if (opts.m_bVerbose) std::cout << "SVM predicting..." << std::endl;
	system("svm-scale -r training.range candidates > candidates.scaled");
	system("svm-predict -b 1 candidates.scaled svm.model pred_out > svm_pred_log");
	if (opts.m_bVerbose) std::cout << "incorporating predictions into score..." << std::endl;
	vector<double> nPredictions;
	string fname_pred = "pred_out";
	loadPredictions(nPredictions, fname_pred);
	getPredictions(nPredictions);
	int ctTest(0);

std::cout << "writing final_junctions file..." << std::endl;
	for (map<string, map<interval, intronInfo> >::iterator it = all_junction.begin(); it != all_junction.end(); it++) {
		for (map<interval, intronInfo> :: iterator it2 = it->second.begin(); it2 != it->second.end(); it2++) {
			Out_junction << it->first <<"\t"<<it2->first.start<<"\t"<<it2->first.end<<"\t";

			if (opts.m_bBrief) {
				if (it2->second.stable_partition==-1)
					Out_junction << it2->second.ori << "\t" << it2->second.final_score << "\n";
				else
					Out_junction << it2->second.ori << "\t" << it2->second.stable_partition << "\n";
			}
			else {
				Out_junction << it2->second.is.spliceDon << "\t" <<
								it2->second.is.spliceAcc << "\t" <<
								it2->second.is.coding << "\t" <<
								it2->second.is.covDiffLeft << "\t" << it2->second.is.covDiffRight << "\t" <<
								it2->second.is.length << "\t" <<
								it2->second.is.depthScore << "\t" <<
								it2->second.is.minMapLen << "\t" <<
								it2->second.is.entropy << "\t" <<
								it2->second.is.mismatches << "\t" <<
								it2->second.is.iframeErr << "\t" <<
								it2->second.is.istrandErr << "\t";

				if (it2->second.stable_partition==-1)
					Out_junction << it2->second.final_score << "\n";
				else
					Out_junction << it2->second.stable_partition << "\n";
			}

			// write a test set
			if (ctTest < ctTestLimit /* && it2->second.stable_partition == -1*/) {
				++ctTest;
				Out_junction_test << it->first <<"\t"<<it2->first.start<<"\t"<<it2->first.end<<"\t";
				for(int i = 0; i<(int)it2->second.score.size(); i++)
					Out_junction_test << it2->second.score[i]<<"\t";

				Out_junction_test << "\n";

				sendOutFeatures( Out_test_set, it2->second.is );
			}
		}
	}

}

void doHeuristicPrediction()
{
	for (map<string, map<interval, intronInfo> >::iterator it = all_junction.begin(); it != all_junction.end(); it++) {
		for (map<interval, intronInfo> :: iterator it2 = it->second.begin(); it2 != it->second.end(); it2++) {
			Out_junction << it->first <<"\t"<<it2->first.start<<"\t"<<it2->first.end<<"\t";

			if (it2->second.stable_partition==-1)
				if (it2->second.is.depthScore > 3.) { Out_junction << it2->second.ori << "\t1\n"; }
				else if (it2->second.is.depthScore > 1.6) { Out_junction << it2->second.ori << "\t0.5\n"; }
				else { Out_junction << it2->second.ori << "\t0\n"; }
			else
				Out_junction << it2->second.ori << "\t" << it2->second.stable_partition << "\n";

		}
	}
}

void doSvmPredictionMulti()
{
	if (opts.m_bVerbose) std::cout << "forming multi prediction set..." << std::endl;
	createPredictionSetMulti();

	if (opts.m_bVerbose) std::cout << "SVM predicting (multi)..." << std::endl;
	system("svm-scale -r training.range candidates_multi > candidates_multi.scaled");
	system("svm-predict  -b 1 candidates_multi.scaled svm.model pred_out_multi > svm_pred_multi_log");
	if (opts.m_bVerbose) std::cout << "incorporating predictions into score..." << std::endl;

	string fname_pred = "pred_out_multi";
	vector<double> nPredictions;

	loadPredictions(nPredictions, fname_pred);
	getPredictionsMulti(nPredictions);

	string chr;
	vector<string> info;
	int b1, b2;

	for (int i=0; i<(int)candidate.size(); i++)
	{
		istringstream ss(candidate[i].SAM);
		string temp;
		info.clear();
		for(int k = 1; k<= 14; k++){
				ss >> temp;
				info.push_back(temp);
		}

		for(map<int, vector<sInfoScore > >::iterator ib = candidate[i].blocks.begin(); ib != candidate[i].blocks.end(); ++ib)
		{
			// loop over one block of alignments
			for (int t=0; t<(int)ib->second.size(); t++)		// ib->second is vector<sInfoScore>
			{
				//if (ib->second[t].score_info.empty())
				//	continue;

				b1 = ib->second[t].SAM_info[0] + ib->second[t].SAM_info[1];
				b2 = ib->second[t].SAM_info[0]+ib->second[t].SAM_info[1]+ib->second[t].SAM_info[2];
				Out_junction_multi << info[2] << "\t" << b1 << "\t" << b2 << "\t";
				for (int j=0; j<(int)ib->second[t].score_info.score.size(); ++j)
					Out_junction_multi << ib->second[t].score_info.score[j] << "\t";		// final score is appended to feature vector
				Out_junction_multi << ib->second[t].score_info.final_score << "\n";
			}
		}
	}

	// create multi.solid file
	//createMulti();

	system("svm-scale -r training.range feat_test_set > feat_test_set.scaled");
}


string getBinDir(const char *cmd)
{
	string bin = cmd;
	bin = bin.substr(0, bin.rfind("/")+1);
	return bin;
}


int main(int argc, const char *argv[])
{
	bool bTroubleshoot(false);

	if (opts.parseOptions(argc, argv))
	{
		if (argc > 1) {
			std::cout << "Unacceptable argument(s) given.\n";
			return 0;
		}
		return 0;
	}

	//opts.m_bAddPositives = (opts.m_uiFeatureMask & _FEAT_CODING) != 0;

	BIN = getBinDir( argv[0] );

	string modelDir;
	if (opts.m_modInDir.empty()) {
		modelDir = BIN;
		modelDir += "heuristic";
	}
	else {
		modelDir = opts.m_modInDir;
	}

	if (opts.m_bVerbose) std::cout << "loading models in: " << modelDir << std::endl;
	if ( initSeqScore(modelDir) )
	{
		cout << "error initializing coding model!" << endl;
		cout << endl;
		return 0;
	}

	Out_junction.open("final_junctions");
	if (!opts.m_multiFile.empty()) Out_junction_multi.open("final_junctions_multi");

	Out_junction_test.open("junctions_test_set");		// used for evaluating training strategies (parameters \sigma and C, for example)
	Out_test_set.open("feat_test_set");					
	if (!opts.m_bNoncano) {
		Out_train.open("training_set");						// training set for SVM training
	}
	Out_predict.open("candidates");						// set for prediction

	if (opts.m_bDebug) {
		Out_training_debug.open("training_debug");
	}

// cov diffs
std::cout << "(1) getting covs..." << std::endl;
	ifstream input_fullBWT( "cov_diffs" );
	getCov(input_fullBWT);

// load DNA sequence
std::cout << "(2) getting seq..." << std::endl;
	if (opts.m_bVerbose) std::cout << "loading ref: " << opts.m_seqFile << std::endl;
	ifstream input_ref( opts.m_seqFile.c_str() );
	getRef(input_ref);

// load junctions
std::cout << "(3) getting junctions..." << std::endl;
	int ctTrain = loadJunctions( opts.m_solidTrainFile );

//train model
	if (!opts.m_bNoncano) {
		// train
//std::cout << "(4) training..." << std::endl;
		if (opts.m_extTrainFile.empty()) {
			if (opts.m_bVerbose) std::cout << ctTrain << " training set samples to be taken from: " << opts.m_solidTrainFile << std::endl;
			std::cout << "Auto-training..." << std::endl;
			auto_train( ctTrain );
		}
		else {
			std::cout << "Training from external file " << opts.m_extTrainFile << "..." << std::endl;
			if (ext_train()) {
				std::cout << "Aborting." << std::endl;
				return 0;
			}
		}
		Out_train.close();
	}

	if (opts.m_bHeuristic) {
		doHeuristicPrediction();
		goto FINISH;
	}

	if (!opts.m_bNoncano)
		doSvmTraining();

	if (opts.m_bTrainOnly)
		goto FINISH;

	doSvmPrediction();

	if ( !opts.m_multiFile.empty() )
	{
		if (opts.m_bVerbose) std::cout << "getCandidate: " << opts.m_multiFile << std::endl;
		ifstream input_can( opts.m_multiFile.c_str() );	//argv[1]);
		getCandidate(input_can);
		input_can.close();

		doSvmPredictionMulti();
	}

FINISH:
	Out_junction.close();
	if (!opts.m_multiFile.empty()) Out_junction_multi.close();

	Out_junction_test.close();
	Out_test_set.close();

	if (opts.m_bDebug) {
		Out_training_debug.close();
	}

	return 0;
}

