
#include "targetver.h"
#include <string>

using std::string;


#define _FEAT_CODING		0x000001
#define _FEAT_SPLICE		0x000002
#define _FEAT_DIFF_COV		0x000004
#define _FEAT_LENGTH		0x000010
#define _FEAT_DEPTH			0x000020
#define _FEAT_OVERHANG		0x000040
#define _FEAT_ENTROPY		0x000100
#define _FEAT_AMBIGUITY		0x000200
#define _FEAT_MISMATCH		0x000400
#define _FEAT_SPLICE_DON	0x001000
#define _FEAT_SPLICE_ACC	0x002000
#define _FEAT_STRAND_ERROR	0x004000
#define _FEAT_FRAMESHIFT_ERROR	0x008000

enum { _CLASSIFIER_LR = 0, _CLASSIFIER_SVM };

class cOptions
{
public:
	cOptions()
	{
		m_bVerbose = false;
		m_bVerify = false;

		m_bHeuristic = false;
		m_bBrief = false;

		// defaults
		m_nOrderSite = 0;

		m_nDonWidth = 9;
		m_nDonOffset = 3;
		m_nAccWidth = 21;
		m_nAccOffset = 18;

		m_bAddPositives = true;

		m_nCovThresh = 5;

		m_nType = 1;

		// by default, customized to classify genic introns
		m_svmGamma = "0.002479";
		m_svmC = "0.002479";

		m_nFragments = 0;
		m_nReadLength = 0;
		m_nSegmentLength = 0;

		m_bHuman = false;

		m_featureSet = "csvldoeamth";		// full list of features
		featSet2Mask();
		m_uiSetMask = 0xfffff000;

		m_bPassThru = false;
		m_nPassThruLength = 10000;

		m_bTrainOnly = false;
		m_bCoarseGate = false;

		m_bNoncano = false;
		m_bSoft = false;

		m_p_delta = 0.05;

		m_bDebug = false;
	};
	~cOptions() {  };

	bool parseOptions(int argc, const char** argv);
	void featSet2Mask(void);

public:
	// just make everything accessible for now

	// options
	bool m_bVerify;
	bool m_bVerbose;
	bool m_bHuman;		// human genome, so use special model selection criteria for supervised models
	bool m_bDebug;
	bool m_bNoncano;
	bool m_bHeuristic;
	bool m_bBrief;		// brief output format
	bool m_bSoft;		// weaker conditions for positive set selection (for simulated reads or shallow sequencing)

	bool m_bPassThru;	// just convert input 'solid' file to a 'final_junctions' output file
	bool m_bTrainOnly;	// just find positive and negative sets, and train SVM parameters
	bool m_bCoarseGate;

	string m_extTrainFile;	// external training set provided (used e.g. for iterative training)
	double m_p_delta;
	
	string m_multiFile;
	string m_seqFile;
	string m_solidTrainFile;
	string m_solidPredFile;
	string m_sortedSamFile;
	// HMM model directory
	string m_modInDir;

	int m_nOrderSite;

	int m_nType;

	int m_nDonWidth;
	int m_nAccWidth;
	int m_nDonOffset;
	int m_nAccOffset;

	int m_nCovThresh;
	int m_nPassThruLength;

	bool m_bAddPositives;

	// SVM parameters (strings)
	string m_svmGamma;
	string m_svmC;

	int m_nFragments;
	int m_nReadLength;
	int m_nSegmentLength;

	string m_featureSet;
	unsigned int m_uiFeatureMask;
	unsigned int m_uiSetMask;
};
