
#include "structs.h"
#include "model.h"
#include "options.h"
#include "structs.h"
#include "enum.h"
#include <assert.h>
#include <math.h>
#include <string.h>

#ifdef _WINDOWS_
 #define SLASH '\\'		// Windows
 //#include <sys/types.h>
#else
 #define SLASH '/'		// Unix
#endif

#ifdef _WINDOWS_
	// these includes are necessary to support directory existence checks in windows
	#include <direct.h>
	#include <io.h>
#else
	#include <dirent.h>
#endif

using std::vector;

// struct for managing file sets
typedef struct tagFileRec {
	string strName;
	int size;
} sFileRec;


// globals
vector<cModel> heur_models;
extern cOptions opts;
extern char alphabet[256];

extern void getSeq(string& seq, string chr, int ori, int b, int e);
extern void reset();

//extern FILE *pfDonSite;
//extern FILE *pfAccSite;

extern map<string, string> ref;


int getIx(int gc)
{
	int ix(0);

	if (opts.m_bHuman) {
		if (gc <= 40)
			ix = 0;
		else if (gc > 40 && gc <= 49)
			ix = 1;
		else
			ix = 2;
	}
	else {
		ix = gc - 20;
	}

	// constrain ix to lie withing array bounds
	int nSize = (int) heur_models.size();
	if (ix < 0) ix = 0;
	else if (ix >= nSize) ix = nSize-1;

	return ix;
}

int getGc(interval tron, string& chr)
{
	int b, e;
	if (tron.start < tron.end) { b = tron.start; e = tron.end; }
	else { b = tron.end; e = tron.start; }

	if (ref[chr].empty())
		return 0;

	int dL = b - 1000 >= 0 ? 1000 : b;
	int dR = e + 1000 >= (int)ref[chr].length() ? (int)ref[chr].length() - e : 1000;

	string seqL, seqR;

	//std::cout << "getGc: len=" << ref[chr].length() << std::endl;
	//std::cout << "b=(" << b-dL << "," << b << ")" << std::endl;
	//std::cout << "e=(" << e << "," << e+dR << ")" << std::endl;

	if (b<0 || b>=(int)ref[chr].length())
		return 0;
	else
		seqL = ref[chr].substr(b - dL, dL);

	if (e<0 || e>=(int)ref[chr].length())
		return 0;
	else
		seqR = ref[chr].substr(e, dR);
 

	int i, nGC(0);

	for (i=0; i<(int)seqL.length(); ++i)
	{
		if (seqL[i]=='C' || seqL[i]=='G') ++nGC;
	}

	for (i=0; i<(int)seqR.length(); ++i)
	{
		if (seqR[i]=='C' || seqR[i]=='G') ++nGC;
	}

	int L = (int)seqL.length() + (int)seqR.length();
	double lfGC = 100. * double(nGC) / double(L);

	nGC = int(lfGC);

	// limit to interval [20,70] as per ES convention
	if (nGC < 20)
		nGC = 20;
	if (nGC > 70)
		nGC = 70;

	return nGC;
}

int getGc(int pos, string& chr) { return getGc(interval(pos, pos+1), chr); }

//----------------------------------------------
// a - 0 c - 1 g - 2 t - 3
//----------------------------------------------
void SetAlphabet( char* ptr )
{
	memset( ptr, 4, 256 );

	*(ptr + 'a' ) = 0;
	*(ptr + 'A' ) = 0;
	*(ptr + 'c' ) = 1;
	*(ptr + 'C' ) = 1;
	*(ptr + 'g' ) = 2;
	*(ptr + 'G' ) = 2;
	*(ptr + 't' ) = 3;
	*(ptr + 'T' ) = 3;
	*(ptr + 'u' ) = 3;
	*(ptr + 'U' ) = 3;
};


int getMaxIx(double lf0, double lf1, double lf2)
{
	double lfM;
	int ix(-1);

	if (lf0 > lf1) {
		lfM = lf0; ix = 0;
	}
	else {
		lfM = lf1; ix = 1;
	}

	if (lfM < lf2) {
		lfM = lf2; ix = 2;
	}

	return ix;
}

bool checkMax(int& ix, double& log0_5, double& log1_5, double& log2_5, double& log0_3, double& log1_3, double& log2_3)
{
	ix = -1;

	int i5 = getMaxIx(log0_5, log1_5, log2_5);
	int i3 = getMaxIx(log0_3, log1_3, log2_3);

	ix = getMaxIx(log0_5+log0_3, log1_5+log1_3, log2_5+log2_3);

	return i5==i3;
}


bool getDonSite( string& seqSite, string chr, interval tron )
{
	int b = tron.start - 1;
	if (tron.ori < 0)
		b--;

	if (tron.ori > 0)
		getSeq(seqSite, chr, tron.ori, b-opts.m_nDonOffset, b-opts.m_nDonOffset+opts.m_nDonWidth-1);
	else
		getSeq(seqSite, chr, tron.ori, b-opts.m_nDonWidth+opts.m_nDonOffset+1, b+opts.m_nDonOffset);

	//std::cout << "getDonSite: " << chr << "\t" << tron.ori << "\t" << b << "\t" << seqSite << std::endl;

	//if (pfDonSite != NULL) fprintf(pfDonSite, "%d\t%s\n", tron.ori, seqSite.c_str());

	return false;
}

bool getAccSite( string& seqSite, string chr, interval tron )
{
	int b = tron.end - 1;
	if (tron.ori < 0)
		b--;

	if (tron.ori > 0)
		getSeq(seqSite, chr, tron.ori, b-opts.m_nAccOffset-2, b-opts.m_nAccOffset+opts.m_nAccWidth-1-2);
	else
		getSeq(seqSite, chr, tron.ori, b-opts.m_nAccWidth+opts.m_nAccOffset+1+2, b+opts.m_nAccOffset+2);

	//std::cout << "getAccSite: " << chr << "\t" << tron.ori << "\t" << b << "\t" << seqSite << std::endl;
	//if (pfAccSite != NULL) fprintf(pfAccSite, "%d\t%s\n", tron.ori, seqSite.c_str() );

	return false;
}

bool isValidChr(string chr)
{
	map<string,string>::iterator it = ref.find( chr );
	return it!=ref.end();
}

bool getStrand(interval& tron, int& iJcnType, string chr, int donor, int acceptor)
{
	iJcnType = INTRON_BAD;		// { INTRON_GT_AG, INTRON_GC_AG, INTRON_AT_AC }
	tron.ori = 0;
	string str5, str3;

	int b = donor - 1;
	int e = acceptor - 1;
	int flag(0);

	if (donor < acceptor)
	{
		getSeq(str5, chr, 1, b-1, b+2);
		getSeq(str3, chr, 1, e-3, e);

		// GT_AG
		if (str5.at(1)=='G' && str5.at(2)=='T' &&
			str3.at(1)=='A' && str3.at(2)=='G') {
			tron.ori = 1;
			iJcnType = INTRON_GT_AG;
			return false;
		}
		else if (str5.at(1)=='C' && str5.at(2)=='T' &&
				 str3.at(1)=='A' && str3.at(2)=='C') {
			tron.ori = -1;
			tron.start = acceptor;
			tron.end = donor;
			iJcnType = INTRON_GT_AG;
			return false;
		}
		// GC_AG
		else if (str5.at(1)=='G' && str5.at(2)=='C' &&
				 str3.at(1)=='A' && str3.at(2)=='G') {
			tron.ori = 1;
			tron.start = acceptor;
			tron.end = donor;
			iJcnType = INTRON_GC_AG;
			return false;
		}
		else if (str5.at(1)=='C' && str5.at(2)=='T' &&
				 str3.at(1)=='G' && str3.at(2)=='C') {
			tron.ori = -1;
			tron.start = acceptor;
			tron.end = donor;
			iJcnType = INTRON_GC_AG;
			return false;
		}
		// AT_AC
		else if (str5.at(1)=='G' && str5.at(2)=='T' &&
				 str3.at(1)=='A' && str3.at(2)=='C') {
			tron.ori = 1;
			tron.start = acceptor;
			tron.end = donor;
			iJcnType = INTRON_AT_AC;
			return false;
		}
		else if (str5.at(1)=='G' && str5.at(2)=='T' &&
				 str3.at(1)=='A' && str3.at(2)=='T') {
			tron.ori = -1;
			tron.start = acceptor;
			tron.end = donor;
			iJcnType = INTRON_AT_AC;
			return false;
		}
		else {
			flag = 1;
		}
	}
	else
	{
		getSeq(str5, chr, 1, b-3, b+2);
		getSeq(str3, chr, 1, e-3, e+2);

		// GT_AG
		if (str5.at(1)=='A' && str5.at(2)=='G' &&
			str3.at(3)=='G' && str3.at(4)=='T') {
			iJcnType = INTRON_GT_AG;
			tron.ori = 1;
			tron.start = acceptor;
			tron.end = donor;
			return false;
		}
		else if (str5.at(1)=='A' && str5.at(2)=='C' &&
			     str3.at(3)=='C' && str3.at(4)=='T') {
			tron.ori = -1;
			tron.start = donor;
			tron.end = acceptor;
			iJcnType = INTRON_GT_AG;
			return false;
		}
		// GC_AG
		else if (str5.at(1)=='A' && str5.at(2)=='G' &&
			str3.at(3)=='G' && str3.at(4)=='C') {
			iJcnType = INTRON_GC_AG;
			tron.ori = 1;
			tron.start = acceptor;
			tron.end = donor;
			return false;
		}
		else if (str5.at(1)=='G' && str5.at(2)=='C' &&
			     str3.at(3)=='C' && str3.at(4)=='T') {
			tron.ori = -1;
			tron.start = donor;
			tron.end = acceptor;
			iJcnType = INTRON_GC_AG;
			return false;
		}
		// AT_AG
		else if (str5.at(1)=='A' && str5.at(2)=='C' &&
			str3.at(3)=='A' && str3.at(4)=='T') {
			iJcnType = INTRON_AT_AC;
			tron.ori = 1;
			tron.start = acceptor;
			tron.end = donor;
			return false;
		}
		else if (str5.at(1)=='A' && str5.at(2)=='T' &&
			     str3.at(3)=='G' && str3.at(4)=='T') {
			tron.ori = -1;
			tron.start = donor;
			tron.end = acceptor;
			iJcnType = INTRON_AT_AC;
			return false;
		}
		else {
			flag = 1;
		}
	}

	return false;
}

bool get_splice_scores( double& don_score, double& acc_score, interval tron, int nJcnType, string chr )
{
	string don, acc;
	int ix(0);
	double d0, d1, d2, a0, a1, a2;

	int gc = getGc(tron, chr);
	ix = getIx(gc);

	getDonSite( don, chr, tron );
	getAccSite( acc, chr, tron );
	if (don.length()!=opts.m_nDonWidth || acc.length()!=opts.m_nAccWidth) {
		//if (opts.m_bVerbose) {
		//	std::cout << "empty splice site(s)! chr=" << chr << ", tron=(" << tron.start << "," << tron.end << "," << tron.ori << ")" << std::endl;
		//	std::cout << "don=" << don << ", " << "acc =" << acc << std::endl;
		//}
		if (don.length()!=opts.m_nDonWidth) don_score = -100.;
		if (acc.length()!=opts.m_nAccWidth) acc_score = -100.;
		return false;
	}

	heur_models[ix].scoreDonSite3( d0, d1, d2, don, nJcnType );
	heur_models[ix].scoreAccSite3( a0, a1, a2, acc, nJcnType );

	ix = getMaxIx( d0+a0, d1+a1, d2+a2 );
	don_score = acc_score = 1e-12;
	//switch (ix) {
	//	case 0:
	//		don_score = d0; acc_score = a0; break;
	//	case 1:
	//		don_score = d1; acc_score = a1; break;
	//	case 2:
	//		don_score = d2; acc_score = a2; break;
	//};
	don_score = d0; acc_score = a0;

	//if (fabs(don_score)<1e-6) don_score = 0.;
	//if (fabs(acc_score)<1e-6) acc_score = 0.;

	//if (opts.m_bVerbose) std::cout << "splice scores=(" << don_score << "," << acc_score << ")" << std::endl;

	return false;
}

double lengthScore(string& chr, interval& tron)
{
	double log0 = -20.;
	double pc = 1e-7;		// pseudocount (to prevent log of zero). log(1e-7) \approx -16.1

	int gc = getGc(tron, chr);
	int ix = getIx(gc);

	int len = abs(tron.end - tron.start) + 1;

	//if (opts.m_bVerbose) std::cout << "lengthScore, ix=" << ix << std::endl;

	int lmin = heur_models[ix].m_distrIntronLen.front().i;
	int lmax = heur_models[ix].m_distrIntronLen.back().i;

	//if (opts.m_bVerbose) std::cout << "lmin,lmax = " << lmin << "," << lmax << std::endl;

	int ixh = len - lmin;

	//if (opts.m_bVerbose) std::cout << "len,ixh = " << len << "," << ixh << std::endl;

	double score;

	if (ixh < 0 || ixh > (int)heur_models[ix].m_distrIntronLen.size())
		score = log0;
	else
		score = log( heur_models[ix].m_distrIntronLen[ixh].y + pc );

	//if (opts.m_bVerbose) std::cout << "score = " << score << std::endl;

	return score;
}

bool getFilesInDir(vector<sFileRec>& files, string dir, string strNameTag)
{
	files.clear();

	sFileRec file_record;
	string fname;

#ifdef _WINDOWS_
	// windows directory interrogation library

	int ANY_FILE =  _A_NORMAL | _A_RDONLY | _A_HIDDEN | _A_SYSTEM | _A_ARCH;

	struct _finddatai64_t data;
	// First create the filename that will be use to initialize the find.
	// "*.*" are wild card characters that tells the find function to return a 
	// list of all the files and directories.  You can limit this if you wish
	// to just file with specific extensions, for example "*.txt".  If you do that
	// then finder will not return any directory names.

	// start the finder -- on error _findfirsti64() will return -1, otherwise if no
	// error it returns a handle greater than -1.

	//if (dir.at(dir.length()-1)==SLASH)
	//	fname = dir + wildcard;
	//else
	//{
	//	fname = dir + SLASH;
	//	fname += wildcard;
	//}
	fname = dir + "\\*.*";

	long h = _findfirsti64(fname.c_str(),&data);
	if(h >= 0)
	{
		do {
			if( (data.attrib & ANY_FILE) )
			{
				// its a file
				fname = data.name;
				if (strNameTag.compare("*")==0 || strNameTag.empty())
				{
					file_record.strName = data.name;
					file_record.size = (int)data.size;
					files.push_back( file_record );
				}
				else if (fname.find(strNameTag.c_str(),0)!=string::npos)
				{
					file_record.strName = data.name;
					file_record.size = (int)data.size;
					files.push_back( file_record );
				}
			}
			else
			{
				// ignore
			}
		} while ( _findnexti64(h,&data) == 0);
		// close the find handle.  
		_findclose(h);

	}

#else
	// unix directory interrogation library
	#include <sys/types.h>
	#include <sys/stat.h>

    DIR *dp;
    struct dirent *dirp;
    if((dp  = opendir(dir.c_str())) == NULL) {
		return true;
    }

	string str;
    while ((dirp = readdir(dp)) != NULL)
	{
		str = string(dirp->d_name);

		if (str.size()>0)
		{
			if (str.at(0)!='.')
			{
				struct stat st;
				size_t size;
				if ( !stat( dirp->d_name, &st ) )
					size = st.st_size;

				if (strNameTag.compare("*")==0 || strNameTag.empty())
				{
					file_record.strName = string(dirp->d_name);
					file_record.size = (int) size;
					files.push_back( file_record );
				}
				else if (str.find(strNameTag.c_str(),0)!=string::npos)
				{
					file_record.strName = string(dirp->d_name);
					file_record.size = (int) size;
					files.push_back( file_record );
				}
			}
		}
    }
    closedir(dp);

#endif

	//sort (files.begin(), files.end(), fileRecLessThan);

    return false;
}

bool getFilesInDir(vector<sFileRec>& files, string dir)
{
	string allfiles = "*";
    return getFilesInDir(files, dir, allfiles);
}

bool initSeqScore(string& dir)
{
	vector<sFileRec> files;

	if (getFilesInDir(files, dir))
		return true;

	heur_models.clear();
	cModel mod;
	cSite site;

	mod.init( &site );

	string fullFile;
	int gc(20);

	for (int i=0; i<(int)files.size(); ++i)
	{
		fullFile = dir + SLASH;
		fullFile.append( files[i].strName );

		if (mod.load( fullFile ))
		{
			std::cout << "Error loading file: " << files[i].strName << std::endl;
			return true;
		}

		heur_models.push_back( mod );
		reset();
	}

	// set alphabet
	SetAlphabet( alphabet );

	return false;
}
