
#include "multi.h"
#include "targetver.h"
#include <fstream>
#include <algorithm>
#include <sstream>
#include <iostream>
#include <math.h>

using namespace std;

bool setLessThan(sIntronSet s1, sIntronSet s2)
{
	if (s1.seq.compare(s2.seq)==0) {
		if (s1.set.empty()) return true;
		if (s2.set.empty()) return false;
		return s1.set[0].intron.ibeg < s2.set[0].intron.ibeg;
	}
	else {
		return s1.seq.compare(s2.seq) < 0;
	}
}

void cMulti::loadOld(string& fname)
{
	ifstream input_can(fname.c_str());

	string can_line, sam_line;
	while(!input_can.eof())
	{
		getline(input_can, can_line);
		if(can_line.length() < 1)
			continue;

		if(can_line == ">")
		{
			int flag = 0;
			getline(input_can, sam_line);
			//			cout << sam_line << endl;
			potJunc junc;
			junc.SAM = sam_line;

			while(!input_can.eof())
			{
				getline(input_can, can_line);
				if(can_line == ">")
					break;

				istringstream ss(can_line);
				string temp;
				ss >> temp;
				int part = atoi(temp.c_str());
				vector<int> in;
				for (int i=1; i<=8; i++)
				{
					ss >> temp;
					in.push_back(atoi(temp.c_str()));
				}
				if (atoi(temp.c_str()) > 1){
					flag = 2;
					break;
				}
				
				junc.SAM_info[part].push_back(in);
				in.clear();
			}
			if (flag == 2)
			{
				while(!input_can.eof())
				{
					getline(input_can, can_line);
					if(can_line == ">")
						break;
				}
				continue;
			}
			candidate.push_back(junc);	
		}
	}
}

bool cMulti::load(string& fname)
{
	ifstream input_can(fname.c_str());

	if (!input_can.is_open())
		return true;

	sIntronSet set;
	sIntronCand cand;

	int ctAln(0), ctSet(0);

	cand.bKeep = true;
	int flag;

	string can_line, sam_line;
	while(!input_can.eof())
	{
		getline(input_can, can_line);
		if(can_line.length() < 1)
			continue;

		if(can_line == ">")
		{
			flag = 0;
			getline(input_can, sam_line);
			//			cout << sam_line << endl;
			set.SAM_line = sam_line;
			if (!sam_line.empty())
				set.seq = getSeqName( sam_line );
			else
				break;

			//if (set.seq.compare("Chr1")!=0)
			//	flag = 1;

			while(!input_can.eof())
			{
				getline(input_can, can_line);
				if(can_line == ">")
					break;

				++ctAln;

				istringstream ss(can_line);
				string temp;
				ss >> temp;
				int part = atoi(temp.c_str());
				vector<int> in;
				in.push_back(part);
				for (int i=1; i<=8; i++)
				{
					ss >> temp;
					in.push_back(atoi(temp.c_str()));
				}
				if (atoi(temp.c_str()) > 1){
					flag = 2;
					break;
				}

				if (in.size() != 9)
					flag = 1;
				
				cand.multi_info = in;
				cand.nStatus = _ALN_PROPERTY_BENIGN;
				set.set.push_back( cand );
				//junc.SAM_info[part].push_back(in);
				in.clear();
			}
			if (flag == 2)
			{
				while(!input_can.eof())
				{
					getline(input_can, can_line);
					if(can_line == ">")
						break;
				}
				continue;
			}
			//candidate.push_back(junc);	
		}

		++ctSet;

		m_sets.push_back( set );
		set.set.clear();

	}	// while !eof

	input_can.close();

	std::cout << ctSet << "," << ctAln << " (sets,alignments) loaded from file: " << fname << std::endl;
	return false;
}

bool cMulti::loadSam(string& fname)
{
	std::ifstream data_file( fname.c_str() );

	if ( !data_file.is_open() )
	{
		std::cout << PROG_NAME << " : error, opening file " << fname << "\n"; 
		return true;
	}

	string line;

	while( !data_file.eof() )
	{
		getline( data_file, line );
		if ( parseSamLine( line ) )
			std::cout << PROG_NAME << " : warning, file " << fname << " line ignored : " << line << "\n";
	}

	data_file.close();

	return false;
}

vector<string> cMulti::split(string& line, int maxCols)
{
	std::basic_string <char>::size_type L, R;
	vector< string > cols;

	L = line.find_first_not_of( " \t" );
	
	// skip white space and comments
	if (( L == -1 )||( line[L] == '#' ))
		return cols;

	R = line.find_first_of( " \t", L );

	for( int i = 0 ; ((int)cols.size()<(maxCols-1))||(L == string::npos); ++i )
	{
		cols.push_back( line.substr( L, R - L ) );

		L = line.find_first_not_of( " \t", R );
		R = line.find_first_of( " \t", L );
	}
	// all remaining text
	cols.push_back( line.substr( L, string::npos ) );

	return cols;
}

bool cMulti::parseSamLine( string& line )
{
	alignRec rec;
	//gffLine line_rec;
	std::basic_string <char>::size_type L, R;

	if (line.empty())
		return false;

	if (line.at(0) == '@')
		return false;
	
	L = line.find_first_not_of( " \t" );
	
	// skip white space and comments
	if (( L == -1 )||( line[L] == '#' ))
		return false;

	vector< string > cols;
	cols.reserve(12);

	R = line.find_first_of( " \t", L );

	for( int i = 0 ; (i < 11)&&(L != -1); ++i )
	{
		cols.push_back( line.substr( L, R - L ) );

		L = line.find_first_not_of( " \t", R );
		R = line.find_first_of( " \t", L );
	}

	// all remaining text
	cols.push_back( line.substr( L, string::npos ) );

	if ( cols.size() != 12 )
	{
		std::cout << "Data format error: " << line << std::endl;
		return true;
	}

//typedef struct tagIntronCand
//{
//	vector<int> multi_info;
//
//	sIntron intron;
//	double donScore, accScore;
//	double fs;
//	int nConflict;
//	int nStatus;
//	bool bKeep;
//} sIntronCand;
//
//typedef struct tagIntronSet
//{
//	string SAM_line;
//	string seq;
//	int nContig;
//	bool bGenic;
//	vector<sIntronCand> set;
//} sIntronSet;

	rec.qname = cols[0];			// sequence name
	rec.flag = atoi( cols[1].c_str() );
	rec.rname = cols[2];
	rec.istart = atoi( cols[3].c_str() );		// start coordinate
	rec.mapqual = atoi( cols[4].c_str() );
	rec.cigar = cols[5];

	procCigar( rec );
	
	rec.seq = cols[9];
	rec.qual = cols[10];

	setScore( rec );

	// save new record
	//m_aligns.push_back( rec );

	sIntronSet newset;
	sIntronCand cand;

	newset.bGenic = false;
	newset.SAM_line = line;
	newset.seq = rec.rname;
	cand.accScore = cand.donScore = -1;
	cand.bKeep = true;
	cand.intron.extr_I = cand.intron.extr_L = cand.intron.extr_R = -1;
	cand.nStatus = _ALN_PROPERTY_BENIGN;

	int npart = int(rec.pieces[0].length / n_nSegmentLength) + 1;
	npart = npart>m_nAlnSegments ? m_nAlnSegments : npart;

	int nFlankL = rec.pieces[0].length - (npart-1)*n_nSegmentLength;
	int nFlankR = rec.pieces[1].length-(m_nAlnSegments-npart)*n_nSegmentLength;

	cand.multi_info.clear();
	cand.multi_info.push_back(npart);							// segment no.
	cand.multi_info.push_back(rec.istart);					// beg
	//cand.multi_info.push_back(rec.pieces[0].length);		// aln left flank length
	cand.multi_info.push_back(nFlankL);						// aln left flank length
	cand.multi_info.push_back(rec.pieces[1].start - rec.pieces[0].length);		// intron length
	//cand.multi_info.push_back(rec.pieces[1].length);		// aln right flank length
	cand.multi_info.push_back(nFlankR);						// aln right flank length
	cand.multi_info.push_back(0);							// no. mismatches (left flank)
	cand.multi_info.push_back(0);							// no. mismatches (right flank)
	cand.multi_info.push_back(1);							// 1=canonical, 0=non-canonical
	cand.multi_info.push_back(1);							// 1=pos strand, 0=neg strand

	int flag(0);
	if (rec.pieces.size()>2)
		flag=1;

	for (int i=1; i<(int)rec.pieces.size(); ++i)
	{
		cand.intron.ibeg = rec.istart + rec.pieces[i-1].length;
		cand.intron.iend = rec.istart + rec.pieces[i].start - 1;
		newset.set.push_back(cand);
		// add new "set" (should always be one intron)
		m_sets.push_back( newset );
		// clear temps
		newset.set.clear();
	}

	return false;
}

void cMulti::procCigar(alignRec& rec)
{
	// determine rec.pieces : the regions of the query sequence which cover DNA nt
	//
	size_t pos = rec.cigar.find_first_of(m_cigarChars);
	size_t plast;
	if (pos == string::npos)
		return;

	string s = rec.cigar.substr(0,pos);
	int ilast(-1), n, len(0);
	piece p;
	bool bGap(false);
	
	char ch;

	while (pos != string::npos)
	{
		ch = rec.cigar.at(pos);
		
		switch (ch)
		{
		case 'M': case '=':
			p.start = ilast + 1;
			p.length = atoi( s.c_str() );
			len += p.length;
			ilast = p.start + p.length;
			rec.pieces.push_back( p );
			break;

		case 'I':		// insertion to the reference
			n = atoi( s.c_str() );
			len += n;
			break;

		case 'D':		// deletion from the reference
			break;

		case 'N':		// skipped region from the reference (intron)
			n = atoi( s.c_str() );
			ilast += n - 1;
			bGap = true;
			break;

		case 'S': case 'H':		// soft clipping and hard clipping
			break;

		case 'P':		// padding (silent deletion from padded reference)
			break;

		case 'X':		// sequence mismatch
			p.start = ilast + 1;
			p.length = atoi( s.c_str() );
			len += p.length;
			rec.pieces.push_back( p );
			break;
		};

		plast = pos;
		pos = rec.cigar.find_first_of(m_cigarChars, plast + 1);

		s = rec.cigar.substr(plast + 1,pos);

	}	// while pos != string::npos

	rec.nLength = len;
	rec.bGap = bGap;
}

bool cMulti::loadScoreFile( string& fname )
{
	std::ifstream data_file( fname.c_str() );

	if ( !data_file.is_open() )
	{
		std::cout << PROG_NAME << " : error, opening file " << fname << "\n"; 
		return true;
	}

	string line;

	while( !data_file.eof() )
	{
		getline( data_file, line );

		if (data_file.fail())
			continue;

		if ( parseScoreLine( line ) )
			std::cout << PROG_NAME << " : warning, file " << fname << " line ignored : " << line << "\n";
	}

	data_file.close();


	//std::sort( m_introns.begin(), m_introns.end(), intronLessThan2 );
	//assignContigs();

	return false;
}

bool cMulti::parseScoreLine( string& line )
{
	sIntron intron;
	std::basic_string <char>::size_type L, R;
	
	string delim = "\t ";

	L = line.find_first_not_of( delim );
	
	// skip white space and comments
	if (( L == -1 )||( line[L] == '#' ))
		return false;

	vector< string > cols;
	cols.reserve(16);
	int i;

	R = line.find_first_of( delim, L );

	for( i = 0 ; L != -1; ++i )
	{
		cols.push_back( line.substr( L, R - L ) );

		L = line.find_first_not_of( delim, R );
		R = line.find_first_of( delim, L );
	}

	if ( cols.size() < 4 )
	{
		std::cout << "Data format error: " << line << std::endl;
		return true;
	}

	intron.extr_I = intron.extr_L = intron.extr_R = 0;
	intron.iframe = -2;

	intron.strId = cols[0];			// sequence name
	intron.ibeg = atoi( cols[1].c_str() ) + 1;
	intron.iend = atoi( cols[2].c_str() );
	intron.iStrand = 0;		// unknown for the time being
	intron.nCoverage = 0;
	//intron.score = atof( cols[cols.size()-1].c_str() );

	intron.nFpL = intron.nFpR = -99;

	intron.ibeg += -1;
	intron.iend += -1;

	intron.feat.insert( intron.feat.begin(), cols.begin(), cols.end() );

	sIntronSet newset;
	sIntronCand cand;

	newset.seq = cols[0];
	cand.bKeep = true;
	cand.nStatus = _ALN_PROPERTY_BENIGN;
	cand.donScore = atof( cols[3].c_str() );
	cand.accScore = 0.;
	cand.intron = intron;
	
	newset.set.push_back( cand );
	newset.SAM_line = line;

	// save new record
	m_sets.push_back( newset );

	return false;
};

//void cJunc::assignContigs()
//{
//	m_nContig = -1;
//	m_chrNames.clear();
//
//	for (int i=0; i<(int)m_introns.size(); ++i)
//	{
//		if (findName(m_introns[i].strId))
//			m_introns[i].nContig = m_nContig;
//		else {
//			m_introns[i].nContig = ++m_nContig;
//			m_chrNames.push_back( m_introns[i].strId );
//		}
//	}
//}
//
//bool cJunc::findName(string &name)
//{
//	if (m_chrNames.empty())
//		return false;
//
//	for (int i=m_chrNames.size()-1; i>=0; --i)
//	{
//		if (name.compare( m_chrNames[i] ) == 0)
//			return true;		// found it
//	}
//
//	return false;
//}


void cMulti::setScore(alignRec& rec)
{
	vector<double> vd;
	int i, iq;
	double lf, lscore(0.);

	for (i=0; i<(int)rec.qual.length(); ++i)
	{
		iq = rec.qual.at(i) - 33;
		lf = -10. * log10( double(iq) );
		lscore += lf;
	}

	// normalize by length
	rec.score = lscore / double(rec.qual.length());
}

void cMulti::save(string& fname, int nFilter)
{
	ofstream out(fname.c_str());
	vector<int> info;

	int i, j;
	bool bAny;
	int ctAln(0), ctSet(0);
	int iframeFilt, istrandFilt;

	for (i=0; i<(int)m_sets.size(); ++i)
	{
		if (m_sets[i].bGenic || throwOut(i))	// skip genic sets (they are saved separately)
			continue;

		bAny = false;
		for (j=0; j<(int)m_sets[i].set.size(); ++j)
		{
			if ((m_sets[i].set[j].nStatus & nFilter) == 0) {
				bAny = true;
				break;
			}
		}

		if (!bAny)
			continue;

		ctAln += m_sets[i].set.size();
		++ctSet;

		out << ">" << endl;
		out << m_sets[i].SAM_line << endl;

		for (j=0; j<(int)m_sets[i].set.size(); ++j)
		{
			if ((m_sets[i].set[j].nStatus & nFilter) == 0) {
				info = m_sets[i].set[j].multi_info;

				iframeFilt = (m_sets[i].set[j].nStatus | _ALN_PROPERTY_FRAME_SHIFT) != 0;
				istrandFilt = (m_sets[i].set[j].nStatus | _ALN_PROPERTY_STRAND_CONFLICT) != 0;

				out << info[0] << "\t" << info[1] << "\t" << info[2] << "\t" << info[3] << "\t" << info[4]
					<< "\t" << info[5] << "\t" << info[6] << "\t" << info[7] << "\t" << info[8]
					<< "\t" << iframeFilt << "\t" << istrandFilt << endl;
			}
		}
		out << ">" << endl;
	}

	out.close();
	std::cout << "(" << ctSet << "," << ctAln << ") (sets,alignments) written to file: " << fname << std::endl;
}

void cMulti::saveN(string& fname, int nFilter)
{
	ofstream out(fname.c_str());
	vector<int> info;

	int i, j;
	bool bAny;
	int ctAln(0), ctSet(0);

	for (i=0; i<(int)m_sets.size(); ++i)
	{
		if (m_sets[i].bGenic || throwOut(i))	// skip genic sets (they are saved separately)
			continue;

		bAny = false;
		for (j=0; j<(int)m_sets[i].set.size(); ++j)
		{
			if ((m_sets[i].set[j].nStatus & nFilter) != 0) {
				bAny = true;
				break;
			}
		}

		if (!bAny)
			continue;

		ctAln += m_sets[i].set.size();
		++ctSet;

		out << ">" << endl;
		out << m_sets[i].SAM_line << endl;

		for (j=0; j<(int)m_sets[i].set.size(); ++j)
		{
			if ((m_sets[i].set[j].nStatus & nFilter) != 0) {
				info = m_sets[i].set[j].multi_info;
				out << info[0] << "\t" << info[1] << "\t" << info[2] << "\t" << info[3] << "\t" << info[4]
					<< "\t" << info[5] << "\t" << info[6] << "\t" << info[7] << "\t" << info[8] << endl;
			}
		}
		out << ">" << endl;
	}

	out.close();

	std::cout << "(" << ctSet << "," << ctAln << ") (sets,alignments) written to file: " << fname << std::endl;
}

void cMulti::saveMed(string fname, int nFilter)
{
	ofstream out(fname.c_str());
	vector<int> info;

	int i, j, status, ct(0);
	string sam_line;

	for (i=0; i<(int)m_sets.size(); ++i)
	{
		if (throwOut(i))
			continue;

		for (j=0; j<(int)m_sets[i].set.size(); ++j)
		{
			status = m_sets[i].set[j].nStatus;
			if ((status & nFilter) == 0 && (status & _ALN_PROPERTY_MED_SPLICE_SCORE) != 0) {
				//
				out << ">" << endl;
				getSamLine(sam_line, i);
				out << sam_line << endl;
				//
				info = m_sets[i].set[j].multi_info;
				//getSamInfo(info, i);
				out << info[0] << "\t" << info[1] << "\t" << info[2] << "\t" << info[3] << "\t" << info[4]
					<< "\t" << info[5] << "\t" << info[6] << "\t" << info[7] << "\t" << info[8] << endl;
				out << ">" << endl;

				++ct;
			}
		}
	}
	out.close();

	std::cout << "ga_filter: " << ct << " entries written to file " << fname << std::endl;
}

void cMulti::getSamLine(string& sam_line, int i)
{	// replace start coordinate with "begin"
	// and CIGAR string with "CIGAR"
	string line = m_sets[i].set[0].intron.strId;
	vector<string> blocks = split(line,12);

	sam_line = blocks[0]; sam_line.append("\t"); sam_line.append(blocks[1]); sam_line.append("\t"); sam_line.append(blocks[2]); sam_line.append("\t");
	sam_line.append("begin"); sam_line.append("\t");
	sam_line.append(blocks[4]); sam_line.append("\t");
	sam_line.append("CIGAR"); sam_line.append("\t"); sam_line.append("*"); sam_line.append("\t");
	sam_line.append(blocks[7]); sam_line.append("\t"); sam_line.append(blocks[8]); sam_line.append("\t"); 
	sam_line.append(blocks[9]); sam_line.append("\t"); sam_line.append(blocks[10]); sam_line.append("\t"); 
	sam_line.append(blocks[11]);
}

void cMulti::getSamInfo(vector<int>& info, int i)
{
	info.clear();
	int istrand;
	
	if (m_sets[i].set[0].intron.iStrand==1) istrand = 1;
	else istrand = 0;

	info.push_back(1);							// segment no.
	info.push_back(m_sets[i].set[0].intron.ibeg);		// beg
	info.push_back(10);							// aln left flank length
	info.push_back(m_sets[i].set[0].intron.iend - m_sets[i].set[0].intron.ibeg + 1);		// intron length
	info.push_back(10);							// aln right flank length
	info.push_back(0);							// no. mismatches (left flank)
	info.push_back(0);							// no. mismatches (right flank)
	info.push_back(1);							// 1=canonical, 0=non-canonical
	info.push_back(istrand);					// 1=pos strand, 0=neg strand
}

void cMulti::saveGenic(string fname)
{	// save all fully genic intron sets, which also have no identified frame shift
	ofstream out(fname.c_str());
	vector<int> info;

	int i, j;
	int ctAln(0), ctSet(0);

	for (i=0; i<(int)m_sets.size(); ++i)
	{
		if (!m_sets[i].bGenic || throwOut(i))
			continue;

		out << ">" << endl;
		out << m_sets[i].SAM_line << endl;

		ctAln += m_sets[i].set.size();
		++ctSet;

		for (j=0; j<(int)m_sets[i].set.size(); ++j)
		{
			//if ((m_sets[i].set[j].nStatus & _ALN_PROPERTY_FRAME_SHIFT) == 0) {
				info = m_sets[i].set[j].multi_info;
				out << info[0] << "\t" << info[1] << "\t" << info[2] << "\t" << info[3] << "\t" << info[4]
					<< "\t" << info[5] << "\t" << info[6] << "\t" << info[7] << "\t" << info[8] << endl;
			//}
		}
		out << ">" << endl;
	}

	out.close();

	std::cout << "(" << ctSet << "," << ctAln << ") (sets,alignments) written to file: " << fname << std::endl;
}

void cMulti::saveScore(string fname)
{
	ofstream out(fname.c_str());
	vector<int> info;

	int i, j, k, ct(0);
	int iframeFilt, istrandFilt;

	for (i=0; i<(int)m_sets.size(); ++i)
	{
		for (j=0; j<(int)m_sets[i].set.size(); ++j)
		{
			++ct;
			out << m_sets[i].set[j].intron.feat[0];
			for (k=1; k<(int)m_sets[i].set[j].intron.feat.size(); ++k)
				out << "\t" << m_sets[i].set[j].intron.feat[k];

			iframeFilt = (m_sets[i].set[j].nStatus & _ALN_PROPERTY_FRAME_SHIFT) != 0;
			istrandFilt = (m_sets[i].set[j].nStatus & _ALN_PROPERTY_STRAND_CONFLICT) != 0;
			//if ( (m_sets[i].set[j].nStatus & _ALN_PROPERTY_CODING) != 0 )
			//	icodingFilt = 2;
			//else if ( (m_sets[i].set[j].nStatus & _ALN_PROPERTY_PARTIAL_CODING) != 0 )
			//	icodingFilt = 1;
			//else
			//	icodingFilt = 0;

			//out << "\t" << iframeFilt << "\t" << istrandFilt << "\t" << icodingFilt << "\n";
			out << "\t" << iframeFilt << "\t" << istrandFilt << "\n";
		}
	}
	out.close();

	std::cout << "gapMark: " << ct << " lines written to file " << fname << std::endl;
}

void cMulti::padZeros(string fname)
{
	ofstream out(fname.c_str());
	vector<int> info;

	int i, j, k, ct(0);
	int iframeFilt, istrandFilt;

	for (i=0; i<(int)m_sets.size(); ++i)
	{
		for (j=0; j<(int)m_sets[i].set.size(); ++j)
		{
			++ct;
			out << m_sets[i].set[j].intron.feat[0];
			for (k=1; k<(int)m_sets[i].set[j].intron.feat.size(); ++k)
				out << "\t" << m_sets[i].set[j].intron.feat[k];

			out << "\t" << 0 << "\t" << 0 << "\n";
		}
	}
	out.close();

	std::cout << "gapMark: " << ct << " lines written to file " << fname << std::endl;
}

void cMulti::saveGenicSam(string fname, int nFilter)
{	// save all fully genic intron sets, which also pass all filters
	ofstream out(fname.c_str());
	vector<int> info;

	int i, j, ct(0);
	bool bAny;

	for (i=0; i<(int)m_sets.size(); ++i)
	{
		if (!m_sets[i].bGenic || throwOut(i))
			continue;

		bAny = false;
		for (j=0; j<(int)m_sets[i].set.size(); ++j)
		{
			if (m_sets[i].set[j].nStatus & _ALN_PROPERTY_NONCANONICAL)	// skip non-canonical introns here
				bAny = true;
		}

		if (bAny) continue;

		for (j=0; j<(int)m_sets[i].set.size(); ++j)
		{
			//if (m_sets[i].set[j].nStatus & _ALN_PROPERTY_NONCANONICAL)	// skip non-canonical introns here
			//	continue;

			if ((m_sets[i].set[j].nStatus & nFilter) == 0) {
				out << m_sets[i].SAM_line << endl;
				++ct;
			}
		}
	}
	out.close();

	std::cout << "ga_filter: " << ct << " SAM lines written to file " << fname << std::endl;
}


void cMulti::saveNonSam(string fname, int nFilter)
{	// save all non-genic intron sets, which also pass all filters
	ofstream out(fname.c_str());
	vector<int> info;

	int i, j, ct(0);
	bool bAny;

	for (i=0; i<(int)m_sets.size(); ++i)
	{
		if (m_sets[i].bGenic || throwOut(i))
			continue;

		bAny = false;
		for (j=0; j<(int)m_sets[i].set.size(); ++j)
		{
			if (m_sets[i].set[j].nStatus & _ALN_PROPERTY_NONCANONICAL)	// skip non-canonical introns here
				bAny = true;
		}

		if (bAny) continue;

		for (j=0; j<(int)m_sets[i].set.size(); ++j)
		{
			//if (m_sets[i].set[0].nStatus & _ALN_PROPERTY_NONCANONICAL)	// skip non-canonical introns here
			//	continue;

			if ((m_sets[i].set[j].nStatus & nFilter) == 0) {
				out << m_sets[i].SAM_line << endl;
				++ct;
			}
			//else
			//	flag=1;
		}
	}
	out.close();

	std::cout << "ga_filter: " << ct << " SAM lines written to file " << fname << std::endl;
}

void cMulti::saveGenicSamN(string fname, int nFilter)
{	// save all fully genic intron sets, which also fail any of the filters
	ofstream out(fname.c_str());
	vector<int> info;

	int i, j, ct(0);
	bool bAny;

	for (i=0; i<(int)m_sets.size(); ++i)
	{
		if (!m_sets[i].bGenic || throwOut(i))
			continue;

		bAny = false;
		for (j=0; j<(int)m_sets[i].set.size(); ++j)
		{
			if ((m_sets[i].set[j].nStatus & nFilter) != 0) {
				bAny = true;
				break;
			}
		}

		if (!bAny)
			continue;

		for (j=0; j<(int)m_sets[i].set.size(); ++j)
		{
			if ((m_sets[i].set[0].nStatus & nFilter) != 0) {
				out << m_sets[i].SAM_line << endl;
				++ct;
			}
			//else
			//	flag=1;
		}
	}
	out.close();

	std::cout << "ga_filter: " << ct << " SAM lines written to file " << fname << std::endl;
}


void cMulti::saveSam(string& fname, int nFilter)
{
	ofstream out(fname.c_str());
	vector<int> info;

	int i, j;
	bool bAny;

	for (i=0; i<(int)m_sets.size(); ++i)
	{
		if (throwOut(i))	// skip genic sets (they are saved separately)
			continue;

		bAny = false;
		for (j=0; j<(int)m_sets[i].set.size(); ++j)
		{
			if ((m_sets[i].set[j].nStatus & nFilter) == 0) {
				bAny = true;
				break;
			}
		}

		if (!bAny)
			continue;

		for (j=0; j<(int)m_sets[i].set.size(); ++j)
		{
			if ((m_sets[i].set[j].nStatus & nFilter) == 0)
				out << m_sets[i].SAM_line << endl;
		}
	}

	out.close();
}

void cMulti::saveSamN(string& fname, int nFilter)
{
	ofstream out(fname.c_str());
	vector<int> info;

	int i, j, ct(0);
	bool bAny;

	for (i=0; i<(int)m_sets.size(); ++i)
	{
		if (throwOut(i))
			continue;

		bAny = false;
		for (j=0; j<(int)m_sets[i].set.size(); ++j)
		{
			if ((m_sets[i].set[j].nStatus & nFilter) != 0) {
				bAny = true;
				break;
			}
		}

		if (!bAny)
			continue;

		for (j=0; j<(int)m_sets[i].set.size(); ++j)
		{
			if ((m_sets[i].set[j].nStatus & nFilter) != 0) {
				out << m_sets[i].SAM_line << endl;
				++ct;
			}
		}
	}

	out.close();

	std::cout << "ga_filter: " << ct << " SAM lines written to file " << fname << std::endl;
}

void cMulti::add(vector<sIntronSet>& sets)
{
	m_sets.insert( m_sets.end(), sets.begin(), sets.end() );
}

vector<sIntronSet> cMulti::getIntronSets(string& seqName, int iContig)
{
	vector<sIntronSet> sets;

	for (int i=0; i<(int)m_sets.size(); ++i)
	{
		if (m_sets[i].seq.compare(seqName)==0) {
			m_sets[i].nContig = iContig;
			sets.push_back( m_sets[i] );
		}
	}

	return sets;
}

void cMulti::sortSets()
{
	std::sort( m_sets.begin(), m_sets.end(), setLessThan );
}

string cMulti::getSeqName(string& sam_line)
{
	// first col
	size_t pos_L = sam_line.find_first_not_of(" \t");
	size_t pos_R = sam_line.find_first_of(" \t",pos_L);
	// second col
	pos_L = sam_line.find_first_not_of(" \t",pos_R);
	pos_R = sam_line.find_first_of(" \t",pos_L);
	// third col
	pos_L = sam_line.find_first_not_of(" \t",pos_R);
	pos_R = sam_line.find_first_of(" \t",pos_L);

	string name = sam_line.substr(pos_L, pos_R - pos_L);
	return name;
}

bool cMulti::throwOut(int ix)
{	// return true if no candidates in set ix are to be kept
	if (ix<0 || ix>=(int)m_sets.size())
		return true;

	return false;

	//bool bThrow = true;

	//for (int i=0; i<m_sets[ix].set.size(); ++i)
	//{
	//	if (m_sets[ix].set[i].bKeep) return false;
	//}

	//return bThrow;
}

//void cMulti::saveGenicWeakN(string fname, int nFilterWeak, int nFilterStrong)
//{	// save all fully genic intron sets which pass weak filter, but fail strong one
//	ofstream out(fname.c_str());
//	vector<int> info;
//
//	int i, j, status;
//	bool bAny;
//
//	// file starts with a single ">" character
//	out << ">" << endl;
//
//	for (i=0; i<m_sets.size(); ++i)
//	{
//		if (!m_sets[i].bGenic || throwOut(i))
//			continue;
//
//		out << m_sets[i].SAM_line << endl;
//
//		for (j=0; j<m_sets[i].set.size(); ++j)
//		{
//			status = m_sets[i].set[j].nStatus;
//			if ((status & nFilterStrong) == 0 && (status & nFilterWeak) != 0) {
//				info = m_sets[i].set[j].multi_info;
//				out << info[0] << "\t" << info[1] << "\t" << info[2] << "\t" << info[3] << "\t" << info[4]
//					<< "\t" << info[5] << "\t" << info[6] << "\t" << info[7] << "\t" << info[8] << endl;
//			}
//		}
//		out << ">" << endl;
//		out << ">" << endl;
//	}
//
//	out.close();
//}
