#pragma once

#include "interval.h"
#include <vector>
#include <string>
#include <iostream>
#include "targetver.h"

using std::vector;
using std::string;

enum { _FEATURE_UNDEFINED = -1, _FEATURE_EXON, _FEATURE_CDS, _FEATURE_INTRON, _FEATURE_ALIGNMENT };


// feature class
//
template<class T>
class cFeature
{
public:
	cFeature() { };
	~cFeature() { };

	string getFeatureString(void) const;
	int getFrame(int ix) const;
	string getFrameString(int ix) const;

	void writeGff(std::ostream& os, string& seqName, char chStrand, string& strParent);

	void add(int beg, int end, int type);
	void add(int beg, int end);

	bool empty() const { return m.empty(); }
	int numel() const { return (int) m.size(); }
	void clear() { m.clear(); m_nType = _FEATURE_UNDEFINED; }

	int length() const;
	int lengthBetween() const;

	void setStrand(char strand) { m_chStrand = strand; }
			
public:
	int m_nType;
	char m_chStrand;
	//string m_strParent;
	vector<T> m;
};

template<class T>
string cFeature<T>::getFeatureString() const
{
	// get feature
	string strFeature = "undef";

	switch (m_nType) {
		case _FEATURE_EXON:
			strFeature = "exon";
			break;
		case _FEATURE_CDS:
			strFeature = "CDS";
			break;
		case _FEATURE_INTRON:
			strFeature = "intron";
			break;
		case _FEATURE_ALIGNMENT:
			strFeature = "align";
			break;
	};

	return strFeature;
}

template<class T>
int cFeature<T>::getFrame(int ix) const
{
	int iframe = -1;

	if (ix<0 || ix>=(int)m.size())
		return iframe;

	switch (m[ix].nType) {
		case _INTERVAL_UNDEFINED: case _INTERVAL_INTRON_NF: case _INTERVAL_EXON_NF:
			iframe = -1;
			break;
		case _INTERVAL_EXON_0: case _INTERVAL_INTRON_0:
			iframe = 0;
			break;
		case _INTERVAL_EXON_1: case _INTERVAL_INTRON_1:
			iframe = 1;
			break;
		case _INTERVAL_EXON_2: case _INTERVAL_INTRON_2:
			iframe = 2;
			break;
	};

	return iframe;
}

template<class T>
string cFeature<T>::getFrameString(int ix) const
{
	int iframe = getFrame(ix);
	string str = ".";

	switch (iframe) {
		case 0:
			str = "0";
			break;
		case 1:
			str = "1";
			break;
		case 2:
			str = "2";
			break;
		case 3:
			str = "3";
			break;
	};

	return str;
}

template<class T>
int cFeature<T>::length() const
{
	int len(0);

	for (int i=0; i<(int)m.size(); ++i)
		len += m[i].second() - m[i].first() + 1;

	return len;
}

template<class T>
int cFeature<T>::lengthBetween() const
{
	int len(0);

	for (int i=1; i<(int)m.size(); ++i)
		len += m[i].first() - m[i-1].second() - 1;

	return len;
}

template<class T>
void cFeature<T>::add(int beg, int end, int type)
{
	T t(beg, end, type);
	m.push_back(t);
}

template<class T>
void cFeature<T>::add(int beg, int end)
{
	T t(beg, end);
	m.push_back(t);
}

template<class T>
void cFeature<T>::writeGff(std::ostream& os, string& seqName, char chStrand, string& strParent)
{
	string feat = getFeatureString();

	for (int i=0; i<(int)m.size(); ++i) {
		os << seqName << "\t"
			<< PROG_NAME << "\t"
			<< feat << "\t"
			<< m[i].first() << "\t"
			<< m[i].second() << "\t"
			<< "." << "\t"
			<< chStrand << "\t"
			<< getFrameString(i) << "\t"
			<< "Parent=" << strParent << ";" << std::endl;
	}
}

