
#pragma once

#include <vector>

using std::pair;
using std::vector;

enum { _INTERVAL_UNDEFINED = -1, _INTERVAL_EXON_0, _INTERVAL_EXON_1, _INTERVAL_EXON_2, _INTERVAL_EXON_NF,
		_INTERVAL_INTRON_0, _INTERVAL_INTRON_1, _INTERVAL_INTRON_2, _INTERVAL_INTRON_NF };


// labelled interval class
//
class cInterval
{
public:
	cInterval() { };
	cInterval(int beg, int end) {
		nType = _INTERVAL_UNDEFINED;
		m = std::make_pair(beg,end);
	}
	cInterval(int beg, int end, int type) {
		nType = type;
		m = std::make_pair(beg,end);
	}
	~cInterval() { };

	bool operator<(cInterval n) const {
		if (m.first < n.m.first)
			return true;
		else if (m.first == n.m.first)
			return m.second < n.m.second;
		else return false;
	}
	bool operator>(cInterval n) const {
		if (m.first > n.m.first)
			return true;
		else if (m.first == n.m.first)
			return m.second > n.m.second;
		else return false;
	}
	bool operator==(cInterval n) const {
		 return m.first==n.m.first && m.second==n.m.second;
	}

	bool isOverlapping(int pos) const {
		return !(m.first > pos || m.second < pos);
	}

	int length() const { return m.second - m.first + 1; }
	int first() const { return m.first; }
	int second() const { return m.second; }

public:
	int nType;
	pair<int,int> m;
};


class cRichInterval : public cInterval
{
public:
	cRichInterval() { m_lenExonL = m_lenExonR = 0; };
	cRichInterval(cInterval n) { m = n.m; nType = n.nType; m_lenExonL = m_lenExonR = 0; }
	~cRichInterval() { };

	int m_lenExonL;
	int m_lenExonR;
};
