
#pragma once

#include "feature.h"
#include <string>
#include <limits>
#include <algorithm>
#include <iostream>
#include "targetver.h"

using std::string;

class cIsoform
{
public:
	cIsoform() { m_ibeg = m_iend = -1; m_cdsBeg = m_cdsEnd = -1; m_chStrand = '*'; };
	~cIsoform() { };

	bool operator<(cIsoform iso) const {
		if (m_ibeg == iso.m_ibeg)
			return m_iend < iso.m_iend;
		else return m_ibeg < iso.m_ibeg;
	}
	bool operator>(cIsoform iso) const {
		if (m_ibeg == iso.m_ibeg)
			return m_iend > iso.m_iend;
		else return m_ibeg > iso.m_ibeg;
	}

	void writeGff( std::ostream& os, string& seqName );

	void addTrans(int ibeg, int iend, int iframe);
	void addTrans(int ibeg, int iend, string& frame);
	void addCoding(int ibeg, int iend, string& frame);

	// get
	void getExons(vector<cInterval>& vfeat) const;
	void getIntrons(vector<cInterval>& vfeat) const;
	void getIntrons(vector<cRichInterval>& vintrons) const;
	void getCodingIntrons(vector<cInterval>& vintrons) const;
	void getCodingIntrons(vector<cRichInterval>& vintrons) const;
	char getStrand(void) const { return m_chStrand; }
	int getCdsBeg() const { return m_cdsBeg; }
	int getCdsEnd() const { return m_cdsEnd; }

	// set
	void setId(string& id) { m_strId = id; }
	void setParent(string& parent) { m_strParent = parent; }
	void setStrand(char ch) { m_chStrand = ch; m_transcript.setStrand(ch); m_coding.setStrand(ch); }
	void setFrames(void);
	void setLimits(void) {
		sort();
		if (!m_transcript.empty()) {
			m_ibeg = m_transcript.m.front().first();
			m_iend = m_transcript.m.back().second();
		}
		else if (!m_coding.empty()) {
			m_ibeg = m_coding.m.front().first();
			m_iend = m_coding.m.back().second();
		}
		else { m_ibeg = m_iend = -1; }

		if (!m_coding.empty()) {
			m_cdsBeg = m_coding.m.front().first();
			m_cdsEnd = m_coding.m.back().second();
		}
		else { m_cdsBeg = m_cdsEnd = -1; }
	}
	void getIntronTypes(vector<cInterval>& vintrons) const;
	void getIntronTypes(vector<cRichInterval>& vrintrons) const;

	int lengthTrans() const { return m_transcript.length(); }
	int lengthCoding() const { return m_coding.length(); }
	int lengthIntrons() const {
		int lT = m_transcript.lengthBetween();
		int lC = m_coding.lengthBetween();
		return lT >= lC ? lT : lC;
	}

	int ctExons() const { return (int)m_transcript.numel(); }
	int ctCds() const { return (int)m_coding.numel(); }

	bool isOverlapping(int ibeg, int iend) const {
		if (m_iend < ibeg || m_ibeg > iend) return false;
		else return true;
	}

	void deriveCoding(int cdsBeg, int cdsEnd);

	void sort(void) {
		std::sort( m_transcript.m.begin(), m_transcript.m.end() );
		std::sort( m_coding.m.begin(), m_coding.m.end() );
	}

	void clear(void) { m_strId = string(""); m_strParent = string(""); m_transcript.clear(); m_coding.clear(); }

public:
	string m_strId;
	string m_strParent;
	char m_chStrand;
	int m_ibeg, m_iend;
	int m_cdsBeg, m_cdsEnd;
	cFeature<cInterval> m_transcript;
	cFeature<cInterval> m_coding;
};


void cIsoform::deriveCoding(int cdsBeg, int cdsEnd)
{	// given CDS start and stop, derive a m_coding member,
	//  and set appropriate interval types (e.g. frame) for all exons
	m_coding.clear();

	if (cdsBeg==cdsEnd)		// NOP
		return;

	int i, ixb, ixe;
	for (i=0; i<m_transcript.numel(); ++i)
	{
		if (m_transcript.m[i].second() < cdsBeg || m_transcript.m[i].first() > cdsEnd)
			m_transcript.m[i].nType = _INTERVAL_EXON_NF;
		else
		{
			if (m_transcript.m[i].isOverlapping(cdsBeg))
				ixb = i;
			if (m_transcript.m[i].isOverlapping(cdsEnd))
				ixe = i;
		}
	}

	int nBeg, nEnd;

	if (ixb == ixe) {
		m_coding.add(cdsBeg, cdsEnd, _INTERVAL_EXON_0);
		setFrames();
		return;
	}
	else {
		nEnd = m_transcript.m[ixb].second();
		m_coding.add(cdsBeg, nEnd, _INTERVAL_EXON_NF);
	}

	for (i=ixb+1; i<ixe; ++i)
	{
		nBeg = m_transcript.m[i].first();
		nEnd = m_transcript.m[i].second();
		m_coding.add(cdsBeg, nEnd, _INTERVAL_EXON_NF);
	}

	nBeg = m_transcript.m[ixe].first();
	m_coding.add(nBeg, cdsEnd, _INTERVAL_EXON_NF);

	setFrames();
}

void cIsoform::setFrames()
{
	if (m_coding.empty()) return;

	int i = m_chStrand=='+' ? 0 : m_coding.numel()-1;
	m_coding.m[i].nType = _INTERVAL_EXON_0;

	int ie = 0, start = m_coding.m[i].first();
	bool bFound(false);
	while (ie<m_transcript.numel()) {
		if (m_transcript.m[ie].isOverlapping(start)) { bFound = true; break; }
		++ie;
	}
	if (!bFound) {	// !! problem with m_transcript !!
		std::cout << "cIsoform::setFrames--discord between m_transcript and m_coding!!" << std::endl;
		return;
	}

	int ilast = i;
	if (m_chStrand=='+') {++i; ++ie;} else {--i; --ie;}

	int cds(0), ph;

	while (i>=0 && i<m_coding.numel())
	{
		cds += m_coding.m[ilast].length();
		ph = cds % 3;
		switch (ph) {
			case 0:
				m_coding.m[i].nType = m_transcript.m[ie].nType = _INTERVAL_EXON_0; break;
			case 1:
				m_coding.m[i].nType = m_transcript.m[ie].nType = _INTERVAL_EXON_1; break;
			case 2:
				m_coding.m[i].nType = m_transcript.m[ie].nType = _INTERVAL_EXON_2; break;
			default:
				m_coding.m[i].nType = m_transcript.m[ie].nType = _INTERVAL_EXON_NF;
		}
		ilast = i;
		if (m_chStrand=='+') {++i; ++ie;} else {--i; --ie;}
	}
}

void cIsoform::addTrans(int ibeg, int iend, int iframe)
{
	switch (iframe) {
		case 0:
			m_transcript.add(ibeg, iend, _INTERVAL_EXON_0); break;
		case 1:
			m_transcript.add(ibeg, iend, _INTERVAL_EXON_1); break;
		case 2:
			m_transcript.add(ibeg, iend, _INTERVAL_EXON_2); break;
		default:
			m_transcript.add(ibeg, iend, _INTERVAL_EXON_NF);
	};
}

void cIsoform::addTrans(int ibeg, int iend, string& frame)
{
	if (frame=="0")
		m_transcript.add(ibeg, iend, _INTERVAL_EXON_0);
	else if (frame=="1")
		m_transcript.add(ibeg, iend, _INTERVAL_EXON_1);
	else if (frame=="2")
		m_transcript.add(ibeg, iend, _INTERVAL_EXON_2);
	else
		m_transcript.add(ibeg, iend, _INTERVAL_EXON_NF);
}

void cIsoform::addCoding(int ibeg, int iend, string& frame)
{
	if (frame=="0")
		m_coding.add(ibeg, iend, _INTERVAL_EXON_0);
	else if (frame=="1")
		m_coding.add(ibeg, iend, _INTERVAL_EXON_1);
	else if (frame=="2")
		m_coding.add(ibeg, iend, _INTERVAL_EXON_2);
	else
		m_coding.add(ibeg, iend, _INTERVAL_EXON_NF);
}

void cIsoform::getExons(vector<cInterval>& vfeat) const
{
	vfeat.clear();
	vfeat.insert( vfeat.begin(), m_transcript.m.begin(), m_transcript.m.end() );
}

void cIsoform::getCodingIntrons(vector<cInterval>& vintrons) const
{	// WARNING: exons MUST be sorted!
	vintrons.clear();

	cInterval intron;
	int ixf;
	int type = _INTERVAL_INTRON_NF;

	for (int i=1; i<m_coding.numel(); ++i)
	{
		ixf = m_chStrand=='+' ? i : i-1;	// intron phase takes on phase of downstream exon
		switch (m_coding.m[ixf].nType) {
			case _INTERVAL_EXON_0:
				type = _INTERVAL_INTRON_0; break;
			case _INTERVAL_EXON_1:
				type = _INTERVAL_INTRON_1; break;
			case _INTERVAL_EXON_2:
				type = _INTERVAL_INTRON_2; break;
			default:
				type = _INTERVAL_INTRON_NF;
		};
		intron = cInterval( m_coding.m[i-1].second()+1, m_coding.m[i].first()-1, type );
		vintrons.push_back( intron );
	}
}

void cIsoform::getCodingIntrons(vector<cRichInterval>& vintrons) const
{	// WARNING: exons MUST be sorted!
	vintrons.clear();

	cRichInterval intron;
	int ixf;
	int type = _INTERVAL_INTRON_NF;

	for (int i=1; i<m_coding.numel(); ++i)
	{
		ixf = m_chStrand=='+' ? i : i-1;	// intron phase takes on phase of downstream exon
		switch (m_coding.m[ixf].nType) {
			case _INTERVAL_EXON_0:
				type = _INTERVAL_INTRON_0; break;
			case _INTERVAL_EXON_1:
				type = _INTERVAL_INTRON_1; break;
			case _INTERVAL_EXON_2:
				type = _INTERVAL_INTRON_2; break;
			default:
				type = _INTERVAL_INTRON_NF;
		};
		intron = cRichInterval( cInterval(m_coding.m[i-1].second()+1, m_coding.m[i].first()-1, type) );
		intron.m_lenExonL = m_coding.m[i-1].length();
		intron.m_lenExonR = m_coding.m[i].length();
		vintrons.push_back( intron );
	}
}

void cIsoform::getIntrons(vector<cInterval>& vintrons) const
{	// WARNING: exons MUST be sorted!
	vintrons.clear();

	if (m_transcript.empty())
		return getCodingIntrons(vintrons);

	cInterval intron;
	int type = _INTERVAL_INTRON_NF;

	for (int i=1; i<m_transcript.numel(); ++i)
	{
		intron = cInterval( m_transcript.m[i-1].second()+1, m_transcript.m[i].first()-1, type );
		vintrons.push_back( intron );
	}

	getIntronTypes( vintrons );
}

void cIsoform::getIntrons(vector<cRichInterval>& vintrons) const
{	// WARNING: exons MUST be sorted!
	vintrons.clear();

	if (m_transcript.empty())
		return getCodingIntrons(vintrons);

	cInterval intron;
	cRichInterval rintron;

	int type = _INTERVAL_INTRON_NF;

	for (int i=1; i<m_transcript.numel(); ++i)
	{
		intron = cInterval( m_transcript.m[i-1].second()+1, m_transcript.m[i].first()-1, type );
		rintron.m = intron.m; rintron.nType = intron.nType;
		rintron.m_lenExonL = m_transcript.m[i-1].length();
		rintron.m_lenExonR = m_transcript.m[i].length();

		vintrons.push_back( rintron );
	}

	getIntronTypes( vintrons );
}

void cIsoform::getIntronTypes(vector<cRichInterval>& vintrons) const
{	// WARNING: vintrons MUST be sorted and unique!
	cRichInterval intron;
	int type = _INTERVAL_INTRON_NF;
	int ix = 0;
	int ixf;

	if (vintrons.empty()) return;

	for (int i=1; i<m_coding.numel(); ++i)
	{
		if (ix>=(int)vintrons.size()) break;

		intron = cRichInterval( cInterval( m_coding.m[i-1].second()+1, m_coding.m[i].first()-1, _INTERVAL_UNDEFINED ) );
		if (vintrons[ix] < intron) {
			vintrons[ix].nType = _INTERVAL_INTRON_NF;
			++ix;
			--i;
			continue;
		}
		else if (vintrons[ix] == intron)
		{
			ixf = m_chStrand=='+' ? i : i-1;	// intron phase takes on phase of downstream exon
			switch (m_coding.m[ixf].nType)
			{
			case _INTERVAL_EXON_0:
				type = _INTERVAL_INTRON_0; break;
			case _INTERVAL_EXON_1:
				type = _INTERVAL_INTRON_1; break;
			case _INTERVAL_EXON_2:
				type = _INTERVAL_INTRON_2; break;
			default:
				type = _INTERVAL_INTRON_NF; break;
			};
			vintrons[ix].nType = type;
		}
		++ix;
	}
}

void cIsoform::getIntronTypes(vector<cInterval>& vintrons) const
{	// WARNING: vintrons MUST be sorted and unique!
	cInterval intron;
	int type = _INTERVAL_INTRON_NF;
	int ix = 0;
	int ixf;

	if (vintrons.empty()) return;

	for (int i=1; i<m_coding.numel(); ++i)
	{
		if (ix>=(int)vintrons.size()) break;

		intron = cInterval( m_coding.m[i-1].second()+1, m_coding.m[i].first()-1, _INTERVAL_UNDEFINED );
		if (vintrons[ix] < intron) {
			vintrons[ix].nType = _INTERVAL_INTRON_NF;
			++ix;
			--i;
			continue;
		}
		else if (vintrons[ix] == intron)
		{
			ixf = m_chStrand=='+' ? i : i-1;	// intron phase takes on phase of downstream exon
			switch (m_coding.m[ixf].nType)
			{
			case _INTERVAL_EXON_0:
				type = _INTERVAL_INTRON_0; break;
			case _INTERVAL_EXON_1:
				type = _INTERVAL_INTRON_1; break;
			case _INTERVAL_EXON_2:
				type = _INTERVAL_INTRON_2; break;
			default:
				type = _INTERVAL_INTRON_NF; break;
			};
			vintrons[ix].nType = type;
		}
		++ix;
	}
}

void cIsoform::writeGff( std::ostream& os, string& seqName )
{
	// mRNA line
	os << seqName << "\t"
		<< PROG_NAME << "\t"
		<< "mRNA" << "\t"
		<< m_ibeg << "\t"
		<< m_iend << "\t"
		<< "." << "\t"
		<< m_chStrand << "\t"
		<< "." << "\t"
		<< "ID=" << m_strId << ";"
		<< "Parent=" << m_strParent << ";" << std::endl;

	m_transcript.writeGff( os, seqName, m_chStrand, m_strId );
	m_coding.writeGff( os, seqName, m_chStrand, m_strId );
}


