
#include "locus.h"
#include "targetver.h"
#include <fstream>
#include <iostream>
#include <algorithm>
#include <limits>

int iglobalcount;

bool oeLessThan(originExon ri, originExon rj)
{
	return ri.istart < rj.istart;
}

bool oiLessThan(originIntron ri, originIntron rj)
{
	//if (ri.introns.empty()) return false;
	//if (rj.introns.empty()) return true;

	if (ri.introns[0].ibeg == rj.introns[0].ibeg)
		return ri.introns[0].iend < rj.introns[0].iend;
	else
		return ri.introns[0].ibeg < rj.introns[0].ibeg;
}

bool intronSegmentLessThan(segment ri, segment rj)
{
	if (ri.ibeg == rj.ibeg)
		return ri.iend < rj.iend;
	else
		return ri.ibeg < rj.ibeg;
}

bool intronSegmentEqualTo(segment ri, segment rj)
{
	return (ri.ibeg == rj.ibeg && ri.iend == rj.iend);
}


bool boundaryLessThan(boundary bi, boundary bj)
{
	return bi.ix < bj.ix;
}

extern bool intronLessThan2 (sIntron i1, sIntron i2);
extern bool intronEqualTo2(sIntron i1, sIntron i2);


void cLoci::deleteOe()
{
	int i;

	for (i=0; i<(int)m_loci.size(); ++i)
	{
		m_loci[i].oe.clear();
	}
}

cLoci::cLoci()
{
	iglobalcount = 0;
	m_ixLast = 0;

	m_nOffset1 = m_nOffset2 = 0;
}

void cLoci::reset()
{
	m_loci.clear();

	m_ixLast = 0;

	m_lr.oe.clear();
	m_lr.oi.clear();
	m_lr.introns.clear();
	m_lr.segments.clear();
	m_lr.isoforms.clear();

	m_len = 0;
}

int cLoci::ctIntrons()
{
	int i, ni, ni_tot(0);

	for (i=0; i<(int)m_loci.size(); ++i)
	{
		ni = m_loci[i].introns.size();
		ni_tot += ni;
	}

	return ni_tot;
}

bool cLoci::addOrigins( vector<alignRec>& aligns )
{
	locusRec rec;
	originExon oe;
	originIntron oi;
	segment intr_seg;
	int ix, i, j;

	for (i=0; i<(int)aligns.size(); ++i)
	{
		ix = findLocus( aligns[i].istart, aligns[i].rname );
		if (ix == -1)
			continue;

		if (isExonic(aligns[i]))
		{
			oe.istart = aligns[i].istart;
			oe.length = aligns[i].nLength;
			oe.score = aligns[i].score;
			// add exonic origin
			m_loci[ix].oe.push_back( oe );
		}
		else
		{
			oi.strId = aligns[i].qname;
			oi.pieces.clear();
			for (j=0; j<(int)aligns[i].pieces.size(); ++j) {
				oe.istart = aligns[i].istart + aligns[i].pieces[j].start;
				oe.length = aligns[i].pieces[j].length;
				oe.score = 0.;
				oi.pieces.push_back( oe );

				// intron segment
				if (j % 2 != 0) {		// j is odd
					intr_seg.ibeg = oi.pieces[j-1].istart + oi.pieces[j-1].length;
					intr_seg.iend = oi.pieces[j].istart - 1;
					m_loci[ix].introns.push_back( intr_seg );
				}
			}
			// add intronic origin
			//m_loci[ix].oi.push_back( oi );
		}
	}

	return false;
}

int cLoci::findLocus(int ntCoord, string& name)
{
	int i;

	for (i=m_ixLast; i<(int)m_loci.size(); ++i)
	{
		if (m_loci[i].name.compare(name) != 0)
			continue;

		if (ntCoord>=m_loci[i].ibeg && ntCoord<=m_loci[i].iend) {
			m_ixLast = i;
			return i;
		}
	}

	// wrap around
	for (i=0; i<m_ixLast; ++i)
	{
		if (m_loci[i].name.compare(name) != 0)
			continue;

		if (ntCoord>=m_loci[i].ibeg && ntCoord<=m_loci[i].iend) {
			m_ixLast = i;
			return i;
		}
	}

	// not found!
	return -1;
}

bool cLoci::isExonic(alignRec& aln)
{
	return !aln.bGap;
}


bool cLoci::checkInterior( int ilocus, int ipt )
{
	// return true if ipt lies interior to an intron at ilocus
	// NOTE: introns MUST be sorted!
	int i;

	for (i=0; i<(int)m_loci[ilocus].introns.size(); ++i)
	{
		if (m_loci[ilocus].introns[i].iend < ipt)
			continue;

		if (m_loci[ilocus].introns[i].ibeg > ipt)
			return false;

		if (m_loci[ilocus].introns[i].ibeg <= ipt &&
			m_loci[ilocus].introns[i].iend >= ipt)
			return true;
	}

	return false;
}

bool cLoci::checkOverlap( int ilocus, int i1, int i2 )
{
	if (m_loci[ilocus].introns[i1].iend < m_loci[ilocus].introns[i2].ibeg ||
		m_loci[ilocus].introns[i1].ibeg > m_loci[ilocus].introns[i2].iend)
		return true;

	return false;
}


bool cLoci::isSpliced( splicePath& sp, int ilocus, int testPt )
{
	// is testPt inside any of the introns in splicePath sp? if so, return true.
	//  otherwise, return false.

	int i;
	segment intron;

	for (i=0; i<(int)sp.ixIntron.size(); ++i)
	{
		intron = m_loci[ilocus].introns[ sp.ixIntron[i] ];
		if (testPt >= intron.ibeg && testPt <= intron.iend)
			return true;
	}

	return false;
}

/*
void cLoci::cleanIntrons()
{
	// remove all redundant introns
	// assumes introns are already sorted
	int i, j, begLast, endLast;

	for (i=0; i<m_loci.size(); ++i)
	{
		if (m_loci[i].introns.empty())
			continue;

		begLast = m_loci[i].introns[0].ibeg;
		endLast = m_loci[i].introns[0].iend;

		for (j=1; j<m_loci[i].introns.size(); ++j)
		{
			if (m_loci[i].introns[j].ibeg == begLast &&
				m_loci[i].introns[j].iend == endLast)
			{	// duplicate copy, remove
				m_loci[i].introns.erase( m_loci[i].introns.begin() + j );
				--j;
			}
			else
			{	// update beg,end
				begLast = m_loci[i].introns[j].ibeg;
				endLast = m_loci[i].introns[j].iend;
			}
		}	// loop over introns
	}	// loop over loci
}
*/

void cLoci::cleanIntrons()
{
	// return all "unique" introns

	vector<segment>::iterator it;
	int nIntrons(0), nBefore(0);
	int ix, ct;

	for (int i=0; i<(int)m_loci.size(); ++i)
	{
		if (m_loci[i].introns.empty())
			continue;

		nBefore += m_loci[i].introns.size();

		sort( m_loci[i].introns.begin(), m_loci[i].introns.end(), intronSegmentLessThan );

		// determine count of alignments to every intron
		ix = 0;
		ct = 1;
		for (int j=1; j<(int)m_loci[i].introns.size(); ++j)
		{
			if (m_loci[i].introns[j].ibeg==m_loci[i].introns[j-1].ibeg &&
				m_loci[i].introns[j].iend==m_loci[i].introns[j-1].iend)
			{
				++ct;
			}
			else {
				// set ct for each alike intron
				for (int k=ix; k<j; ++k) { m_loci[i].introns[k].ct = ct; }
				ct = 1;
				ix = j;
			}
		}

		it = unique( m_loci[i].introns.begin(), m_loci[i].introns.end(), intronSegmentEqualTo );
		m_loci[i].introns.resize( it - m_loci[i].introns.begin() );

		nIntrons += m_loci[i].introns.size();
	}	// loop over loci

	//printf("cLoci::cleanIntrons--%d introns before cleaning\n", nBefore);
	//printf("cLoci::cleanIntrons--%d introns after cleaning\n", nIntrons);
}

void cLoci::sortOe()
{
	int i;

	for (i=0; i<(int)m_loci.size(); ++i)
	{
		sort( m_loci[i].oe.begin(), m_loci[i].oe.end(), oeLessThan );
	}
}

void cLoci::sortIntrons()
{
	int i;

	for (i=0; i<(int)m_loci.size(); ++i)
	{
		//sort( m_loci[i].introns.begin(), m_loci[i].introns.end(), intronSegmentLessThan );
	}
}

bool cLoci::load(string& fname)
{
	std::ifstream data_file( fname.c_str() );

	if ( !data_file.is_open() )
	{
		std::cout << PROG_NAME << " : error, opening file " << fname << "\n"; 
		return true;
	}

	string line;

	while( !data_file.eof() )
	{
		getline( data_file, line );
		if ( parseOrgLine( line ) )
			std::cout << PROG_NAME << " : warning, file " << fname << " line ignored : " << line << "\n";
	}

	data_file.close();

	// add the final locus to m_loci
	m_lr.seqLength = m_len;
	m_loci.push_back( m_lr );		// save new record

	return false;
}

bool cLoci::parseOrgLine( string& line )
{
	alignRec rec;
	//gffLine line_rec;
	std::basic_string <char>::size_type L, R;
	L = line.find_first_not_of( " \t" );
	// skip white space and comments
	if (( L == -1 )||( line[L] == '#' ))
		return false;
	R = line.find_first_of( " \t", L );

	string s = line.substr( L, R - L );

	originExon oe;
	originIntron oi;
	
	if (s.compare("locus")==0) {		// parse locus line
		if (!m_lr.oe.empty() || !m_lr.oi.empty()) {
			m_lr.seqLength = m_len;
			m_loci.push_back( m_lr );		// save new record
			//m_lr.oe.clear();
			//m_lr.oi.clear();
		}

		parseLocusLine( m_lr, line );
	}
	else if (s.compare("oe")==0) {			// parse exon origin line
		if (!parseOeLine( oe, line ))
			m_lr.oe.push_back( oe );
	}
	else {								// parse intron origin line ("oi")
		if (!parseOiLine( oi, line )) {
			m_lr.oi.push_back( oi );
			m_lr.introns.insert( m_lr.introns.end(), oi.introns.begin(), oi.introns.end() );
		}
	}

	return false;
};

bool cLoci::parseLocusLine( locusRec& lr, string& line )
{
	// clear input
	lr.oe.clear();
	lr.oi.clear();
	lr.introns.clear();
	lr.isoforms.clear();

	m_len = 0;

	// locus	Chr1	5914	10193	413	85
	//
	std::basic_string <char>::size_type L, R;
	L = line.find_first_not_of( " \t" );
	
	// skip white space and comments
	if (( L == -1 )||( line[L] == '#' ))
		return false;

	vector< string > cols;
	int nCols(6);
	cols.reserve(nCols);

	R = line.find_first_of( " \t", L );

	lr.isoforms.clear();
	lr.oe.clear();
	lr.oi.clear();

	for( int i = 0 ; (i < nCols-1)&&(L != -1); ++i )
	{
		cols.push_back( line.substr( L, R - L ) );

		L = line.find_first_not_of( " \t", R );
		R = line.find_first_of( " \t", L );
	}

	// all remaining text
	cols.push_back( line.substr( L, string::npos ) );

	if ( cols.size() != nCols )
	{
		std::cout << "Data format error: " << line << std::endl;
		return true;
	}

	lr.name = cols[1];
	lr.ibeg = atoi( cols[2].c_str() );
	lr.iend = atoi( cols[3].c_str() );

	return false;
}

bool cLoci::parseOeLine( originExon& oe, string& line )
{
	// oe	5837	50	-801.03
	//
	std::basic_string <char>::size_type L, R;
	L = line.find_first_not_of( " \t" );
	
	// skip white space and comments
	if (( L == -1 )||( line[L] == '#' ))
		return false;

	vector< string > cols;
	int nCols(4);
	cols.reserve(nCols);

	R = line.find_first_of( " \t", L );

	for( int i = 0 ; (i < nCols-1)&&(L != -1); ++i )
	{
		cols.push_back( line.substr( L, R - L ) );

		L = line.find_first_not_of( " \t", R );
		R = line.find_first_of( " \t", L );
	}

	// all remaining text
	cols.push_back( line.substr( L, string::npos ) );

	if ( cols.size() != nCols )
	{
		std::cout << "Data format error: " << line << std::endl;
		return true;
	}

	oe.istart = atoi( cols[1].c_str() );
	oe.length = atoi( cols[2].c_str() );
	oe.score = atof( cols[3].c_str() );

	// running length total
	m_len += oe.length;

	return false;
}

bool cLoci::parseOiLine( originIntron& oi, string& line )
{
	// oi	2	7029	41	 0.00	7157	9	 0.00
	//

	// clear input
	oi.pieces.clear();
	oi.introns.clear();

	// parse line
	std::basic_string <char>::size_type L, R;
	L = line.find_first_not_of( " \t" );
	
	// skip white space and comments
	if (( L == -1 )||( line[L] == '#' ))
		return false;

	vector< string > cols, segs;
	int nCols(2), i;
	cols.reserve(nCols);

	R = line.find_first_of( " \t", L );

	for( i = 0 ; (i < 2)&&(L != -1); ++i )
	{
		cols.push_back( line.substr( L, R - L ) );

		L = line.find_first_not_of( " \t", R );
		R = line.find_first_of( " \t", L );
	}

	if ( cols.size() != 2 )
	{
		std::cout << "Data format error: " << line << std::endl;
		return true;
	}

	// id
	int nId = m_lr.oi.size() + 1;
	char pch[32];
	sprintf( pch, "%d", nId );
	oi.strId = "oi_";
	oi.strId.append( pch );

	int nPieces = atoi( cols[1].c_str() );		// no. of pieces of read origin
	originExon oi_seg;

	for( i = 0 ; (i < 3*nPieces-1)&&(L != -1); ++i )
	{
		segs.push_back( line.substr( L, R - L ) );

		L = line.find_first_not_of( " \t", R );
		R = line.find_first_of( " \t", L );
	}

	// all remaining text
	segs.push_back( line.substr( L, string::npos ) );

	if ( segs.size() != 3*nPieces )
	{
		std::cout << "Data format error: " << line << std::endl;
		return true;
	}

	segment sint;
	sint.extr_L = sint.extr_R = 1;
	sint.level = 0;
	sint.nOrigins = 0;

	for (i=0; i<nPieces; ++i)
	{
		// ith piece
		oi_seg.istart = atoi( segs[3*i].c_str() );
		oi_seg.length = atoi( segs[3*i+1].c_str() );
		oi_seg.score = atof( segs[3*i+2].c_str() );

		// add intron
		if (i>0)
		{
			sint.ibeg = oi.pieces.back().istart + oi.pieces.back().length;
			sint.iend = oi_seg.istart - 1;

			sint.ibeg += 1;
			//sint.iend -= 1;

			sint.ibeg += m_nOffset1;
			sint.iend += m_nOffset2;
			//sint.iend += m_nYangOffset;

			oi.introns.push_back( sint );
		}

		// add exonic piece to oi
		oi.pieces.push_back( oi_seg );

		// running length total
		m_len += oi_seg.length;
	}

	return false;
}

//void cLoci::countPaths(string fname)
//{
//	int i, j, flag;
//	vector<splicePath> paths;
//	vector<pathNode> graph;
//	island isle;
//
//	for (i=0; i<m_loci.size(); ++i)
//	{
//		isle.ibeg = m_loci[i].ibeg;
//		isle.iend = m_loci[i].iend;
//		isle.ixIntron.clear();
//
//		if (m_loci[i].introns.size()>12)
//		{
//			m_loci[i].nPaths = 9999;
//		}
//		else
//		{
//			for (j=0; j<m_loci[i].introns.size(); ++j)
//				isle.ixIntron.push_back(j);
//
//			getPaths( paths, graph, isle, i );
//			m_loci[i].nPaths = paths.size();
//		}
//	}
//
//	FILE *pf = fopen(fname.c_str(), "w");
//	if (pf==NULL) {
//		printf("Error opening file for writing: %s\n", fname.c_str());
//		return;
//	}
//
//	for (i=0; i<m_loci.size(); ++i)
//	{
//		fprintf( pf, "%d\t%d\t%d\n", i, m_loci[i].introns.size(), m_loci[i].nPaths );
//	}
//
//	fclose( pf );
//}

void cLoci::writeOrigins(string fname)
{
	FILE *pf = fopen(fname.c_str(), "w");
	if (pf==NULL) {
		printf("Error opening file for writing: %s\n", fname.c_str());
		return;
	}

	int i, j;

	for (i=0; i<(int)m_loci.size(); ++i)
	{
		// locus description line
		fprintf(pf, "locus\t%s\t%d\t%d\t%d\t%d\n", m_loci[i].name.c_str(), m_loci[i].ibeg, m_loci[i].iend, (int)m_loci[i].oe.size(), (int)m_loci[i].oi.size());
		// exonic origins
		for (j=0; j<(int)m_loci[i].oe.size(); ++j)
		{
			fprintf(pf, "oe\t%d\t%d\t%5.2f\n", m_loci[i].oe[j].istart, m_loci[i].oe[j].length, m_loci[i].oe[j].score);
		}
		//// intronic origins
		//for (j=0; j<m_loci[i].oi.size(); ++j)
		//{
		//	fprintf(pf, "oi\t%d", m_loci[i].oi[j].pieces.size());
		//	for (k=0; k<m_loci[i].oi[j].pieces.size(); ++k)
		//		fprintf(pf, "\t%d\t%d\t%5.2f", m_loci[i].oi[j].pieces[k].istart, m_loci[i].oi[j].pieces[k].length, m_loci[i].oi[j].pieces[k].score);
		//	fprintf(pf, "\n");
		//}
		// intronic origins
		for (j=0; j<(int)m_loci[i].introns.size(); ++j)
		{
			fprintf(pf, "oi\t2");
			fprintf(pf, "\t%d\t1\t%d\t1\t0.0", m_loci[i].introns[j].ibeg-1, m_loci[i].introns[j].iend+1);
			fprintf(pf, "\n");
		}
	}

	fclose( pf );
}

void cLoci::writeIntronOrigins(string fname)
{
	FILE *pf = fopen(fname.c_str(), "w");
	if (pf==NULL) {
		printf("Error opening file for writing: %s\n", fname.c_str());
		return;
	}

	int i, j;

	for (i=0; i<(int)m_loci.size(); ++i)
	{
		// locus description line
		fprintf(pf, "locus\t%s\t%d\t%d\t%d\t%d\n", m_loci[i].name.c_str(), m_loci[i].ibeg, m_loci[i].iend, 0, (int)m_loci[i].introns.size());
		// intronic origins
		for (j=0; j<(int)m_loci[i].introns.size(); ++j)
		{
			fprintf(pf, "oi\t2\t%d\t1\t0.0\t%d\t1\t0.0", m_loci[i].introns[j].ibeg-1, m_loci[i].introns[j].iend+1);
			fprintf(pf, "\n");
		}
	}

	fclose( pf );
}

vector<sIntron> cLoci::getIntrons( int nContig, string& seqName )
{
	sIntron intron;
	vector<sIntron> introns;
	segment seg;
	int i, j;

	intron.nContig = nContig;
	intron.nIntronsInGene = -1;
	intron.ixIntron = -1;
	intron.extr_L = 0;
	intron.extr_R = 0;
	intron.extr_I = 0;
	intron.iStrand = 0;
	intron.iframe = -2;
	intron.nCoverage = 0;
	intron.posAtg = std::numeric_limits<int>::max();
	intron.posStp = std::numeric_limits<int>::max();
	intron.nFpL = intron.nFpR = -99;

	for (i=0; i<(int)m_loci.size(); ++i)
	{
		if (m_loci[i].name.compare(seqName) == 0)
		{
			// get introns associated with each intron origin
			for (j=0; j<(int)m_loci[i].introns.size(); ++j)
			{
				seg = m_loci[i].introns[j];
				intron.ibeg = seg.ibeg;
				intron.iend = seg.iend;
				intron.nCoverage = seg.ct;

				intron.strId = seqName;
				//intron.flankLeft = oi.pieces[0].length;
				//intron.flankRight = oi.pieces[1].length;

				introns.push_back( intron );
			}
		}
	}

	std::sort( introns.begin(), introns.end(), intronLessThan2 );
	vector<sIntron>::iterator it = std::unique( introns.begin(), introns.end(), intronEqualTo2 );
	introns.resize( it - introns.begin() );

	return introns;
}
