
#pragma once

#include "geneType.h"
#include "origin.h"
#include <vector>

using std::vector;

enum { BOUND_LEFT = 0, BOUND_RIGHT, BOUND_LEFT_RIGHT };

typedef vector<segment> VSEGMENT;
typedef vector<int> VINT;

typedef struct tagBoundary
{
	int ix;
	int side;
	int count;
} boundary;

typedef struct tagOriginExon
{
	int istart;
	int length;
	double score;
} originExon;

typedef struct tagOriginIntron
{
	string strId;
	vector<originExon> pieces;
	vector<segment> introns;
} originIntron;

typedef struct tagGeneIsoform
{
	bool bStrand;
	int cdsStart;
	int cdsStop;
	vector<segment> exons;
} geneIsoform;

typedef struct tagLocusRec
{
	int id;
	string name;
	int ibeg, iend;
	int seqLength;			// total length of read origin alignments
	int nPaths;

	vector<originExon> oe;
	vector<originIntron> oi;
	vector<segment> introns;
	vector<segment> segments;

	vector<geneIsoform> isoforms;
} locusRec;

typedef struct tagSplicePath
{
	int id;
	vector<int> ixIntron;
} splicePath;

typedef struct tagPathNode
{
	int ix;				// which intron
	vector<int> vix;	// which downstream introns we point to
} pathNode;

typedef struct tagIsland
{
	int id;
	vector<int> ixIntron;
	int ibeg;
	int iend;
} island;


class cLoci
{
public:
	cLoci();
	~cLoci() { };

	void reset(void);

	int ctIntrons(void);

	void deleteOe(void);

	void addLocus( locusRec& rec ) { m_loci.push_back( rec ); }
	bool addOrigins( vector<alignRec>& aligns );

	int findLocus(int coord, string& name);
	bool isExonic(alignRec& aln);
	bool isSpliced( splicePath& sp, int ilocus, int testPt );

	bool checkInterior( int ilocus, int ipt );
	bool checkOverlap( int ilocus, int i1, int i2 );
	void cleanIntrons();
	void sortOe(void);
	void sortIntrons(void);

	bool load(string& fname);
	bool parseOrgLine( string& line );
	bool parseLocusLine( locusRec& lr, string& line );
	bool parseOeLine( originExon& oe, string& line );
	bool parseOiLine( originIntron& oi, string& line );

	//void countPaths(string fname);

	void writeOrigins(string fname);
	void writeIntronOrigins(string fname);

	vector<sIntron> getIntrons( int nContig, string& seqName );

	int m_nOffset1, m_nOffset2;

private:
	vector<locusRec> m_loci;

	int m_ixLast;
	locusRec	m_lr;
	int			m_len;
};
