
#pragma once

#include "geneType.h"
#include "origin.h"
#include <string>
#include <vector>
#include <map>

using std::string;
using std::vector;
using std::map;

#define _ALN_PROPERTY_BENIGN				0x00
#define _ALN_PROPERTY_STRAND_CONFLICT		0x01
#define _ALN_PROPERTY_FRAME_SHIFT			0x02
#define _ALN_PROPERTY_LOW_SPLICE_SCORE		0x04
#define _ALN_PROPERTY_MED_SPLICE_SCORE		0x08
#define _ALN_PROPERTY_HIGH_SPLICE_SCORE		0x10
#define _ALN_PROPERTY_NONCANONICAL			0x20
#define _ALN_PROPERTY_GENIC					0x40

//
//typedef struct tagPiece
//{
//	int start;
//	int length;
//} piece;
//
//typedef struct tagAlignRec
//{
//	string qname;			// query seq name
//	string rname;			// ref seq name
//	int flag;				// general info, including strand info
//	int istart;				// alignment start coord. in ref seq
//	int mapqual;			// quality scalar
//	string cigar;			// CIGAR string
//	vector<piece> pieces;	// coord. and length of each piece of alignment
//	string seq;				// query seq
//	string qual;			// quality seq (alignment scores)
//	double score;			// score of whole alignment
//	bool bStrand;
//	int nLength;			// alignment length
//	bool bGap;				// is this a gapped alignment (pieces > 1 could just be insertion!)
//} alignRec;

class potJunc
{
	public:
		string SAM;
		map<int,vector<vector<int> > > SAM_info;			// key is part of segment
		map<int,vector<vector<double> > > score_info;
};

class cMulti
{
public:
	cMulti() {
		m_cigarChars = "MIDNSHP=X";
		m_nOffset1 = m_nOffset2 = 0;
	};
	~cMulti() { };

	void loadOld(string& fname);
	bool load(string& fname);
	bool loadSam(string& fname);
	void save(string& fname, int nFilter);		// save all sets WITH nFilter property
	void saveN(string& fname, int nFilter);		// save all sets WITHOUT nFilter property
	void saveMed(string fname, int nFilter);		// save all sets WITH nFilter property
	void saveGenic(string fname);
	void saveSam(string& fname, int nFilter);
	void saveGenicSam(string fname, int nFilter);
	void saveNonSam(string fname, int nFilter);
	void saveGenicSamN(string fname, int nFilter);
	//void saveGenicWeakN(string fname, int nFilterWeak, int nFilterStrong);

	bool loadScoreFile( string& fname );

	void add(vector<sIntronSet>& sets);

	void sortSets(void);
	vector<sIntronSet> getIntronSets(string& seqName, int iContig);
	vector<sIntron> getIntrons(string& seqName, int iConig);
	
	string getSeqName(string& sam_line);
	bool throwOut(int ix);
	//bool isGenic(cGene& gene, cGeneSet& set, int ix);

	void setGenic(int ix) {
		if (ix<0 || ix>=(int)m_sets.size()) return;
		m_sets[ix].bGenic = true;
	}

private:
	bool parseSamLine( string& line );
	bool parseScoreLine( string& line );
	vector<string> split(string& line, int maxCols);
	void procCigar(alignRec& rec);
	void setScore(alignRec& rec);

	void getSamLine(string& sam_line, int i);
	void getSamInfo(vector<int>& info, int i);

	string m_cigarChars;

public:
	vector<sIntronSet> m_sets;

	int m_nAlnSegments;
	int n_nSegmentLength;
	int m_nOffset1, m_nOffset2;

	vector< potJunc > candidate;
};

