
#pragma once

#include <string>
#include <vector>
#include <iostream>
#include <fstream>

using std::vector;
using std::string;

typedef struct tagPiece
{
	int start;
	int length;
} piece;

typedef struct tagAlignRec
{
	string qname;			// query seq name
	string rname;			// ref seq name
	int flag;				// general info, including strand info
	int istart;				// alignment start coord. in ref seq
	int mapqual;			// quality scalar
	string cigar;			// CIGAR string
	vector<piece> pieces;	// coord. and length of each piece of alignment
	string seq;				// query seq
	string qual;			// quality seq (alignment scores)
	double score;			// score of whole alignment
	bool bStrand;
	int nLength;			// alignment length
	bool bGap;				// is this a gapped alignment (pieces > 1 could just be insertion!)
} alignRec;


class cOrigin
{
public:
	cOrigin()
	{
		m_streamPos = 0;
		m_bDone = false;
		m_cigarChars = "MIDNSHP=X";
	};
	~cOrigin() { };

	bool loadSam(string& fname);
	bool loadSamBlock(string& fname, int nLines);
	bool parseDataLine( string& line );

	void procCigar(alignRec& rec);
	void setScore(alignRec& rec);

	void reset(void);
	void Sort(void);

	bool IsDone(void) { return m_bDone; }

public:
	vector<alignRec> m_aligns;
	std::streampos   m_streamPos;
	bool             m_bDone;

	string           m_cigarChars;
};
