/*
	This class implements performance metrics suitable for short read alignment programs.
	The class will handle calculations necessary to compute intron-level Sn/Sp.

	Paul D. Burns, 2011
*/


#pragma once

#include "geneType.h"
#include <vector>
#include <string>

using std::vector;
using std::string;

class cMetricsSR
{
public:
	cMetricsSR()
	{
		reset();
		m_cacheLen = 10;
	}
	~cMetricsSR() { }

	void reset(void);

	bool markAgreement( vector<sIntron>& ref_introns, vector<sIntron>& query_introns, int iType );
	bool checkOverlap( sIntron& intron1, sIntron& intron2 );
	bool markIntron( sIntron& intron, sIntron& query_intron );
	bool markIntronPh( sIntron& intron, sIntron& query_intron );

	bool getNN( vector<sIntron>& ref_introns, vector<sIntron>& query_introns );
	bool donNN( vector<sIntron>& ref_introns, vector<sIntron>& query_introns );
	bool accNN( vector<sIntron>& ref_introns, vector<sIntron>& query_introns );
	
	int nextDon( int curDon, int& lastDon, int pos, int& ix, vector<sIntron>& introns );
	int nextAcc( int curAcc, int& lastAcc, int pos, int& ix, vector<sIntron>& introns );
	
	bool resetDonDist( vector<sIntron>& query_introns );
	bool resetAccDist( vector<sIntron>& query_introns );

	void scoreIntron( sIntron& intron, int iType );

	void writeMetrics( string fname, int nRefIntrons, int nDonRef, int nAccRef, int nPredIntrons, int nDonQ, int nAccQ, int nUtr );

	bool findDon(sIntron& intron);
	bool findAcc(sIntron& intron);
	bool findIntron(sIntron& intron);

public:
	int m_cacheLen;
	vector<sIntron> m_intronCache;

	string m_chr;

	// metrics: ref, general feature counts
	int nDon, nAcc, nIntron;
	// metrics: predictions, general feature counts
	int nDonP, nAccP, nIntronP;

	// metrics: ref, these are counts of 'good' predictions
	int nDonOk, nAccOk, nIntronOk, nUtrIntronOk;
	int nRefIntrons;

	// metrics: query, these are counts of 'good' predictions
	int nDonPred, nAccPred, nIntronPred, nUtrPred;

	// metrics: other
	int nDonMiss, nAccMiss;
	int nDonFalse, nAccFalse;
	int nIntronMiss, nFalseIntron;
};
