#include <numeric>
#include <functional>
#include <iostream>
#include <string>
#include <vector>
#include <map>
#include <iomanip>
#include <fstream>
#include <cctype>
#include <set>
#include <iterator>
#include <algorithm>
#include <sstream>

using namespace std;

ofstream Out_full("new_full.sam");
ofstream Out("solid"), Out_multi("multi"), Out_NON("non_solid"), Out_NON_multi("non_multi");
double err_rate = 0.2; // if a segment has 2 errs, it should be longer than 10bp
map<string, int> ALL_READ;
int only_canonical = 0;
int N_core = 1;
string single_ref;

int lenMarkov = 80;

set<string> total_chr;
int min_intron = 20;
int Intron_length = 200000;
int N_fragment;
int search_flag = 0, search_flag_2 = 0;
int read_length;
struct interval{
     int start, end; // [start, end]
     interval(int  s, int  e) : start(s), end(e) {}
};

bool operator<(interval const& a, interval const& b){
     return a.start < b.start;
}


struct Part{
        int Ori, Pos, part, segMis, locat;
	int flag;
        string Chr, name;
        Part(int a, int b, int c, int d, string f, string e, int g = 0, int h = -1) : Ori(a),Pos(b),part(c),segMis(d),Chr(f),name(e),flag(g),locat(h) {}
};


struct SAM{
	string name, seq, qua, chr, cigar;
	int ori, b;
};


vector< Part > Temp;
map< int, vector<int> > PartHash;

typedef map<interval, set < int > > intervalMap;
typedef map<string, intervalMap> CHRintervalMap;

CHRintervalMap CandidateList;
vector< vector<Part> > CandidateVector, local_CandidateVector;
vector< SAM > Final;
struct job{
	string chr;
	int b,e;
	set < int > list;
};
vector< job > job_list;

interval const*find(intervalMap& s, int  point)
{
     intervalMap::iterator Itinterval = s.lower_bound(interval(point,point));
     if(Itinterval == s.end() || point < Itinterval->first.start) {
         if(Itinterval == s.begin())
             return NULL;
         --Itinterval;
         return point <= Itinterval->first.end && point >= Itinterval->first.start ? &(Itinterval->first) : NULL;
     }
     return point <= Itinterval->first.end && point >= Itinterval->first.start ? &(Itinterval->first) : NULL;

}

interval const*overlap(intervalMap& s, int  b, int e)
{
     intervalMap::iterator Itinterval = s.lower_bound(interval(e,e));
     if(Itinterval == s.end() || e < Itinterval->first.start) {
         if(Itinterval == s.begin())
             return NULL;
         --Itinterval;
         return e >= Itinterval->first.end && b <= Itinterval->first.start ? &(Itinterval->first) : NULL;
     }
     return e >= Itinterval->first.end && b <= Itinterval->first.start ? &(Itinterval->first) : NULL;

}


void merge_interval(intervalMap* map, int  b, int  e, int ID){
	set< int > merge_temp;
        if((*map).find(interval(b,e)) != (*map).end()){
                if((*map)[interval(b,e)].find(ID) !=(*map)[interval(b,e)].end())
                        return;
                else{
                        (*map)[interval(b,e)].insert(ID);
                        return;
                }
        }
        else{
		merge_temp.insert(ID);
                interval const*i = find(*map, b);
		
		if (i != NULL ){
			
			if( (*i).end >= e ){
				(*map)[*i].insert(ID);
				return;
			}
			else{
				b = (*i).start;
				for(set<int>::iterator m=(*map)[*i].begin(); m!= (*map)[*i].end(); m++){
					merge_temp.insert( *m );
				}
				(*map).erase((*i));
			}
		}
				
		interval const*j = find(*map, e);
		if (j != NULL ){
			e = (*j).end;
			for(set<int>::iterator m=(*map)[*j].begin(); m!= (*map)[*j].end(); m++){
                                        merge_temp.insert( *m );
                                }
                                (*map).erase((*j));
		}
		interval const*p = overlap(*map, b, e);
		while(p != NULL){
			for(set<int>::iterator m=(*map)[*p].begin(); m!= (*map)[*p].end(); m++){
                                        merge_temp.insert( *m );
                                }
                                (*map).erase((*p));
			p = overlap(*map, b, e);
		}
		
		(*map)[interval(b,e)] = merge_temp;
		
        }
}






map<string, vector<int> >::iterator ItRefIndex;
//string fa_whole;//string of ref.fa
string ref_name;

char CHR1[500];
int CHR1_begin, CHR1_end;
//char ref_ori_1, ref_ori_2;//orientation of reference gene

int totalMis;
int halfLength;
int mask;


map<string,string> InputFastq; //input reads
map<string, vector< Part > > InputFastqPart;//patial information    part(1,2) orien(1,2) position(int)
map<string, vector< Part > >::iterator ItFastqPart;
map<string,string> InputQ;//quality

//map<string, vector<vector<int> > > Gapped;//orien(1,2) start alignment1 gap alignment2
//map<string, vector<vector<int> > >::iterator ItGapped;

vector<int> PosiPair;//list of mapped region
set<int> Annotation1;//list of annotated exon from ucsc
set<int> Annotation2;
int tile_length;

pair<int,int> JunctionPair;
map<pair<int,int>, vector<string> > Junction;
map<pair<int,int>, vector<string> >::iterator ItJunction;

set<vector<int> > PCR;

set<vector<int> >::iterator ItPCR;

//map<string, vector<string> > Junction;
//map<string, vector<string> >::iterator ItJunction; 



int MINOVLP;

void usage() {
        cout<<
                "GappedAlignment - find gapped alignment and output CIGAR\n"
                "    ";
}

string revString(const std::string & Forward){
        string Reverse;
        for(int i=Forward.length()-1;i>=0;i--){
                if(Forward[i]=='A')
                        Reverse.push_back('T');
                else if(Forward[i]=='T')
                        Reverse.push_back('A');
                else if(Forward[i]=='G')
                        Reverse.push_back('C');
                else if(Forward[i]=='C')
                        Reverse.push_back('G');
		else 
			Reverse.push_back(Forward[i]);
        }
        return (Reverse);
}

string revQ(const std::string & A){
	string B;
	for(int i=A.length()-1;i>=0;i--)
		B.push_back(A[i]);
	return (B);
}

//index the merged fa file with length 6


void getRefIndex(int& b, int& e, map<string, set<int> >& RefIndex, string& fa_whole){
	
	int length = single_ref.length();
	if(b < 0)
		b = 0;
	if(e>length)
		e = length;
	fa_whole = single_ref.substr(b,e-b);
//	int length = fa_line.length();
//	if(mask != 1) transform(fa_whole.begin(),fa_whole.end(),fa_whole.begin(),::toupper);
	for(int i=0;i<int(fa_whole.length()) - tile_length;i++){
		RefIndex[ fa_whole.substr(i, tile_length) ].insert(i + b);
	}
}

int HammingDis(const std::string& s1, const std::string& s2, int mis){
    if (s1.size() != s2.size()){
      return 100;
    }
    return inner_product(s1.begin(), s1.end(), s2.begin(), 0, plus<int>(), not2(equal_to<string::value_type>()));
}			





int extendSegment1(int cano, int Part,string singleRead, int start, int segLength, int ori, int mismatch, int segMis, map<int, vector<vector<int> > >& Gapped, int b, int e, map<string, set<int> >& RefIndex,const std::string & fa_whole){
	int found = 0;
	int flag = 0;
	int Hamm1, Hamm2;
	int Len = singleRead.length();
	int spliceOri = 1;
	if(start < b || start + Len > e)
                return 0;
//	int tempHamm;
	if(ori == 2){
		singleRead=revString(singleRead);
	}
	
	int Len_ref = fa_whole.length();
	vector<int> information;
	for(int i=0; i<=Len-segLength-MINOVLP;i++){
		if(start+segLength+i+2-b > Len_ref || start+segLength+i - b < 0 ) continue;
		string canonical1 = fa_whole.substr(start+segLength+i - b, 2);
		if(cano)
			if(canonical1 != "GT" && canonical1 != "CT")
				continue;
		if(canonical1 == "CT")
			spliceOri = 0;
		else if(canonical1 == "GT")
                        spliceOri = 1;
		else
			spliceOri = 2;
		if(segLength-tile_length+i<0||segLength+i>Len){continue;}
		string s = singleRead.substr(segLength-tile_length+i,tile_length);
		int p_1 = start+segLength-tile_length+i;
		if(RefIndex.find(s) == RefIndex.end()){
			continue;
		}
		if(RefIndex[s].find(p_1) != RefIndex[s].end()){
//			if(*it == start+segLength-tile_length+i){
			
				if(start+segLength-tile_length+i -b > Len_ref){
					flag = 1;
					continue;
				}
				//cout <<"sssssssssssss\n";
				if(i <= tile_length)
					Hamm1 = segMis;
				else{
					if(start+segLength-b<0 || start+segLength-b +  i-tile_length > Len_ref || segLength + i-tile_length > Len){
						
						continue;
					}
					Hamm1 = HammingDis(singleRead.substr(0+segLength, i-tile_length), fa_whole.substr(start+segLength-b, i-tile_length),mismatch) + segMis;
				}
				//	Hamm1 += HammingDis(singleRead.substr(0+segLength+i-tile_length-1, 1), fa_whole.substr(start+segLength+i-tile_length-1, 1), mismatch);
				if(Hamm1 <= mismatch && i * err_rate > Hamm1){
					if(segLength+i+tile_length > Len){continue;}
					string s2 = singleRead.substr(segLength+i,tile_length);
					if(RefIndex.find(s2) == RefIndex.end()){
						continue;
					}
					set<int>::iterator it2 =  RefIndex[s2].lower_bound(p_1);
					while(it2 != RefIndex[s2].end()){
					//	cout << *it2 << endl;
						if(*it2 - p_1 < min_intron + tile_length){it2++; continue;}
							
						if(*it2 - p_1 > Intron_length){break;}
						if(*it2+tile_length+Len-(segLength+i+tile_length)-b  > Len_ref || *it2-2 - b  < 0 || *it2- b > Len_ref || segLength+i+tile_length >= Len){it2++; continue;}
						string canonical2 = fa_whole.substr (*it2-2 - b,2);
						if( (canonical1 == "GT"&& canonical2 == "AG") || (canonical1 == "CT"&& canonical2 == "AC") || cano == 0){
							Hamm2 = HammingDis(singleRead.substr(segLength+i+tile_length,Len-(segLength+i+tile_length)), fa_whole.substr(*it2+tile_length-b, Len-(segLength+i+tile_length)),mismatch-Hamm1);
							
							
						if(Hamm2 <= mismatch-Hamm1 && (Len-segLength-i) * err_rate > Hamm2){
							
							information.clear();
							information.push_back(Part);
							information.push_back(start + segLength+1);// care here, historical reasons
							information.push_back(i);//al1
							information.push_back(*it2 - p_1 -tile_length);//gap
							information.push_back(Len-segLength-i);//al2
							information.push_back(Hamm1);
							information.push_back(Hamm2);
							if(cano == 0)
								if((canonical1 == "GC"&& canonical2 == "AG") || (canonical1 == "CT"&& canonical2 == "GC") || (canonical1 == "AT"&& canonical2 == "AC") || (canonical1 == "GT"&& canonical2 == "AT")){
                                                                        if((canonical1 == "GC"&& canonical2 == "AG") || (canonical1 == "AT"&& canonical2 == "AC") )
                                                                                spliceOri = 1;//semi canonical

                                                                        else
                                                                                spliceOri = 0;

                                                                        cano = 2;//semi canonical
                                                                        vector< vector<int> >::iterator it2;
                                                                        for(it2 = Gapped[Part].begin(); it2!=Gapped[Part].end();){
                                                                                if((*it2)[7] == 0){
                                                                                        Gapped[Part].erase(it2);
                                                                                }
                                                                                else it2++;
                                                                        }
                                                        }
								
							information.push_back(cano);
							information.push_back(spliceOri);
							Gapped[Part].push_back(information);
							if(Gapped[Part].size() > 100){
								return 0;
							}
							found = 1;

						}
						}
						it2++;
					}

				}
				else flag = 1;
				//				break;
				//			}
		}
		//		if(flag == 1) break;
	}

	return found;
}
				
int extendSegment2(int cano, int Part, string singleRead, int start, int segLength, int ori, int mismatch, int segMis, map<int, vector<vector<int> > >& Gapped, int b, int e, map<string, set<int> >& RefIndex, const std::string & fa_whole){
	int found = 0;
	int Hamm1, Hamm2;
        int flag = 0;
	int spliceOri = 1;
	int Len = singleRead.length();
	if(Len < segLength)
		return 0;
	if(start < b || start + Len > e)
                return 0;
	if(ori == 2) singleRead=revString(singleRead);
	int Len_ref = fa_whole.length();
        vector<int> information;
        for(int i=0; i<=segLength-MINOVLP;i++){
		if(start-i-2 -b < 0 || start-i - b > Len_ref) continue;
		string canonical1 = fa_whole.substr(start-i-2 - b, 2);
		if(cano)
        	        if(canonical1 != "AG" && canonical1 != "AC")
                	        continue;
		if(canonical1 == "AC")
                        spliceOri = 0;
		else if(canonical1 == "AG")
                        spliceOri = 1;
		else
			spliceOri = 2;
		if(Len-segLength-i < 0 || Len-segLength-i+tile_length > Len){continue;}
                string s = singleRead.substr(Len-segLength-i,tile_length);
		int p_1 = start-i;
		if(RefIndex.find(s) == RefIndex.end()){
			continue;
		}
//		cout << singleRead << endl;
//                vector<int>::iterator it = RefIndex[s].begin();
                if(RefIndex[s].find(p_1) != RefIndex[s].end()){
                        //cout << p_1 <<"\t"<<i<< endl;
				
				if(start-i+tile_length - b<0){
                                        flag = 1;
                                        continue;
                                }
				if(i <= tile_length)
                                        Hamm1 = segMis;
                                else{
					if(start-i+tile_length-b <0 || start - b > Len_ref || Len-segLength-i+tile_length <0){
							continue;
							
					}
                                       Hamm1 = HammingDis(singleRead.substr(Len-segLength-i+tile_length, i-tile_length), fa_whole.substr(start-i+tile_length-b, i-tile_length), mismatch) + segMis;
				}
                                if(Hamm1 <= mismatch && i * err_rate > Hamm1){
					if(Len-segLength-i-tile_length < 0){continue;}
                                        string s2 = singleRead.substr(Len-segLength-i-tile_length,tile_length);
					if(RefIndex.find(s2) == RefIndex.end()){
						continue;
					}
                                        //vector<int>::iterator it2 =  RefIndex[s2].begin();
					set<int>::const_iterator it2 = RefIndex[s2].upper_bound(p_1);
					if(it2 == RefIndex[s2].begin())
						break;
					it2--;
					
                                        while(it2 != RefIndex[s2].begin()){
					//	cout << *it2 << endl;
						if(p_1 - *it2 < min_intron + tile_length){it2--; continue;}
                                                if( p_1 - *it2 > Intron_length){break;}
						if(*it2-(Len-segLength-i-tile_length)-b<0 || *it2-b > Len_ref || *it2+tile_length - b < 0 || *it2+tile_length - b +2 > Len_ref || Len-segLength-i-tile_length < 0){
							if(it2 == RefIndex[s2].begin())
                                                        	break;
							it2--; continue;
						}
						string canonical2 = fa_whole.substr (*it2+tile_length - b,2);
						
                                                if( (canonical1 == "AG"&& canonical2 == "GT") || (canonical1 == "AC"&& canonical2 == "CT" ) || cano == 0){
//						cout << *it2-(Len-segLength-i-tile_length)-b <<"\t"<<Len-segLength-i-tile_length<<endl;
						if(Len-segLength-i-tile_length <= 0){
							if(it2 == RefIndex[s2].begin())
                                                        	break;
							--it2;
							continue;
						}
                                                Hamm2 = HammingDis(singleRead.substr(0,Len-segLength-i-tile_length), fa_whole.substr(*it2-(Len-segLength-i-tile_length)-b, Len-segLength-i-tile_length), mismatch-Hamm1); 
						
                                                if(Hamm2 <= mismatch-Hamm1 && (Len-segLength-i) * err_rate > Hamm2){
			
				//cout << "sd\n";
                                                        information.clear();
	
                                                        information.push_back(Part);
                                                        information.push_back(*it2-(Len-segLength-i-tile_length)+1);
                                                        information.push_back(Len-segLength-i);//al1
                                                        information.push_back(p_1 - *it2 -tile_length);//gap
                                                        information.push_back(i);//al2
                                                        information.push_back(Hamm2);
                                                        information.push_back(Hamm1);
							if(cano == 0)
                                                                if((canonical1 == "GC"&& canonical2 == "AG") || (canonical1 == "CT"&& canonical2 == "GC") || (canonical1 == "AT"&& canonical2 == "AC") || (canonical1 == "GT"&& canonical2 == "AT")){
                                                                        if((canonical1 == "GC"&& canonical2 == "AG") || (canonical1 == "AT"&& canonical2 == "AC") )
                                                                                spliceOri = 1;//semi canonical

                                                                        else
                                                                                spliceOri = 0;

                                                                        cano = 2;//semi canonical
                                                                        vector< vector<int> >::iterator it2;
                                                                        for(it2 = Gapped[Part].begin(); it2!=Gapped[Part].end();){
                                                                                if((*it2)[7] == 0){
                                                                                        Gapped[Part].erase(it2);
                                                                                }
                                                                                else it2++;
                                                                        }
                                                        }
                                                                        

							information.push_back(cano);

							information.push_back(spliceOri);
                                                        Gapped[Part].push_back(information);
							if(Gapped[Part].size() > 100){                                                                                                  
								return 0;
							}
							found = 1;
                                                     
                                                }
						}
						if(it2 == RefIndex[s2].begin())
							break;
                                                it2--;
                                        }
                                }
                       		else flag = 1;
//				break;
 //                      } 
                        
                }
//		if(flag == 1) break;
        }
	
	return found;
	
}
			
int filGap(int cano, int Part, string singleRead, int start, int end, int segLength, int ori, int mismatch, map<int, vector<vector<int> > >& Gapped, const std::string & fa_whole, int b, int e, map<string, set<int> >& RefIndex){
	int found = 0;
        int Hamm1, Hamm2;
        int flag = 0;
	int spliceOri = 1;
	if(start < b || end > e)
		return 0;
        if(ori == 2) singleRead=revString(singleRead);
        int Len = singleRead.length();
	if(Len < segLength)
		return 0;
	int Len_ref = fa_whole.length();
        vector<int> information;
//	cout << singleRead << endl;
        for(int i=0; i<=segLength;i++){
//		cout << i <<endl;
		if(start+ tile_length + i - b < 0 || start+ tile_length + i - b + 2 > Len_ref)
			continue;
		string canonical1 = fa_whole.substr(start+ tile_length + i - b, 2);
		string canonical2 = fa_whole.substr (end-(segLength-i)-2 - b,2);
//		cout << canonical1 <<"\t"<<canonical2<<endl;
                if(cano)
                        if(!( (canonical1 == "GT"&& canonical2 == "AG") || (canonical1 == "CT"&& canonical2 == "AC")))
                                continue;
		if(canonical1 == "CT")
                        spliceOri = 0;//-
		else if(canonical1 == "GT")
			spliceOri = 1;//+
		else
			spliceOri = 2;//undecided
		
		Hamm1 = HammingDis(singleRead.substr(0+tile_length, i), fa_whole.substr(start+tile_length-b, i),mismatch);
		if(Hamm1 > mismatch)
			continue;
		if(HammingDis(singleRead.substr(0+i , tile_length), fa_whole.substr(start-b+i, tile_length),mismatch) > 1){
			Hamm1 = 100;
		}
//		cout << Hamm1 << endl;
		if(Hamm1 <= mismatch){
			Hamm2 = HammingDis(singleRead.substr(i+tile_length,segLength-i), fa_whole.substr(end-(segLength-i)-b, segLength-i),mismatch-Hamm1);
			if(Hamm2 > mismatch-Hamm1)
				continue;
			if(HammingDis(singleRead.substr(i+tile_length,tile_length), fa_whole.substr(end-(segLength-i)-b, tile_length),mismatch-Hamm1) > 1){
				Hamm2 = 100;
			}
			//			cout << Hamm2 << endl;
			if(Hamm2 <= mismatch-Hamm1){
				information.clear();
				information.push_back(Part);
				information.push_back(start+tile_length+1);//+1?
				information.push_back(i);//al1
				information.push_back(end-(segLength-i) - (start+i) -tile_length);//gap
				information.push_back(segLength-i);//al2
				information.push_back(Hamm1);
				information.push_back(Hamm2);
				if(cano == 0)
					if((canonical1 == "GC"&& canonical2 == "AG") || (canonical1 == "CT"&& canonical2 == "GC") || (canonical1 == "AT"&& canonical2 == "AC") || (canonical1 == "GT"&& canonical2 == "AT")){
						if((canonical1 == "GC"&& canonical2 == "AG") || (canonical1 == "AT"&& canonical2 == "AC") )
							spliceOri = 1;//semi canonical

						else
							spliceOri = 0;

						cano = 2;//semi canonical
						vector< vector<int> >::iterator it2;
						for(it2 = Gapped[Part].begin(); it2!=Gapped[Part].end();){
							if((*it2)[7] == 0){
								Gapped[Part].erase(it2);
							}
							else it2++;
						}
					}

				information.push_back(cano);
				information.push_back(spliceOri);
				Gapped[Part].push_back(information);
				if(Gapped[Part].size() > 100){                                                                                                  
					return 0;
				}
				found = 1;

			}
		}
	}
	return found;
}
int new_filGap(int cano, int Part, string singleRead, int start, int end, int ori, int mismatch, map<int, vector<vector<int> > >& Gapped, const std::string & fa_whole, int b, int e, int flag){
//10 bp on either side
//===========================||=======================||======================
//------------------------------------------#############//#######----------------------------------
           //       start------20bp--- end
        int found = 0;
        int Hamm1, Hamm2;
      //  int flag = 0;
        int spliceOri = 1;
        if(start < b || end > e)
                return 0;
        if(ori == 2) singleRead=revString(singleRead);
        int Len = singleRead.length();
        int Len_ref = fa_whole.length();
        vector<int> information;
//      cout << singleRead << endl;
	if(flag == 1){   //0---------------------------1
		for(int i=1; i<=8;i++){ // set 20 as default
			//              cout << i <<endl;
			string canonical1 = fa_whole.substr(start + i - b, 2);
			string canonical2 = fa_whole.substr(end +i - b -2, 2);
			if(cano)
				if(!((canonical1 == "GT"&& canonical2 == "AG") || (canonical1 == "CT"&& canonical2 == "AC")))
					continue;
			if(canonical1 == "CT")
				spliceOri = 0;//-
			else if(canonical1 == "GT")
				spliceOri = 1;//+
			else
				spliceOri = 2;//undecided

			Hamm1 = HammingDis(singleRead.substr(0, i), fa_whole.substr(start-b, i),mismatch);
			if(Hamm1 <= mismatch){
				Hamm2 = HammingDis(singleRead.substr(i,10-i), fa_whole.substr(end + i - b, 10-i ),mismatch-Hamm1);
				if(Hamm2 <= mismatch-Hamm1){
					information.clear();
					information.push_back(Part);
					information.push_back(start + 1);//+1?
					information.push_back(i);//al1
					information.push_back(end - start);//gap
					information.push_back(Len - i);//al2
					information.push_back(Hamm1);
					information.push_back(Hamm2);
					if(cano == 0)
						if((canonical1 == "GC"&& canonical2 == "AG") || (canonical1 == "CT"&& canonical2 == "GC") || (canonical1 == "AT"&& canonical2 == "AC") || (canonical1 == "GT"&& canonical2 == "AT")){
							if((canonical1 == "GC"&& canonical2 == "AG") || (canonical1 == "AT"&& canonical2 == "AC") )
								spliceOri = 1;//semi canonical

							else
								spliceOri = 0;

							cano = 2;//semi canonical
							vector< vector<int> >::iterator it2;
							for(it2 = Gapped[Part].begin(); it2!=Gapped[Part].end();){
								if((*it2)[7] == 0){
									Gapped[Part].erase(it2);
								}
								else it2++;
							}
						}
					information.push_back(cano);
					information.push_back(spliceOri);
					Gapped[Part].push_back(information);
					if(Gapped[Part].size() > 100){                                                                                                  
						return 0;
					}
					found = 1;
				}
			}
		}
		return found;
	}
	if(flag == 0){
		for(int i=0; i<=8;i++){ // set 20 as default
			//              cout << i <<endl;
			string canonical1 = fa_whole.substr(start - i - b, 2);
			string canonical2 = fa_whole.substr(end - i - b -2, 2);
			if(cano)
				if(!((canonical1 == "GT"&& canonical2 == "AG") || (canonical1 == "CT"&& canonical2 == "AC")))
					continue;
			if(canonical1 == "CT")
				spliceOri = 0;//-
			else if(canonical1 == "GT")
				spliceOri = 1;//+
			else
				spliceOri = 2;//undecided
		
			Hamm1 = HammingDis(singleRead.substr(Len-10, 10-i), fa_whole.substr(start-b-10, 10-i),mismatch);
			if(Hamm1 <= mismatch){
				Hamm2 = HammingDis(singleRead.substr(Len - i,i), fa_whole.substr(end - i - b, i ),mismatch-Hamm1);
				if(Hamm2 <= mismatch-Hamm1){
					information.clear();
					information.clear();
					information.push_back(Part);
					information.push_back(start - Len + 1);//+1?
					information.push_back(Len - i);//al1
					information.push_back(end - start);//gap
					information.push_back(i);//al2
					information.push_back(Hamm1);
					information.push_back(Hamm2);
					if(cano == 0)
						if((canonical1 == "GC"&& canonical2 == "AG") || (canonical1 == "CT"&& canonical2 == "GC") || (canonical1 == "AT"&& canonical2 == "AC") || (canonical1 == "GT"&& canonical2 == "AT")){
							if((canonical1 == "GC"&& canonical2 == "AG") || (canonical1 == "AT"&& canonical2 == "AC") )
								spliceOri = 1;//semi canonical

							else
								spliceOri = 0;

							cano = 2;//semi canonical
							vector< vector<int> >::iterator it2;
							for(it2 = Gapped[Part].begin(); it2!=Gapped[Part].end();){
								if((*it2)[7] == 0){
									Gapped[Part].erase(it2);
								}
								else it2++;
							}
						}
					information.push_back(cano);
					information.push_back(spliceOri);
					Gapped[Part].push_back(information);
					if(Gapped[Part].size() > 100){                                                                                                  
						return 0;
					}
					found = 1;

				}
			}
		}
		return found;
	}
}




void readFq(ifstream& input_fq){
	string fq_line;
	string Name_fq, Name_temp;
	int flag=0;
	while(!input_fq.eof()){
		getline(input_fq, fq_line);
		if(fq_line.length()==0)  //in case of a blank line
			break;
		istringstream ss(fq_line);
		ss >> Name_temp;
		if(Name_temp.length()<2)
			break;
		Name_fq = Name_temp.substr(1);
		//		cout << Name_fq <<endl;
		getline(input_fq, fq_line);
		if(flag == 0){
			read_length = fq_line.length();
			flag = 1;
		}
		if(ALL_READ.find(Name_fq) != ALL_READ.end()){
			InputFastq[Name_fq] = fq_line;
			getline(input_fq, fq_line);
			getline(input_fq, fq_line);
			InputQ[Name_fq] = fq_line;
		}
		else{
			getline(input_fq, fq_line);
			getline(input_fq, fq_line);
		}
	}
	cout << ALL_READ.size() << endl;
	ALL_READ.clear();

	//	cout << read_length-(N_fragment-1)*halfLength << endl;
}

int search(int flag, int i, int pos, int ori, string Chr, string name);

void readBwt(ifstream& input_bwt){
	string bwt_line;
	string Name_Index;
	char Orien;
	int Position, part;
	while(!input_bwt.eof()){
		getline(input_bwt, bwt_line);
		if(bwt_line.length()==0)//in case of blank line
			break;
		istringstream ss(bwt_line);
		string Chr,temp1,temp2,err;

		ss >> Name_Index >> Orien >> Chr >> Position >> temp1 >> temp2 >> err;

		part = atoi( Name_Index.substr(Name_Index.length()-1).c_str() );
		string Name = Name_Index.substr(0,Name_Index.length()-2);
		int Ori;
		if(Orien == '+'){Ori=1;}
		if(Orien == '-'){Ori=2;}
		Part P(Ori, Position, part, int(err.length()/5), Chr,Name);
		InputFastqPart[Name].push_back(P);
	}

	for(ItFastqPart = InputFastqPart.begin();ItFastqPart != InputFastqPart.end();ItFastqPart++){
		Temp.clear();
		PartHash.clear();
		string Name = ItFastqPart->first;
		if(N_fragment >= 3){
			if(int(ItFastqPart->second.size()) > N_fragment*20){
				continue;
			}
		}

		for(int i=0;i!=int(ItFastqPart->second.size());i++){

			PartHash[ItFastqPart->second[i].part].push_back(i);
			ItFastqPart->second[i].locat = i;
		}
		local_CandidateVector.clear();
		search_flag = 0;
		search_flag_2 = 0;
		int lack_segment = 0;
		if(N_fragment > 2){
			for(int jj = 1; jj < N_fragment; jj++){
				if(PartHash.find(jj) == PartHash.end() && PartHash.find(jj+1) == PartHash.end()){
					lack_segment = 1;
					break;
				}
			}
			if(lack_segment == 1){
				continue;
			}
		}

		int i=1;
		if(PartHash.find(i) == PartHash.end()){
			if(PartHash.find(i+1) == PartHash.end())
				continue;
			else{
				Part P(-1,-1,-1,-1,"fake", Name);
				Temp.push_back(P);
				for(int j=0;j != PartHash[i+1].size(); j++){
					Part P = ItFastqPart->second[PartHash[i+1][j]];
					Temp.push_back(P);
					search(1,3,P.Pos,P.Ori,P.Chr,P.name);
					Temp.pop_back();
				}
			}
		}
		else{

			for(int j=0;j != PartHash[i].size(); j++){
				//	int ori = InputFastqPart[Name][ PartHash[i][j] + 1];
				//	int Chr = InputFastqPart[Name][ PartHash[i][j] + 2];	
				Part P = ItFastqPart->second[PartHash[i][j]];
				Temp.push_back(P);



				search(1,2,P.Pos,P.Ori,P.Chr,P.name);

				Temp.pop_back();
			}
			Part P(-1,-1,-1,-1,"fake", Name);
			Temp.push_back(P);
			for(int j=0;j != PartHash[i+1].size(); j++){
				if(ItFastqPart->second[PartHash[i+1][j]].flag == 0){
					Part P = ItFastqPart->second[PartHash[i+1][j]];
					Temp.push_back(P);
					search(1,3,P.Pos,P.Ori,P.Chr,P.name);
					Temp.pop_back();
				}
			}
		}

		PartHash.clear();
		Temp.clear();
		//		cout << "ssssssssss\n";
		if(search_flag_2 != 0){
			local_CandidateVector.clear();
		}              ///
		else{
			CandidateVector.insert(CandidateVector.end(), local_CandidateVector.begin(), local_CandidateVector.end() );
		}
	}
	InputFastqPart.clear();

	for (int i = 0; i < CandidateVector.size(); i++){
		string chr = CandidateVector[i][1].Chr;
		total_chr.insert(chr);
		ALL_READ[CandidateVector[i][1].name] = 1;
		int b,e;
		if(CandidateVector[i][0].Pos > 0)
			b = CandidateVector[i][0].Pos;
		else
			b = CandidateVector[i][1].Pos;
		e = CandidateVector[i][N_fragment-1].Pos;
		//		cout << i <<"\t"<<b<<"\t"<<e << endl;

		if(CandidateVector[i][1].Ori == 1) {

			merge_interval(&CandidateList[chr],b-10000,e+halfLength+10000,i);
		}
		else{
			merge_interval(&CandidateList[chr],e-10000,b+halfLength+10000,i);
		}
	}

}




int action(string chr){
	job_list.clear();
	for(intervalMap::iterator it = CandidateList[chr].begin(); it!=CandidateList[chr].end(); it++){
		job J;
		J.chr = chr;
		J.b = it->first.start;
		J.e = it->first.end;
		J.list = it->second;     
		job_list.push_back(J);
	}
#pragma omp parallel for num_threads(N_core)
	for(int i=0; i< job_list.size(); i++){
		map< string, set<int> > RefIndex;
		RefIndex.clear();
		string fa_whole;
		fa_whole.clear();
		int b = job_list[i].b - Intron_length;
		int e = job_list[i].e + Intron_length;
		getRefIndex(b, e, RefIndex, fa_whole);
		vector <string> T, T_multi, NON, NON_multi;//
		for(set<int>::iterator it = job_list[i].list.begin(); it != job_list[i].list.end(); it++){
			int Num = *it; 
			map<int, vector<vector<int> > > Gapped;
			map<int, vector<vector<int> > >::iterator ItGapped;
			Gapped.clear();
			string name = CandidateVector[Num][1].name;
			int job_ori = CandidateVector[Num][1].Ori;
			int find=0;
			if(InputFastq[name].length() < (N_fragment-1)*halfLength)
				continue;
			int err = 0;
			for(int t=0;t<N_fragment;t++){
				if(t < N_fragment - 1){
					if(CandidateVector[Num][t+1].part == -2){
						err += 0;
						continue;
					}
				}

				if(CandidateVector[Num][t].part == -1 || CandidateVector[Num][t].part == -3){
					err += 0;
				}
				else
					err += CandidateVector[Num][t].segMis;

			}
			if(err > totalMis)
				continue;
			// MAY 30
			for(int t=0;t<N_fragment;t++){	
				if( (CandidateVector[Num][0].part == -1 || CandidateVector[Num][1].part == -2) && t==0){
					if(CandidateVector[Num][1].Pos < b || CandidateVector[Num][1].Pos > e)
						break;
					if(CandidateVector[Num][1].Ori == 1){


						find = extendSegment2(1, 1, InputFastq[name].substr(0,2*halfLength), CandidateVector[Num][1].Pos , halfLength, 1 , totalMis-err , 0, Gapped , b, e, RefIndex, fa_whole);
						if(find == 0){
							if(only_canonical != 1){
								find = extendSegment2(0, 1, InputFastq[name].substr(0,2*halfLength), CandidateVector[Num][1].Pos , halfLength, 1 , 0 , 0, Gapped , b, e, RefIndex, fa_whole);
							}
						}
					}

					if(CandidateVector[Num][1].Ori == 2){
						find = extendSegment1(1, 1, InputFastq[name].substr(0,2*halfLength), CandidateVector[Num][1].Pos , halfLength, 2 , totalMis-err , 0, Gapped , b, e, RefIndex, fa_whole);
						if(find == 0){
							if(only_canonical != 1){
								find = extendSegment1(0, 1, InputFastq[name].substr(0,2*halfLength), CandidateVector[Num][1].Pos , halfLength, 2 , 0 , 0, Gapped , b, e, RefIndex, fa_whole);
							}
						}
					}
					if(find == 0)
						break;
					continue;
				}
				if( CandidateVector[Num][N_fragment-1].part == -1 && t==N_fragment-1){
					if(CandidateVector[Num][N_fragment-1].Pos < b || CandidateVector[Num][N_fragment-1].Pos > e)
						break;
					if(CandidateVector[Num][N_fragment-2].Ori == 1){
						find = extendSegment1(1, N_fragment, InputFastq[name].substr((N_fragment-2)*halfLength), CandidateVector[Num][N_fragment-1].Pos , halfLength, 1 , totalMis-err , 0, Gapped , b, e, RefIndex, fa_whole);
						if(find == 0 && only_canonical != 1)
							find = extendSegment1(0, N_fragment, InputFastq[name].substr((N_fragment-2)*halfLength), CandidateVector[Num][N_fragment-1].Pos , halfLength, 1 , 0 , 0, Gapped , b, e, RefIndex, fa_whole);
					}

					if(CandidateVector[Num][N_fragment-2].Ori == 2){
						find = extendSegment2(1, N_fragment, InputFastq[name].substr((N_fragment-2)*halfLength), CandidateVector[Num][N_fragment-1].Pos , halfLength, 2 , totalMis-err , 0, Gapped ,b, e, RefIndex, fa_whole);
						if(find == 0 && only_canonical != 1)
							find = extendSegment2(0, N_fragment, InputFastq[name].substr((N_fragment-2)*halfLength), CandidateVector[Num][N_fragment-1].Pos , halfLength, 2 , 0 , 0, Gapped ,b, e, RefIndex, fa_whole);
					}
					if(find == 0)
						break;
					continue;
				}
				if( ( CandidateVector[Num][t].part == -1 || CandidateVector[Num][t+1].part == -2)){
					if(!(Gapped.find(t) == Gapped.end() && Gapped.find(t+2) == Gapped.end())){
						find = 0;
						break;
					}
					if(t*halfLength - tile_length<0 || (t+1)*halfLength + tile_length > InputFastq[name].length())
						break;
					if(CandidateVector[Num][t].Ori == 1){
						if(CandidateVector[Num][t-1].Pos + halfLength-tile_length < b || CandidateVector[Num][t+1].Pos > e)
							break;
						find = filGap(1,t+1, InputFastq[name].substr(t*halfLength - tile_length, halfLength + 2*tile_length), CandidateVector[Num][t-1].Pos + halfLength-tile_length, CandidateVector[Num][t+1].Pos, halfLength, CandidateVector[Num][t].Ori, totalMis - err, Gapped, fa_whole, b, e, RefIndex);
						if(find == 0 && only_canonical != 1)
							find = filGap(0,t+1, InputFastq[name].substr(t*halfLength - tile_length, halfLength + 2*tile_length), CandidateVector[Num][t-1].Pos + halfLength-tile_length, CandidateVector[Num][t+1].Pos, halfLength, CandidateVector[Num][t].Ori, 0, Gapped, fa_whole, b, e, RefIndex);
					}

					else{
						if(CandidateVector[Num][t+1].Pos-tile_length + halfLength < b || CandidateVector[Num][t-1].Pos > e)
							break;
						find = filGap(1,t+1, InputFastq[name].substr(t*halfLength - tile_length, halfLength + 2*tile_length), CandidateVector[Num][t+1].Pos-tile_length + halfLength, CandidateVector[Num][t-1].Pos, halfLength, CandidateVector[Num][t].Ori, totalMis - err, Gapped, fa_whole, b, e, RefIndex);
						if(find == 0 && only_canonical != 1)
							find = filGap(0,t+1, InputFastq[name].substr(t*halfLength - tile_length, halfLength + 2*tile_length), CandidateVector[Num][t+1].Pos-tile_length + halfLength, CandidateVector[Num][t-1].Pos, halfLength, CandidateVector[Num][t].Ori, 0, Gapped, fa_whole, b, e, RefIndex);
					}
					if(find == 0)
						break;
					continue;
				}
			}

			if(find == 0)
				continue;

			int multi_flag = 0;
			for(ItGapped = Gapped.begin();ItGapped != Gapped.end(); ItGapped++){
				vector< vector<int> >::iterator it; 
				set<int> p;
				p.clear();
				int min = 1;//from 100 -> 0  06/23/11
				int c = 0;
				for(it = ItGapped->second.begin(); it!=ItGapped->second.end();it++,c++ ){
					int mis = (*it)[5]+(*it)[6];
					if(mis<min) { min=mis;p.clear();p.insert(c);}
					if(mis == min){p.insert(c);}// major difference here, set p as single value 
				}
				if(p.size() == 0){
					multi_flag = 1;
					continue;
				}
				c = 0;
				for(it = ItGapped->second.begin(); it!=ItGapped->second.end();c++ ){
					if(p.find(c) == p.end()){
						ItGapped->second.erase(it);
					}
					else it++;
				}
				if(p.size()>1)
					multi_flag = 1;
			}
			int SAM_start;
			int ali=0;
			ostringstream CIGAR;
			CIGAR.clear();
			int mis = err; /// MAY 30 !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
			int cano = -1;
			if(multi_flag == 0){
				int signalOri = 0;

				if(job_ori == 1){
					if(Gapped.find(1) != Gapped.end()){
						SAM_start=Gapped[1][0][1];
						signalOri = Gapped[1][0][8];
					}
					else
						SAM_start = CandidateVector[Num][0].Pos + 1;///////////////////////+1
					for(int t=0;t<N_fragment;t++){
						if(Gapped.find(t+1) != Gapped.end()){
							cano = Gapped[t+1][0][7];
							mis += Gapped[t+1][0][5] + Gapped[t+1][0][6];
							signalOri = signalOri > Gapped[t+1][0][8]?signalOri:Gapped[t+1][0][8];
							ali+=Gapped[t+1][0][2];
							CIGAR << ali <<"M"<<Gapped[t+1][0][3]<<"N";

							if(t == N_fragment-1)
								ali = read_length-halfLength*(N_fragment-1) - Gapped[t+1][0][2];
							else
								ali=Gapped[t+1][0][4];

						}
						else{
							if(t == N_fragment-1)
								ali += read_length-halfLength*(N_fragment-1);
							else
								ali+=halfLength;
						}
					}
					CIGAR << ali <<"M";
				}
				else{
					if(Gapped.find(N_fragment) != Gapped.end()){
						SAM_start=Gapped[N_fragment][0][1];
						signalOri = Gapped[N_fragment][0][8];
					}
					else
						SAM_start = CandidateVector[Num][N_fragment-1].Pos + 1 ;////////////////////////////+1
					for(int t=N_fragment-1;t>=0;t--){
						if(Gapped.find(t+1) != Gapped.end()){
							cano = Gapped[t+1][0][7];
							mis += Gapped[t+1][0][5] + Gapped[t+1][0][6];
							signalOri = signalOri > Gapped[t+1][0][8]?signalOri:Gapped[t+1][0][8];
							if(t == N_fragment-1)
								ali+= read_length-halfLength*(N_fragment-1) - Gapped[t+1][0][4];
							else
								ali+=Gapped[t+1][0][2];

							CIGAR << ali <<"M"<<Gapped[t+1][0][3]<<"N";

							ali=Gapped[t+1][0][4];

						}
						else{
							if(t == N_fragment-1)
								ali += read_length-halfLength*(N_fragment-1);
							else
								ali+=halfLength;
						}
					}
					CIGAR << ali <<"M";
				}

				SAM SAM_temp;
				SAM_temp.name = name;
				SAM_temp.ori = (job_ori == 1)? 0:16;
				SAM_temp.cigar = CIGAR.str();
				if(job_ori == 1){
					SAM_temp.seq = InputFastq[name]; 		
					SAM_temp.qua = InputQ[name];
				}
				else{
					SAM_temp.seq = revString(InputFastq[name]);
					SAM_temp.qua = revQ(InputQ[name]);
				}

				SAM_temp.chr = job_list[i].chr;
				SAM_temp.b = SAM_start;// !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!! +1 based
				ostringstream out_temp;
				out_temp << name <<"\t"<<SAM_temp.ori<<"\t"<<SAM_temp.chr<<"\t"<<SAM_temp.b<<"\t255\t"<<SAM_temp.cigar<<"\t"<<cano<<"\t"<<mis<<"\t1\t"<<SAM_temp.seq<<"\t"<<SAM_temp.qua<<"\tNM:i:1\tXX:A:"<<signalOri<<"\tNH:i:1\n";
				if(signalOri < 2 && cano == 1)
					T.push_back(out_temp.str());
				else
					NON.push_back(out_temp.str());
			}
			else{
				ostringstream out_temp;
				out_temp.clear();
				SAM SAM_temp;
				SAM_temp.name = name;
				SAM_temp.ori = (job_ori == 1)? 0:16;
				if(job_ori == 1){
					SAM_temp.seq = InputFastq[name];
					SAM_temp.qua = InputQ[name];
				}
				else{
					SAM_temp.seq = revString(InputFastq[name]);
					SAM_temp.qua = revQ(InputQ[name]);
				}

				SAM_temp.chr = job_list[i].chr;
				int signalOri = 0;
				out_temp <<">\n";
				out_temp << name <<"\t"<<SAM_temp.ori<<"\t"<<SAM_temp.chr<<"\t"<<"begin"<<"\t255\t"<<"CIGAR"<<"\t*\t"<< err <<"\t0\t"<<SAM_temp.seq<<"\t"<<SAM_temp.qua<<"\tNM:i:1\tXX:A:x\tNH:i:1\n";
				for(int t=0;t<N_fragment-1;t++){
					if(Gapped.find(t+1) != Gapped.end()){
						cano = Gapped[t+1][0][7];
						signalOri = signalOri > Gapped[t+1][0][8]?signalOri:Gapped[t+1][0][8];
						for(int k = 0; k < Gapped[t+1].size(); k++){

							for(int j = 0; j<9; j++){
								out_temp << Gapped[t+1][k][j] <<"\t";
							}
							out_temp <<"\n";
						}
					}
				}
				int t = N_fragment - 1;
				if(Gapped.find(t+1) != Gapped.end()){
					cano = Gapped[t+1][0][7];
					signalOri = signalOri > Gapped[t+1][0][8]?signalOri:Gapped[t+1][0][8];
					for(int k = 0; k < Gapped[t+1].size(); k++){
						if(job_ori == 1){
							Gapped[t+1][k][4] = read_length-halfLength*(N_fragment-1) - Gapped[t+1][k][2];
						}
						else{
							Gapped[t+1][k][2] = read_length-halfLength*(N_fragment-1) - Gapped[t+1][k][4];
						}
						for(int k = 0; k < Gapped[t+1].size(); k++){
							for(int j = 0; j<9; j++){
								out_temp << Gapped[t+1][k][j] <<"\t";
							}
							out_temp <<"\n";
						}

					}
				}
				out_temp <<">\n";
				if(signalOri < 2 && cano == 1)
					T_multi.push_back(out_temp.str());
				else
					NON_multi.push_back(out_temp.str());


			}
		}
#pragma omp critical
		{
			for(int t =0; t < T.size();t++){
				Out << T[t];
			}
			for(int t =0; t < T_multi.size();t++){
				Out_multi << T_multi[t];
			}
			for(int t =0; t < NON.size();t++){
				Out_NON << NON[t];
			}
			for(int t =0; t < NON_multi.size();t++){
			}
		}
	}
}
int run_from_ref(ifstream& input_ref){
	string fa_line,chr,chr_temp;
	int flag = 0;
	int F = 1;
	while(!input_ref.eof()){
		getline(input_ref, fa_line);
		if(fa_line.length() < 1)
			continue;
		if(fa_line.substr(0,1) == ">"){
			istringstream ss(fa_line);
			string fa;
			ss >> fa;
			chr = fa.substr(1);
			if(flag == 0){
				flag = 1;
				chr_temp = chr;
			}
			else{
				if(F == 1){
					transform(single_ref.begin(),single_ref.end(),single_ref.begin(),::toupper);
					action(chr_temp);
					total_chr.erase(chr_temp);
				}
				if(chr_temp.size() == 0)
					return 0;
				chr_temp = chr;
				single_ref = "";
			}

			if(total_chr.find(chr) == total_chr.end())
				F = 0;
			else
				F = 1;
			continue;

		}
		if(F == 1)
			single_ref += fa_line;
	}
	transform(single_ref.begin(),single_ref.end(),single_ref.begin(),::toupper);
	action(chr_temp);
	return 0;
}
int search(int flag, int i, int pos, int ori, string Chr, string name){
	if(i-1 == N_fragment || flag == -1){
		int f=0;
		//cout << Temp[0].name <<"\t";
		int err = 0;
		for(int j=0;j != Temp.size(); j++){
			//cout << Temp[j].part <<"\t"<< Temp[j].Pos <<"\t"<< Temp[j].Ori <<"\t"<<Temp[j].Chr <<"\t";
			if(Temp[j].part<0){f=1;}
			if(j+1 <= Temp.size())
				if(Temp[j+1].part == -2){
					err += 0;
					continue;
				}
			if(Temp[j].part == -1 || Temp[j].part == -3){
				err += 0;
				continue;
			}
			err += Temp[j].segMis;
		}
		if(totalMis < err){
			return 0;
		}
		if(Temp[1].locat != -1){
			int locat = Temp[1].locat;
			//			cout << locat <<"\t"<<ItFastqPart->second[locat].Pos;
			ItFastqPart->second[locat].flag = -1;
		}
		//cout << endl;
		if(f==1){
			local_CandidateVector.push_back(Temp);
		}
		if(f==0){
			int final_ori = (Temp[0].Ori == 1)? 0:16;
			Out_full << Temp[0].name <<"\t"<<final_ori<<"\t"<<Temp[0].Chr<<"\t"<<Temp[0].Pos<<"\t255\t"<<read_length<<"M\t*\t0\t0\t"<<InputFastq[Temp[0].name]<<"\t"<<InputQ[Temp[0].name]<<"\tNM:i:1\tXX:A:1\tNH:i:1\n";
			search_flag_2 = 2;
		}
		search_flag = 1;
		return 1;

	}
	int F=0;
	if(flag == 1){
		if(PartHash.find(i) == PartHash.end()){
			Part P(ori,pos,-1,-1,Chr,name);
			Temp.push_back(P);
			search(0,i+1,pos,ori,Chr,name);
			Temp.pop_back();
		}
		else{
			for(int j=0;j != PartHash[i].size(); j++){
				Part P = ItFastqPart->second[PartHash[i][j]];
				if(P.Chr != Chr || P.Ori != ori){
					continue;
				}
				if(ori == 1 && P.Pos<=pos || ori == 2 && P.Pos>=pos || abs(P.Pos-pos) > Intron_length){
					continue;
				}
				if(ori == 1){
					if(abs(pos + halfLength - P.Pos) <= 2 ){

						Temp.push_back(P);
						//ut_multi << T_multi[t];
						search(1,i+1,P.Pos,ori,Chr,name);
						Temp.pop_back();
						F=1;
						break;
					}
				}
				if(ori == 2){

					if(i != N_fragment){
						if(abs(pos - halfLength - P.Pos) <= 2 ){
							Temp.push_back(P);
							search(1,i+1,P.Pos,ori,Chr,name);
							Temp.pop_back();
							F=1;
							break;
						}
					}
					if(i == N_fragment){
						if(abs(pos - (read_length-(N_fragment-1)*halfLength) - P.Pos) <= 2 ){
							Temp.push_back(P);
							search(1,i+1,P.Pos,ori,Chr,name);
							Temp.pop_back();
							F=1;
							break;
						}
					}
				}
			}
			if(F==0){
				for(int j=0;j != PartHash[i].size(); j++){
					Part P = ItFastqPart->second[PartHash[i][j]];
					if(P.Chr != Chr || P.Ori != ori){
						continue;
					}
					if(ori == 1 && P.Pos<=pos || ori == 2 && P.Pos>=pos || abs(P.Pos-pos) > Intron_length || abs(P.Pos-pos) < min_intron + 2*halfLength){ //////////
						continue;
					}
					if(ori == 1){
						if(abs(pos + halfLength - P.Pos) > 2){
							P.part = -2; // break point on the split site
							Temp.push_back(P);
							search(2,i+1,P.Pos,ori,Chr,name);							
							Temp.pop_back();
							F = 1;
						}
					}
					if(ori == 2){
						if(i == N_fragment){
							if(abs(pos - (read_length-(N_fragment-1)*halfLength) - P.Pos) > 2){
								P.part = -2;
								Temp.push_back(P);
								search(2,i+1,P.Pos,ori,Chr,name);
								Temp.pop_back();
								F = 1;
							}
						}
						if(i != N_fragment){
							if(abs(pos - halfLength - P.Pos) > 2){
								P.part = -2;
								Temp.push_back(P);
								search(2,i+1,P.Pos,ori,Chr,name);
								Temp.pop_back();
								F = 1;
							}
						}
					}
				}
				//				if(F == 0)
				Part P(ori,pos,-1,-1,Chr,name);
				//				if(F == 1)
				//					Part P(ori,pos,-3,-3,Chr,name);
				Temp.push_back(P);
				search(0,i+1,pos,ori,Chr,name);
				Temp.pop_back();




			}
		}
	}
	if(flag == 0 || flag == 2){
		if(PartHash.find(i) == PartHash.end()){
			return 0;
		}
		else{
			for(int j=0;j != PartHash[i].size(); j++){
				Part P = ItFastqPart->second[PartHash[i][j]];
				if(P.Chr != Chr || P.Ori != ori){
					continue;
				}
				if(ori == 1 && P.Pos<=pos || ori == 2 && P.Pos>=pos || abs(P.Pos-pos) > Intron_length || (abs(P.Pos-pos) < min_intron + 2*halfLength && flag == 0)){   /////////////July 13
					continue;
				}
				if(ori == 1){
					if((flag == 2 && (pos + halfLength == P.Pos)) || flag == 0){
						Temp.push_back(P);
						search(1,i+1,P.Pos,ori,Chr,name);
						Temp.pop_back();
						F=1;
						continue;
					}
				}
				if(ori == 2){
					if(i != N_fragment){
						if((flag == 2 && (pos - halfLength == P.Pos)) || flag == 0){
							Temp.push_back(P);
							search(1,i+1,P.Pos,ori,Chr,name);
							Temp.pop_back();
							F=1;
							continue;
						}
					}
					if(i == N_fragment){
						if((flag == 2 && (pos - (read_length-(N_fragment-1)*halfLength) == P.Pos)) || flag == 0){
							Temp.push_back(P);
							search(1,i+1,P.Pos,ori,Chr,name);
							Temp.pop_back();
							F=1;
							continue;
						}
					}


				}
			}
			if(F==0){
				return 0;
			}
		}
	}	
}




int main(int argc, char *argv[]){
	if(argc != 14){
		usage();
		return 1;
	}
	Intron_length = atoi(argv[11]);
	read_length = atoi(argv[13]); 
	min_intron = atoi(argv[12]);
	tile_length = atoi(argv[4]);
	MINOVLP = atoi(argv[5]);
	totalMis = atoi(argv[6]);
	mask = atoi(argv[7]);
	ifstream input_ref_name(argv[8]);
	ifstream input_fq(argv[9]);
	N_fragment = atoi(argv[1]);
	halfLength = atoi(argv[2]);
	ifstream input_bwt(argv[3]);
	N_core = atoi(argv[10]);
	readBwt(input_bwt);
	readFq(input_fq);	
	run_from_ref(input_ref_name);
	return (0);
}


