#include <iostream>
#include <string>
#include <vector>
#include <map>
//#include <iomanip>
#include <stdlib.h>
#include <sstream>
#include <fstream>

using namespace std;


//////////////////////////////////////////////////

#define PROG_NAME "getCov"
#define PROG_VER "v0.1"


ofstream Out_cov;
map<string, map<int, int> > full_cov;

string samFile, junctFile;
int readLength;

bool parseOptions (int argc, const char** argv)
{
	if (argc < 4) {
		std::cout << std::endl;
		std::cout << "Usage: " << PROG_NAME << " <read_len> <sam> <jcns>" << std::endl;
		std::cout << " Where" << std::endl;
		std::cout << "  <read_len>     Read length" << std::endl;
		std::cout << "  <sam>          File containing solid (un-gapped) read alignments (temp.sam.sort)" << std::endl;
		std::cout << "  <jcns>         File containing candidate junctions (output of junctionEva.pl)" << std::endl;
		std::cout << std::endl;
		std::cout << "\nVersion: " << PROG_VER << std::endl;
		std::cout << std::endl;
		std::cout << "This program carries out a EM with logistic regression learning procedure to classify a set" << std::endl;
		std::cout << " of candidate introns as either true (correct transcript origin) or not." << std::endl;
		std::cout << std::endl;
		return true;
	}

	//Parse argv for required options
	int argCount = 0;
	string setNot;

	for (int i=1; i<argc; ++i) {
		std::string opt = (char*)argv[i];

		if (argCount==0)
		{
			readLength = atoi( opt.c_str() );
			++argCount;
		}
		else if (argCount==1)
		{
			samFile = opt;
			++argCount;
		}
		else if (argCount==2)
		{
			junctFile = opt;
			++argCount;
		}
		else
			return true;
	}

	if (argCount != 3) {
		std::cout << "Incorrect number of arguments\n" << std::endl;
		return true;		// incorrect number of arguments
	}

	return false;
}


double getDiffCov(string chr, int pos, int flag){
	// flag = 1, first, 2 , second
	double first = 0, second = 0 , mid = 0;
	for(int i = pos - readLength - 100; i < pos - readLength; i++){
		if(full_cov[chr].find(i) != full_cov[chr].end()){
			first += full_cov[chr][i];
		}
	}
	for(int i = pos; i <= pos + 100; i++){
		if(full_cov[chr].find(i) != full_cov[chr].end()){
			second += full_cov[chr][i];
		}
	}

	if(flag == 1)
		return first/100 - second/100;
	else
		return second/100 - first/100;
}

void writeCovDiff(string& fname)
{
	ifstream input_junctions( fname.c_str() );

	string chr, b ,e, f, junc_line, num, min_over, err, entropy, bad_frame, bad_strand, dc;

	double cov_score_first, cov_score_second;

	while (!input_junctions.eof())
	{
		getline(input_junctions, junc_line);
		istringstream ss(junc_line);
		ss >> chr >> b >> e >> f >> num >> min_over >> entropy >> dc >> err >> bad_frame >> bad_strand;

		int first = atoi(b.c_str());
		int second = atoi(e.c_str());

		if (second-first <= 0)
			continue;

		cov_score_first = getDiffCov(chr, first, 1);
		cov_score_second = getDiffCov(chr, second, 2);

		Out_cov << chr << "\t" << b << "\t" << e << "\t" << cov_score_first << "\t" << cov_score_second << std::endl;
	}

}


void getCov(ifstream& input_fullBWT){
	string line, chr, temp1, temp2, begin;
	int b;
	string chr_temp = "";
	int b_temp = -1;
	while(!input_fullBWT.eof()){
		getline(input_fullBWT, line);
		if(line.length()==0)  //in case of a blank line
			break;
		istringstream ss(line);
		vector<string> T;
		while(ss){
			string s;
			if (!getline( ss, s, '\t' )) break;
			T.push_back(s);
		}

		b = atoi(T[3].c_str());
		chr = T[2];
		full_cov[chr][b]++;//memory issues??

	}
}



string getBinDir(const char *cmd)
{
	string bin = cmd;
	bin = bin.substr(0, bin.rfind("/")+1);
	return bin;
}


int main(int argc, const char *argv[])
{
	bool bTroubleshoot(false);

	if (parseOptions(argc, argv))
	{
		if (argc > 1) {
			std::cout << "Unacceptable argument(s) given.\n";
			return 0;
		}
		return 0;
	}

	ifstream input_fullBWT( samFile.c_str() );
	getCov(input_fullBWT);


	Out_cov.open("cov_diffs");

	writeCovDiff( junctFile );

	Out_cov.close();


	return 0;
}

