
#include "histogram.h"
#include "stdio.h"


// for automating histogram sort algorithm
//extern bool intLessThan (twoInt ti, twoInt tj);
//extern bool dvLessThan (DoubleXY di, DoubleXY dj);
bool intLessThan (twoInt ti, twoInt tj) { return (ti.i<tj.i); }
bool dvLessThan (DoubleXY di, DoubleXY dj) { return (di.x<dj.x); }

void cHistogram::loadHist( intHist& hist, string strFile )
{
	hist.clear();

	FILE *pF = fopen( strFile.c_str(), "r" );
	if (pF==NULL)
		return;		// error

	int ival;
	while ( fscanf(pF, "%d\n", &ival) != EOF )
	{
		updateHistogram( hist, ival );
	}

	fclose( pF );
}

void cHistogram::getHist( intHist& hist, vector<int> data )
{
	hist.clear();

	for ( int i=0; i<(int)data.size(); ++i )
	{
		updateHistogram( hist, data[i] );
	}
}

void cHistogram::saveHist( intDoubleHist& hist, string strFile )
{
	FILE *pF = fopen( strFile.c_str(), "w" );
	if (pF==NULL)
		return;		// error

	for (int i=0; i<(int)hist.size(); ++i)
	{
		fprintf( pF, "%d \t %9.7f\n", hist[i].i, hist[i].y );
	}

	fclose( pF );
}

void cHistogram::saveHist( intHist& hist, string strFile )
{
	FILE *pF = fopen( strFile.c_str(), "w" );
	if (pF==NULL)
		return;		// error

	for (int i=0; i<(int)hist.size(); ++i)
	{
		fprintf( pF, "%d \t %d\n", hist[i].i, hist[i].n );
	}

	fclose( pF );
}

void cHistogram::calcHistFromFile(string strOutFile, string strInFile, int itype, bool bSmooth3, bool bScale)
{
	printf("Loading data file...\n");

	intHist rawHist;
	loadHist( rawHist, strInFile );

	printf("Calculating histogram...\n");
	intDoubleHist outHist;
	string strOut = "default";
	getSmoothHist( outHist, rawHist, 100, 10000, itype, bSmooth3, bScale, strOut );

	printf("Saving file...\n");
	saveHist( outHist, strOutFile );
}

void cHistogram::interpHist( intDoubleHist& outHist, doubleHist& inHist, int nMin, int nMax )
{
	intDouble di;

	outHist.clear();
	double lfx = inHist[0].x;
	double lfy = inHist[0].y;

	int ix = int( lfx );
	int i, j, dj;
	int jcount = 0;
	double df, dx;

	// 0 to inHist[0].x
	for (j=0; j<=ix; ++j)
	{
		di.i = jcount;
		di.y = double(j)/double(ix) * lfy;
		outHist.push_back( di );
		jcount++;
	}

	// inHist[0].x to inHist[end].x
	for (i=1; i<(int)inHist.size(); ++i)
	{
		dx = inHist[i].x - inHist[i-1].x;
		for (j=int(inHist[i-1].x); j<int(inHist[i].x); ++j)
		{
			dj = j - int(inHist[i-1].x);
			di.i = jcount;
			df = double( dj ) / dx ;
			di.y = inHist[i-1].y + df*( inHist[i].y - inHist[i-1].y );
			jcount++;
		}
	}

	// inHist[end].x to nMax
	int n = (int)inHist.size();
	for (j=int(inHist[n-1].x); j<=nMax; ++j)
	{
		di.i = jcount;
		di.y = double(j)/double(ix) * lfy;
		outHist.push_back( di );
		jcount++;
	}
}

void cHistogram::getSmoothHist_Mod1( intDoubleHist& smoothHist, intHist& rawHist, int maxLen, int itype, bool bSmooth3, bool bScale )
{
	doubleHist logHist, frawHist;
	intDoubleHist tmpHist;
	intHist ihist;

	bool bCompress = bSmooth3 || bScale;
	if (bCompress) {
		compressHist( ihist, rawHist, 3 );
		copyHist( rawHist, ihist );
	}

	int nMin = 1;
	int nMax = maxLen;

	int2double( frawHist, rawHist );
	getLogBinHist( logHist, frawHist, 40, false );
	interpHist( smoothHist, logHist, nMin, nMax );

	if (bCompress)
	{
		copyHist( tmpHist, smoothHist );
		expandHist( smoothHist, tmpHist, 3, bScale );
	}

	normalizeHist( smoothHist );
}

void cHistogram::getSmoothHist( intDoubleHist& smoothHist, intHist& rawHist, int imin, int maxLen, int itype, bool bSmooth3, bool bScale, string& outFileBase )
{
	smoothHist.clear();

	// check for empty input
	if (rawHist.empty())
		return;

	// sort the raw input histogram data
	sort (rawHist.begin(), rawHist.end(), intLessThan);

	intDoubleHist tmpHist;
	intHist ihist;

	string strOut;

	bool bCompress = bSmooth3 || bScale;
	if (bCompress) {
		compressHist( ihist, rawHist, 3 );
		if (m_bDebug)
		{
			strOut = outFileBase + "_compressed.txt";
			saveHist( ihist, strOut );
		}
		copyHist( rawHist, ihist );
	}

	double mean, std;
	getStats( mean, std, rawHist );

	intHist cumHist;
	// get cumulative histogram and max (final) value as well
	double lfn = (double) getCumTotal( cumHist, rawHist );

	double K, W;
	getSmoothingParams( K, W, cumHist, std );

	if (itype==0)
	{
		smoothHist0( tmpHist, rawHist, maxLen, K, W, outFileBase );
		if (m_bDebug)
		{
			strOut = outFileBase + "_smooth0.txt";
			saveHist( tmpHist, strOut );
		}

//		smoothHist1( smoothHist, tmpHist, K, W );
		smoothHist2( smoothHist, tmpHist, imin, maxLen, K, W );
		if (m_bDebug)
		{
			strOut = outFileBase + "_smooth2.txt";
			saveHist( smoothHist, strOut );
		}
		normalizeHist( smoothHist );
	}
	else
	{
		smoothHist0( smoothHist, rawHist, maxLen, K, W, outFileBase );
	}

	if (bCompress)
	{
		copyHist( tmpHist, smoothHist );
		expandHist( smoothHist, tmpHist, 3, bScale );
		if (m_bDebug)
		{
			strOut = outFileBase + "_expanded.txt";
			saveHist( smoothHist, strOut );
		}
	}

	normalizeHist( smoothHist );

}

void cHistogram::compressHist( intHist& scaleHist, intHist& rawHist, int iscale )
{
	int i, ix;
	twoInt ti;

	scaleHist.clear();

	for (i=0; i<(int)rawHist.size(); ++i)
	{
		ti.i = (int) floor( double(rawHist[i].i) / double(iscale) );
		ti.n = rawHist[i].n;
		ix = (int)scaleHist.size() - 1;
		if (ix>=0)
		{
			if (scaleHist[ix].i==ti.i)
				scaleHist[ix].n += ti.n;
			else
				scaleHist.push_back( ti );
		}
		else
		{
			scaleHist.push_back( ti );
		}
	}
}

void cHistogram::expandHist( intDoubleHist& scaleHist, intDoubleHist& rawHist, int iscale, bool bScale )
{
	int i, j;
	intDouble di, dout;

	scaleHist.clear();

	for (i=0; i<(int)rawHist.size(); ++i)
	{
		di.i = rawHist[i].i;
		di.y = rawHist[i].y;
		if (bScale)
		{
			// re-scaled value, skipping (iscale-1) zeros
			dout.i = (int) floor( double(di.i) * double(iscale) );
			if (dout.i<3) continue;
			dout.y = di.y;
			scaleHist.push_back( dout );
		}
		else
		{
			// re-scale by iscale and hold for next iscale values
			dout.y = di.y / double(iscale);
			for (j=0; j<iscale; ++j)
			{
				dout.i = (int) floor( double(di.i) * double(iscale) ) + j;
				scaleHist.push_back( dout );
			}
		}
	}
}

void cHistogram::copyHist( intHist& outHist, intHist& inHist )
{
	int i;
	twoInt ti;

	outHist.clear();

	for (i=0; i<(int)inHist.size(); ++i)
	{
		ti.i = inHist[i].i;
		ti.n = inHist[i].n;
		outHist.push_back( ti );
	}
}

void cHistogram::copyHist( intDoubleHist& outHist, intDoubleHist& inHist )
{
	int i;
	intDouble ti;

	outHist.clear();

	for (i=0; i<(int)inHist.size(); ++i)
	{
		ti.i = inHist[i].i;
		ti.y = inHist[i].y;
		outHist.push_back( ti );
	}
}

void cHistogram::normalizeHist( intDoubleHist& hist )
{
	int i;
	double sum = 0.;

	for (i=0; i<(int)hist.size(); ++i)
	{
		sum += hist[i].y;
	}

	for (i=0; i<(int)hist.size(); ++i)
	{
		hist[i].y /= sum;
	}
}

intDouble cHistogram::getPeakPt( intDoubleHist& hist )
{
	intDouble id;
	double lfmax = -1.;

	for (int i=0; i<(int)hist.size(); ++i)
	{
		if (hist[i].y>lfmax)
		{
			id = hist[i];
			lfmax = id.y;
		}
	}

	return id;
}

//for ( $i = $minB; $i <= $maxB; ++$i )
//{
//  # starting new window
//  $weight = $data[$i];
//  $left = $i;
//  $right = $i;
//
//  # increase size of window, until at list $K points included
//  while ( $weight < $K )
//  {
//    --$left;
//    ++$right;
//
//    # if maximum possible $weight < $K, stop program
//    if (( $left < $minB )&&( $right > $maxB ))
//      { die "error in setting of smoothing parameter K = $K\n"; }
//
//    # update weight
//    if ( $left >= $minB )
//      { $weight += $data[$left]; }
//
//    if ( $right <= $maxB )
//      { $weight += $data[$right]; }
//  }
//
//  $histo[$i] = $weight/($right - $left + 1);
//}
// smoothHist0 mimics behavior of above algorithm (method /k/) taken from histogram.pl, by A. Lomsadze
//
void cHistogram::smoothHist0( intDoubleHist& outHist, intHist& inHist, int maxLen, double K, double W, string& outBase )
{
	int i, ix;
	int imin = 0;
	int imax = inHist[inHist.size()-1].i;
	if (imax>=maxLen)
		imax = maxLen-1;

	int ixleft, ixright;
	int ileft, iright;
	bool bLeftLimit = false;
	bool bRightLimit = false;
	double wt = 0.;

	int dbg_flag;

	intDouble id;
	outHist.clear();

	bool bPrint = false;
	FILE *pf;
	if (bPrint)
	{
		string out = outBase + "_0_dbg.txt";
		pf = fopen( out.c_str(), "w" );
		if (pf==NULL)
		{
			printf("unable to open debug file for printing.");
			bPrint = false;
		}
	}


	bool bError = false;
	i = 0;
	int iNext;
	int ixNext = inHist[0].i;
	while (ixNext<imin) {
		++i;
		if (i>=(int)inHist.size()) {
			bError = true;
			break;
		}
		iNext = inHist[i].i;
	}
	if (bError) return;

	for (ix=imin; ix<=imax; ++ix)
	{
		// initialize wt
		wt = 0.;

		if (ix==ixNext) {
			// get weight
			wt = double( inHist[i].n );
			// set ptrs to next histogram support point
			++i;
			if (i>=(int)inHist.size()) {
				ixNext = -1;		// no more histogram points
			}
			else {
				ixNext = inHist[i].i;
			}
		}

		ixleft = ixright = ix;
		if (i>1)
		{
			ileft = i-2;			// <--- nearest support point (to left of ix)
			iright = i;				// <--- nearest support point (to right of ix)
		}
		else
		{
			ileft = i-1;			// <--- nearest support point (to left of ix)
			iright = i;				// <--- nearest support point (to right of ix)
		}

		while (wt < K)
		{
			--ixleft;
			++ixright;

			if ((ixright-ixleft+1)>1000)
				dbg_flag = 1;

			if (ixleft>=imin)
			{
				if (ileft>=0) {
					if (ixleft<inHist[ileft].i && ixleft>inHist[0].i)
						dbg_flag = 1;
					if (ixleft==inHist[ileft].i) {
						wt += double(inHist[ileft].n);
						--ileft;
					}
				}
			}
			else { bLeftLimit = true; };

			if (ixright<=imax)
			{
				if (iright<(int)inHist.size()) {
					if (ixright>inHist[iright].i && ixright<inHist[inHist.size()-1].i)
						dbg_flag = 1;
					if (ixright==inHist[iright].i) {
						wt += double(inHist[iright].n);
						++iright;
					}
				}
			}
			else { bRightLimit = true; };

			if (bLeftLimit && bRightLimit) { break; }
		}

		if (bPrint)
			fprintf( pf, "%d  %d  %d  %d  %d  %f  %f  %f\n", ix, i, ixleft, ixright, ixright-ixleft+1, wt, wt / double(ixright-ixleft+1), K );

		// reset limit flags
		bLeftLimit = false;
		bRightLimit = false;

		id.i = ix;
		id.y = wt / double(ixright-ixleft+1);
		outHist.push_back( id );
	}

	if (bPrint)
		fclose(pf);
}

// method /kw/: modeled after histogram.pl, by A. Lomsadze
//
//#->
//if ($verbose) { print "calculating naive histogram\n"; }
//for ( $i = $minB; $i <= $maxB; ++$i )
//{
//  # starting new window
//  $weight = $histo[$i];
//  $left = $i;
//  $right = $i;
//
//  # increase size of window, until width W
//  for ( ; $right - $left + 1 <= $W ; --$left, ++$right)
//  {
//    # if out of range, stop program
//    if (( $left < $minB )&&( $right > $maxB ))
//      { die "error in setting of smoothing parameter W = $W\n"; }
//
//    # update weight
//    if ( $left >= $minB )
//      { $weight += $histo[$left]; }
//
//    if ( $right <= $maxB )
//      { $weight += $histo[$right]; }
//  }
//
//  $naive[$i] = $weight/($right - $left + 1);
//}
//#<-
void cHistogram::smoothHist1( intDoubleHist& outHist, intDoubleHist& inHist, int maxLen, double K, double W )
{
	int i, ix, j;
	int imin = 0;
	int W2 = int( W/2 );
	int imax = inHist[inHist.size()-1].i;
	if (imax>=maxLen)
		imax = maxLen-1;
	int ixleft, ixright;
	int ileft, iright;
	bool bLeftLimit = false;
	bool bRightLimit = false;
	double wt = 0.;

	intDouble id;
	outHist.clear();

	bool bError = false;
	i = 0;
	int iNext;
	int ixNext = inHist[0].i;
	while (ixNext<imin) {
		++i;
		if (i>=(int)inHist.size()) {
			bError = true;
			break;
		}
		iNext = inHist[i].i;
	}
	if (bError) return;

	for (ix=imin; ix<=imax; ++ix)
	{
		// initialize wt
		wt = 0.;

		if (ix==ixNext) {
			// get weight
			wt = inHist[i].y;
			// set ptrs to next histogram support point
			++i;
			if (i>=(int)inHist.size()) {
				ixNext = -1;		// no more histogram points
			}
			else {
				ixNext = inHist[i].i;
			}
		}

		ixleft = ixright = ix;
		ileft = i-1;			// <--- nearest support point (to left of ix)
		iright = i;				// <--- nearest support point (to right of ix)

		for (j=0; j<W2; ++j)		// while (wt < K)
		{
			//if (!bLeftLimit) { --ixleft; }
			//if (!bRightLimit) { ++ixright; }
			--ixleft;
			++ixright;

			if (ixleft>=imin)
			{
				if (ileft>=0) {
					if (ixleft==inHist[ileft].i) {
						wt += inHist[ileft].y;
						--ileft;
					}
				}
			}
			else { bLeftLimit = true; };

			if (ixright<=imax)
			{
				if (iright<(int)inHist.size()) {
					if (ixright==inHist[iright].i) {
						wt += inHist[iright].y;
						++iright;
					}
				}
			}
			else { bRightLimit = true; };

			if (bLeftLimit && bRightLimit) { break; }
		}

		// reset limit flags
		bLeftLimit = false;
		bRightLimit = false;

		id.i = ix;
		id.y = wt / double(ixright-ixleft+1);
		outHist.push_back( id );
	}

}

// smoothing by convolution with rectangular window of width W
//
void cHistogram::smoothHist2( intDoubleHist& outHist, intDoubleHist& inHist, int imin, int maxLen, double K, double W )
{
	int i, j;
	int W2 = int( W/2 );
	int iw = int( W );
	int imax = inHist[inHist.size()-1].i;
	if (imax>=maxLen)
		imax = maxLen-1;
	double wt = 0.;

	intDouble id;
	outHist.clear();

	// prepare data vector
	double *data =  new double[imax];
	int iNext = 0;
	for (i=0; i<imax; ++i)
	{
		data[i] = 0;
		if (i==inHist[iNext].i)
		{
			data[i] = inHist[iNext].y;
			++iNext;
		}
	}

	// prepare kernel
	int iw2 = iw/2;
	double *window = new double[iw];
	for (i=0; i<iw; ++i)
	{
		window[i] = 1./W;
	}

	// convolve
	int ixd;
	bool bInDataRange;
	for (i=imin; i<imax; ++i)
	{
		wt = 0.;
		for (j=0; j<iw; ++j)
		{
			ixd = i + j - iw2;
			bInDataRange = ixd>=imin && ixd<imax;
			if (bInDataRange)
				wt += data[ixd]*window[j];
		}

		id.i = i;
		id.y = wt;
		outHist.push_back( id );
	}

	delete []data;
	delete []window;
}

void cHistogram::getSmoothingParams( double& K, double& W, intHist& cumHist, double std )
{
	int nHist = (int)cumHist.size();
	double lfn = (double) cumHist[nHist-1].n;

	// quartile points
	double lf0p25 = lfn * 0.25;
	double lf0p75 = lfn * 0.75;
	int iq1 = getNearestPoint( cumHist, lf0p25 );
	int iq3 = getNearestPoint( cumHist, lf0p75 );
	int iqr = iq3 - iq1;

	// smoothing parameters
	K = sqrt( lfn );
	W;
	if (std < double(iqr)/1.34)
		W = floor( 0.9*std/pow(lfn,0.2) );
	else
		W = floor( 0.9*double(iqr)/(1.34*pow(lfn,0.2)) );

	if (W==0.)
		W = 3.;

	W *= 2.;
}


int cHistogram::getNearestPoint( intHist& cumHist, double lfval )
{
	if (lfval<=double(cumHist[0].n))
		return 0;

	if (lfval>=double(cumHist[cumHist.size()-1].n))
		return (int)cumHist.size()-1;

	// search for point in cumulative histogram
	int i = 0;
	while (lfval>double(cumHist[i].n))
		++i;

	return cumHist[i].i;
}

void cHistogram::getStats( double& mean, double& std, intHist& hist )
{
	double sum = 0.;
	mean = 0.;
	std = 0.;

	double lfx, lfy;

	double lfytot = double( getCumTotaly(hist) );

	// mean
	for (int i=0; i<(int)hist.size(); ++i) {
		lfy = double( hist[i].n );
		lfx = double( hist[i].i );
		sum += lfx*lfy;
	}

	mean = sum / lfytot;

	// std deviation
	for (int i=0; i<(int)hist.size(); ++i) {
		lfx = double( hist[i].i );
		lfy = double( hist[i].n );
		std += (lfx - mean)*(lfx - mean)*lfy;
	}

	// MMSE
	std = sqrt( std/(lfytot - 1.) );
	// BLUE
	// std =/ lfytot;
}

int cHistogram::getCumTotal( intHist& icumhist, intHist& ihist )
{
	int ival = 0;
	icumhist.clear();
	twoInt ti;

	for (int i=0; i<(int)ihist.size(); ++i) {
		ival += ihist[i].n;
		ti.i = ihist[i].i;
		ti.n = ival;
		icumhist.push_back( ti );
	}

	return ival;
}

int cHistogram::getCumTotalx( intHist& ihist )
{
	int ival = 0;

	for (int i=0; i<(int)ihist.size(); ++i) {
		ival += ihist[i].i;
	}

	return ival;
}

int cHistogram::getCumTotaly( intHist& ihist )
{
	int ival = 0;

	for (int i=0; i<(int)ihist.size(); ++i) {
		ival += ihist[i].n;
	}

	return ival;
}

double cHistogram::getCumTotalx( doubleHist& dhist )
{
	double lfval = 0.;

	for (int i=0; i<(int)dhist.size(); ++i) {
		lfval += dhist[i].x;
	}

	return lfval;
}

void cHistogram::int2double( doubleHist& dhist, intHist& ihist )
{
	// Directly convert int histogram to double histogram

	// clear old data
	dhist.clear();

	DoubleXY dxy;

	for (int i=0; i<(int)ihist.size(); ++i) {
		dxy.x = double( ihist[i].i );
		dxy.y = double( ihist[i].n );
		dhist.push_back( dxy );
	}
}


void cHistogram::doubleInt2intHst( intHist& ihist, intDoubleHist& dhist )
{
	// Directly convert doubleInt histogram to int histogram

	// clear old data
	ihist.clear();

	twoInt ti;

	for (int i=0; i<(int)dhist.size(); ++i) {
		ti.i = dhist[i].i;
		ti.n = int( dhist[i].y );
		ihist.push_back( ti );
	}
}


void cHistogram::int2intDouble( intDoubleHist& dhist, intHist& ihist )
{
	// Directly convert int histogram to int-double histogram

	// clear old data
	dhist.clear();

	intDouble id;

	for (int i=0; i<(int)ihist.size(); ++i) {
		id.i = ihist[i].i;
		id.y = double( ihist[i].n );
		dhist.push_back( id );
	}
}

void cHistogram::updateHistogram( intHist& hist, int inew )
{
	for ( int i=0; i<(int)hist.size(); ++i )
		if (hist[i].i==inew) {
			hist[i].n++;
			return;
		}

	// getting here means this is a new element
	twoInt newElem;
	newElem.i = inew;
	newElem.n = 1;
	hist.push_back( newElem );
}

string cHistogram::printHistogram(string strHeader, intHist& hist, string strSep)
{
	// print out histogram data
	//
	ostringstream out;
	out.setf( ios::fixed, ios::floatfield );
	out.precision( 6 );

	// sort histogram
	sortHist( hist );

	out << strHeader << "\n";

	int nTot = 0;

	for (int i=0; i<(int)hist.size(); ++i)
	{
		nTot += hist[i].n;
		out << hist[i].i << strSep << hist[i].n << strSep << nTot << "\n";
	}
	out << "\n";

	return out.str();
}

string cHistogram::printHistogram(string strHeader, doubleHist& hist, double fBinSize, bool bCumulative)
{
	// print out histogram data
	//
	ostringstream out;
	out.setf( ios::fixed, ios::floatfield );
	out.precision( 6 );

	doubleHist::iterator h_beg_itr;
	doubleHist::iterator h_end_itr;

	DoubleXY dxy;

	// sort the histograms
	double fGlobalMin(numeric_limits<double>::infinity()), fGlobalMax(0);

	sort (hist.begin(), hist.end(), dvLessThan);

	h_beg_itr = hist.begin();
	h_end_itr = hist.end();

	if ((*h_beg_itr).x<fGlobalMin)
		fGlobalMin = (*h_beg_itr).x;

	if ((*(h_end_itr-1)).x>fGlobalMax)
		fGlobalMax = (*(h_end_itr-1)).x;

	// header
	out << strHeader << "\n";
	int ir;		// indices for row (output is a matrix)
	int nBins = int( (fGlobalMax-fGlobalMin)/fBinSize ) + 1;
	double fBinVal, fBinStart;
	
	// keep track of histogram support
	int nNextSample = 0;
	double fyCounts;
	double fxCounts;
	double fxTotals = 1;

	for ( ir=0; ir<nBins; ++ir )
	{
		fBinStart = fGlobalMin + ir*fBinSize;
		// count instances between ir and (ir+iinc)
		getCount( fyCounts, fxCounts, nNextSample, fBinStart, fBinSize, hist );
		fBinVal = fBinStart + fBinSize/2.;

		out << fBinVal << "\t\t";

		if (bCumulative) {
			dxy = hist[nNextSample-1];
			out << dxy.y;
		}
		else {
			out <<  fyCounts;
		}

		if (bCumulative) {
			fxTotals += fxCounts;

			out <<  "\t\t" << fxTotals;
		}

		out << "\n";
	}

	return out.str();
}


void cHistogram::sortHist( intHist& inHist )
{
	sort (inHist.begin(), inHist.end(), intLessThan);
}

void cHistogram::sortHist( double& fGlobalMin, double& fGlobalMax, doubleHist& rawHist )
{
	sort (rawHist.begin(), rawHist.end(), dvLessThan);

	doubleHist::iterator h_beg_itr = rawHist.begin();
	doubleHist::iterator h_end_itr = rawHist.end();

	if ((*h_beg_itr).x<fGlobalMin)
		fGlobalMin = (*h_beg_itr).x;

	if ((*(h_end_itr-1)).x>fGlobalMax)
		fGlobalMax = (*(h_end_itr-1)).x;
}

void cHistogram::getBinHist( doubleHist& binHist, doubleHist& rawHist, int nBins, bool bCumulative )
{
	// sort the histogram and reteurn min and max ordinate values
	double fGlobalMin(numeric_limits<double>::infinity()), fGlobalMax(-numeric_limits<double>::infinity());
	sortHist( fGlobalMin, fGlobalMax, rawHist );

	double fBinSize = (fGlobalMax-fGlobalMin) / double(nBins);

	calcBinHist( binHist, rawHist, fGlobalMin, fBinSize, nBins, bCumulative );
}


void cHistogram::getBinHist( doubleHist& binHist, doubleHist& rawHist, double fBinSize, bool bCumulative )
{
	// sort the histogram and reteurn min and max ordinate values
	double fGlobalMin(numeric_limits<double>::infinity()), fGlobalMax(-numeric_limits<double>::infinity());
	sortHist( fGlobalMin, fGlobalMax, rawHist );

	int nBins = int( (fGlobalMax-fGlobalMin)/fBinSize ) + 1;

	calcBinHist( binHist, rawHist, fGlobalMin, fBinSize, nBins, bCumulative );
}


void cHistogram::getLogBinHist( doubleHist& outHist, doubleHist& rawHist, int nBins, bool bCumulative )
{
	// get log hist
	doubleHist logHist, binHist;
	Log( logHist, rawHist );

	// sort the histogram and reteurn min and max ordinate values
	double fLogGlobalMin(numeric_limits<double>::infinity()), fLogGlobalMax(-numeric_limits<double>::infinity());
	sortHist( fLogGlobalMin, fLogGlobalMax, logHist );

	calcBinHist( binHist, logHist, fLogGlobalMin, fLogGlobalMax, nBins, bCumulative );

	// convert back to linear
	Exp( outHist, binHist );
}


void cHistogram::Log( doubleHist& outHist, doubleHist& inHist )
{
	DoubleXY dxy;
	outHist.clear();

	for (int i=0; i<(int)inHist.size(); ++i)
	{
		dxy.x = log( inHist[i].x );
		dxy.y = inHist[i].y;
		outHist.push_back( dxy );
	}
}

void cHistogram::Exp( doubleHist& outHist, doubleHist& inHist )
{
	DoubleXY dxy;
	outHist.clear();

	for (int i=0; i<(int)inHist.size(); ++i)
	{
		dxy.x = exp( inHist[i].x );
		dxy.y = inHist[i].y;
		outHist.push_back( dxy );
	}
}


void cHistogram::calcBinHist( doubleHist& binHist, doubleHist& rawHist,
							  double fGlobalMin, double fBinSize, int nBins, bool bCumulative )
{
	double fBinVal, fBinStart;

	// keep track of histogram support
	int nNextSample = 0;
	double fyCounts;
	double fxCounts;

	DoubleXY dxy;
	dxy.x = 0.;
	dxy.y = 0.;

	int ir;

	for ( ir=0; ir<nBins; ++ir )
	{
		fBinStart = fGlobalMin + ir*fBinSize;
		// count instances between ir and (ir+iinc)
		getCount( fyCounts, fxCounts, nNextSample, fBinStart, fBinSize, rawHist );
		fBinVal = fBinStart + fBinSize/2.;

		if (bCumulative) {
			dxy.x += fxCounts;
			dxy.y += fyCounts;
		}
		else {
			dxy.x = fxCounts;
			dxy.y = fyCounts;
		}

		binHist.push_back(dxy);
	}
}

//--------------------------------------------------//
void cHistogram::getCount( double& yCount, double& xCount, int& nLastSample,
						  double fStart, double fInc, doubleHist& hist )
{
	// accumulate total number of instances between istart and istart+iinc
	yCount = 0.;
	xCount = 0.;
	if ( nLastSample<(int)hist.size() ) {
		while (hist[nLastSample].x<(fStart+fInc)) {
			yCount += hist[nLastSample].y;
			xCount += hist[nLastSample].x;
			nLastSample++;
			if (nLastSample>=(int)hist.size())
				break;
		}
	}

}

void cHistogram::calcSlope( double& slope, doubleHist& dhist, double xMaxThresh, double yMinThresh )
{
	int i;
	vector<double> px, py;

	// only consider data above threshold
	for (i=0; i<(int)dhist.size(); ++i) {
		if (dhist[i].x<xMaxThresh && dhist[i].y>yMinThresh)
		{
			px.push_back( dhist[i].x );
			py.push_back( dhist[i].y );
		}
	}

	double x[2];
	// get slope (x[0]) and y-intercept (x[1])
	lsFit( x, px, py );
	slope = x[0];
}

void cHistogram::lsFit( double x[2], vector<double>& px, vector<double>& py )
{
	assert( px.size()==py.size() );

	double lfn = double( px.size() );

	// intermediate terms
	double xx = innerProd( px, px );
	double xy = innerProd( py, px );
	double sx = sumVect( px );
	double sy = sumVect( py );
	
	// determinant
	double det = lfn*xx - sx*sx;

	// slope and y-intercept
	double a = lfn*xy - sx*sy;
	double b = -sx*xy + xx*sy;

	x[0] = a / det;
	x[1] = b / det;
}

double cHistogram::sumVect( vector<double>& x )
{
	int n = (int)x.size();
	double out = 0.;

	for (int i=0; i<n; ++i) {
		out += x[i];
	}

	return out;
}

double cHistogram::innerProd( vector<double>& x, vector<double>& y )
{
	assert( x.size()==y.size() );

	int n = (int)x.size();
	double out = 0.;

	for (int i=0; i<n; ++i) {
		out += x[i] * y[i];
	}

	return out;
}