//**************************************************
//  file:     markov.cpp
//  project:  ehmm
//**************************************************
#include <stdio.h>
#include <math.h>
#include <memory.h>

#include "markov.h"
//#include "common.h"

//--------------------------------------------------
bool cMarkovCod::Set( int order, double probN )
{
	bool x;

	x = AbsToFirst( d1a, &d1, order );    if (!x) return false;
	x = AbsToFirst( d2a, &d2, order );    if (!x) return false;
	x = AbsToFirst( d3a, &d3, order );    if (!x) return false;
	
	x = RevFromDir( d1a, &r1a, order );   if (!x) return false;
	x = RevFromDir( d2a, &r2a, order );   if (!x) return false;
	x = RevFromDir( d3a, &r3a, order );   if (!x) return false;

	x = AbsToFirst( r1a, &r1, order );    if (!x) return false;
	x = AbsToFirst( r2a, &r2, order );    if (!x) return false;
	x = AbsToFirst( r3a, &r3, order );    if (!x) return false;


	int n = 4<<(2*order);

	d1a[n] = d2a[n] = d3a[n] = probN;
	r1a[n] = r2a[n] = r3a[n] = probN;

	d1[n] = d2[n] = d3[n] = probN;
	r1[n] = r2[n] = r3[n] = probN;

	++n;
	
	Log( d1a, n ); Log( d2a, n ); Log( d3a, n );
	Log( d1,  n ); Log( d2,  n ); Log( d3,  n );
	Log( r1a, n ); Log( r2a, n ); Log( r3a, n );
	Log( r1,  n ); Log( r2,  n ); Log( r3,  n );

	dph0.mat = d1;
	dph1.mat = d2;
	dph2.mat = d3;

	dph0.matA = d1a;
	dph1.matA = d2a;
	dph2.matA = d3a;

	dph0.next = &dph1;
	dph1.next = &dph2;
	dph2.next = &dph0;

	switch ( order%3 ) {
	case 0: 
		rph0.mat = r1;
		rph1.mat = r2;
		rph2.mat = r3;

		rph0.matA = r1a;
		rph1.matA = r2a;
		rph2.matA = r3a;
		break;
	case 1:
		rph0.mat = r3;
		rph1.mat = r1;
		rph2.mat = r2;

		rph0.matA = r3a;
		rph1.matA = r1a;
		rph2.matA = r2a;
		break;
	case 2:
		rph0.mat = r2;
		rph1.mat = r3;
		rph2.mat = r1;

		rph0.matA = r2a;
		rph1.matA = r3a;
		rph2.matA = r1a;
		break;
	}

	rph0.next = &rph2;
	rph2.next = &rph1;
	rph1.next = &rph0;


	dirArr[0] = &dph0;
	dirArr[1] = &dph1;
	dirArr[2] = &dph2;

	revArr[0] = &rph2;
	revArr[1] = &rph0;
	revArr[2] = &rph1;

	return true;
};

bool cMarkovCod::Reset()
{
	if (d1a != NULL) delete []d1a;
	if (d2a != NULL) delete []d2a;
	if (d3a != NULL) delete []d3a;
	d1a = d2a = d3a = NULL;

	if (r1a != NULL) delete []r1a;
	if (r2a != NULL) delete []r2a;
	if (r3a != NULL) delete []r3a;
	r1a = r2a = r3a = NULL;

	// and so on...

	return false;
}

bool cMarkovCod::MkMarkov(int order, char* buf)
{
	int msize = 4<<(2*order);
	char alphabet[256];
	int i;

	m_order = order;
	m_mkModel_d1 = new double[msize];
	m_mkModel_d2 = new double[msize];
	m_mkModel_d3 = new double[msize];

	if (m_mkModel_d1==NULL || m_mkModel_d2==NULL || m_mkModel_d3==NULL)
		return false;		// error

	for (i=0; i<msize; ++i)
	{
		*(m_mkModel_d1+i)=1;
		*(m_mkModel_d2+i)=1;
		*(m_mkModel_d3+i)=1;
	}
	SetAlphabet( alphabet );

	int n1 = Calc( m_mkModel_d1, buf, alphabet, order, 1 , (int)strlen(buf), false);
	int n2 = Calc( m_mkModel_d2, buf, alphabet, order, 2 , (int)strlen(buf), false);
	int n3 = Calc( m_mkModel_d3, buf, alphabet, order, 3 , (int)strlen(buf), false);

	// normalize
	for (i=0; i<msize; ++i) {
		m_mkModel_d1[i] /= double(n1);
		m_mkModel_d2[i] /= double(n2);
		m_mkModel_d3[i] /= double(n3);
	}

	return true;			// success
}

bool cMarkovCod::MkMarkovNorm( cMarkovCod& markovCod_1, cMarkovCod& markovCod_0 )
{
	int order = 1;
	int msize = 4<<(2*order);
	int i, ix;

	m_order = order;
	m_mkModel_d1 = new double[msize];
	m_mkModel_d2 = new double[msize];
	m_mkModel_d3 = new double[msize];
	if (m_mkModel_d1==NULL || m_mkModel_d2==NULL || m_mkModel_d3==NULL)
		return false;		// error

	for (i=0; i<msize; ++i)
	{
		ix = (int) fmod( floor( double(i)/(pow(4.,double(order))) ), 4. );
		m_mkModel_d1[i] = markovCod_1.m_mkModel_d1[i] / markovCod_0.m_mkModel_d1[ix];
		m_mkModel_d2[i] = markovCod_1.m_mkModel_d2[i] / markovCod_0.m_mkModel_d2[ix];
		m_mkModel_d3[i] = markovCod_1.m_mkModel_d3[i] / markovCod_0.m_mkModel_d3[ix];
	}

	return true;			// success
}

bool cMarkovCod::SetMarkov( cMarkovCod& markovCod_1 )
{
	int order = 1;
	int msize = 4<<(2*order);
	int i;

	m_order = order;
	m_mkModel_d1 = new double[msize];
	m_mkModel_d2 = new double[msize];
	m_mkModel_d3 = new double[msize];
	if (m_mkModel_d1==NULL || m_mkModel_d2==NULL || m_mkModel_d3==NULL)
		return false;		// error

	for (i=0; i<msize; ++i)
	{
		m_mkModel_d1[i] = markovCod_1.m_mkModel_d1[i];
		m_mkModel_d2[i] = markovCod_1.m_mkModel_d2[i];
		m_mkModel_d3[i] = markovCod_1.m_mkModel_d3[i];
	}

	return true;			// success
}

bool cMarkovCod::SetMkFromAbs()
{
	int msize = 4<<(2*m_order);
	int i;

	m_mkModel_d1 = new double[msize];
	m_mkModel_d2 = new double[msize];
	m_mkModel_d3 = new double[msize];
	if (m_mkModel_d1==NULL || m_mkModel_d2==NULL || m_mkModel_d3==NULL)
		return false;		// error

	for (i=0; i<msize; ++i)
	{
		m_mkModel_d1[i] = exp( d1a[i] );
		m_mkModel_d2[i] = exp( d2a[i] );
		m_mkModel_d3[i] = exp( d3a[i] );
	}

	return true;			// success
}

//--------------------------------------------------
bool cMarkovIntr::Set( int order, double probN )
{
	bool x;

	x = AbsToFirst( da, &d, order );    if (!x) return false;
	x = RevFromDir( da, &ra, order );   if (!x) return false;
	x = AbsToFirst( ra, &r, order );    if (!x) return false;

// extension to lower order start
	if ( ddd != 0 )
	{
		printf( "add check point here\n" );
		return false;
	}
	ddd = new double* [ order ];
	rrr = new double* [ order ];
	if ((!ddd )||(!rrr))
	{
		printf( "error, out of memory\n" );
		return false;
	}
	memset( ddd, 0, order*sizeof(double*) );
	memset( rrr, 0, order*sizeof(double*) );

	int i;
	double* curABS = da;

	for ( i = order - 1; i >= 0; --i )
	{
		x = HiToLo( curABS, ddd + i, i + 1 );
		if (!x) return false;
		curABS = ddd[i];
	}

	for ( i = order - 1; i >= 0; --i )
	{
		x = RevFromDir( ddd[i], rrr + i, i );
		if (!x) return false;
	}

	for ( i = order - 1; i >= 0; --i )
	{
		ddd[i][4<<(2*i)] = probN;
		rrr[i][4<<(2*i)] = probN;
		Log(ddd[i], (4<<(2*i)) + 1);
		Log(rrr[i], (4<<(2*i)) + 1);
	}
//stop

	int size = 4<<(2*order);

	da[size] = d[size] = probN;
	ra[size] = r[size] = probN;

	++size;

	Log( da, size ); Log( d, size ); 
	Log( ra, size ); Log( r, size );

	return true;
};

bool cMarkovIntr::MkMarkov(int order, char* buf)
{
	int msize = 4<<(2*order);
	char alphabet[256];
	int i;

	m_order = order;
	m_mkModel_1 = new double[msize];
	m_mkModel_2 = new double[msize];
	m_mkModel_3 = new double[msize];

	if (m_mkModel_1==NULL || m_mkModel_2==NULL || m_mkModel_3==NULL)
		return false;		// error

	for (i=0; i<msize; ++i)
	{
		*(m_mkModel_1+i)=1;
		*(m_mkModel_2+i)=1;
		*(m_mkModel_3+i)=1;
	}
	SetAlphabet( alphabet );

	int n1 = Calc( m_mkModel_1, buf, alphabet, order, 1 , (int)strlen(buf), true);
	int n2 = Calc( m_mkModel_2, buf, alphabet, order, 2 , (int)strlen(buf), true);
	int n3 = Calc( m_mkModel_3, buf, alphabet, order, 3 , (int)strlen(buf), true);

	// normalize
	for (i=0; i<msize; ++i) {
		m_mkModel_1[i] /= double(n1);
		m_mkModel_2[i] /= double(n2);
		m_mkModel_3[i] /= double(n3);
	}

	return true;			// success
}

bool cMarkovIntr::MkMarkovNorm( cMarkovIntr& markovIntr_1, cMarkovIntr& markovIntr_0 )
{
	int order = 1;
	int msize = 4<<(2*order);
	int i, ix;

	m_order = order;
	m_mkModel_1 = new double[msize];
	m_mkModel_2 = new double[msize];
	m_mkModel_3 = new double[msize];
	if (m_mkModel_1==NULL || m_mkModel_2==NULL || m_mkModel_3==NULL)
		return false;		// error

	for (i=0; i<msize; ++i)
	{
		ix = (int) fmod( floor( double(i)/(pow(4.,double(order))) ), 4. );
		m_mkModel_1[i] = markovIntr_1.m_mkModel_1[i] / markovIntr_0.m_mkModel_1[ix];
		m_mkModel_2[i] = markovIntr_1.m_mkModel_2[i] / markovIntr_0.m_mkModel_2[ix];
		m_mkModel_3[i] = markovIntr_1.m_mkModel_3[i] / markovIntr_0.m_mkModel_3[ix];
	}

	return true;			// success
}


//--------------------------------------------------
bool cMarkovNon::Set( int order, double probN )
{
	if ( !AbsToFirst( na, &n, order ) )
		return false;

	int size = 4<<(2*order);

	na[size] = n[size]= probN;

	++size;

	Log( na, size ); Log( n, size ); 

	return true;
};
//--------------------------------------------------

bool cMarkovNon::MkMarkov(int order, char* buf)
{
	int msize = 4<<(2*order);
	char alphabet[256];
	int i;

	m_order = order;
	m_mkModel = new double[msize];

	if (m_mkModel==NULL)
		return false;		// error

    for (i=0; i<msize; ++i)
    {
		*(m_mkModel+i)=1;
	}
	SetAlphabet( alphabet );

	int n1 = Calc( m_mkModel, buf, alphabet, order, 1 , (int)strlen(buf), true);
	int n2 = Calc( m_mkModel, buf, alphabet, order, 2 , (int)strlen(buf), true);
	int n3 = Calc( m_mkModel, buf, alphabet, order, 3 , (int)strlen(buf), true);

	for (i=0; i<msize; ++i) {
		m_mkModel[i] /= double(n3);
	}

	return true;			// success
}

bool cMarkovNon::MkMarkovNorm( cMarkovNon& markovNon_1, cMarkovNon& markovNon_0 )
{
	int order = 1;
	int msize = 4<<(2*order);
	int i, ix;

	m_order = order;
	m_mkModel = new double[msize];
	if (m_mkModel==NULL)
		return false;		// error

	for (i=0; i<msize; ++i)
	{
		ix = (int) fmod( floor( double(i)/(pow(4.,double(order))) ), 4. );
		m_mkModel[i] = markovNon_1.m_mkModel[i] / markovNon_0.m_mkModel[ix];
	}

	return true;			// success
}

bool cMarkovNon::SetMarkov( cMarkovNon& markovNon_1 )
{
	int order = 1;
	int msize = 4<<(2*order);
	int i;

	m_order = order;
	m_mkModel = new double[msize];
	if (m_mkModel==NULL)
		return false;		// error

	for (i=0; i<msize; ++i)
	{
		m_mkModel[i] = markovNon_1.m_mkModel[i];
	}

	return true;			// success
}

bool cMarkovNon::SetMkFromAbs()
{
	int msize = 4<<(2*m_order);
	int i;

	m_mkModel = new double[msize];
	if (m_mkModel==NULL)
		return false;		// error

	for (i=0; i<msize; ++i)
	{
		m_mkModel[i] = exp( na[i] );
	}

	return true;			// success
}
