//**************************************************
//  file:     markov.h
//  project:  ehmm
//**************************************************

#ifndef __MARKOV_H__
#define __MARKOV_H__

#include <stdio.h>
#include "matrixmod.h"
#include <string.h>

//--------------------------------------------------
class cMarkovCod : public cMatrixMod
{
public:
	cMarkovCod()
	{
		d1=d2=d3=d1a=d2a=d3a=r1=r2=r3=r1a=r2a=r3a = NULL;
		m_mkModel_d1 = m_mkModel_d2 = m_mkModel_d3 = NULL;
		m_mkModel_r1 = m_mkModel_r2 = m_mkModel_r3 = NULL;
	};
	~cMarkovCod(){
		clear();
	}

	cMarkovCod( const cMarkovCod& m )
	{
		clear();

		m_order = m.m_order;
		int size = 4<<(2*m_order);

		m_bRelLeft = m.m_bRelLeft;

		if (m.m_mkModel_d1 != NULL) { m_mkModel_d1 = new double[size]; memcpy( m_mkModel_d1, m.m_mkModel_d1, size*sizeof(double) ); }
		if (m.m_mkModel_d2 != NULL) { m_mkModel_d2 = new double[size]; memcpy( m_mkModel_d2, m.m_mkModel_d2, size*sizeof(double) ); }
		if (m.m_mkModel_d3 != NULL) { m_mkModel_d3 = new double[size]; memcpy( m_mkModel_d3, m.m_mkModel_d3, size*sizeof(double) ); }

		if (m.d1a != NULL) { d1a = new double[size+1]; memcpy( d1a, m.d1a, (size+1)*sizeof(double) ); }
		if (m.d2a != NULL) { d2a = new double[size+1]; memcpy( d2a, m.d2a, (size+1)*sizeof(double) ); }
		if (m.d3a != NULL) { d3a = new double[size+1]; memcpy( d3a, m.d3a, (size+1)*sizeof(double) ); }

		if (m.d1 != NULL) { d1 = new double[size+1]; memcpy( d1, m.d1, (size+1)*sizeof(double) ); }
		if (m.d2 != NULL) { d2 = new double[size+1]; memcpy( d2, m.d2, (size+1)*sizeof(double) ); }
		if (m.d3 != NULL) { d3 = new double[size+1]; memcpy( d3, m.d3, (size+1)*sizeof(double) ); }

		if (m.r1a != NULL) { r1a = new double[size+1]; memcpy( r1a, m.r1a, (size+1)*sizeof(double) ); }
		if (m.r2a != NULL) { r2a = new double[size+1]; memcpy( r2a, m.r2a, (size+1)*sizeof(double) ); }
		if (m.r3a != NULL) { r3a = new double[size+1]; memcpy( r3a, m.r3a, (size+1)*sizeof(double) ); }

		if (m.r1 != NULL) { r1 = new double[size+1]; memcpy( r1, m.r1, (size+1)*sizeof(double) ); }
		if (m.r2 != NULL) { r2 = new double[size+1]; memcpy( r2, m.r2, (size+1)*sizeof(double) ); }
		if (m.r3 != NULL) { r3 = new double[size+1]; memcpy( r3, m.r3, (size+1)*sizeof(double) ); }
	}
	void operator=( const cMarkovCod& m )
	{
		clear();

		m_order = m.m_order;
		int size = 4<<(2*m_order);

		m_bRelLeft = m.m_bRelLeft;

		if (m.m_mkModel_d1 != NULL) { m_mkModel_d1 = new double[size]; memcpy( m_mkModel_d1, m.m_mkModel_d1, size*sizeof(double) ); }
		if (m.m_mkModel_d2 != NULL) { m_mkModel_d2 = new double[size]; memcpy( m_mkModel_d2, m.m_mkModel_d2, size*sizeof(double) ); }
		if (m.m_mkModel_d3 != NULL) { m_mkModel_d3 = new double[size]; memcpy( m_mkModel_d3, m.m_mkModel_d3, size*sizeof(double) ); }

		if (m.d1a != NULL) { d1a = new double[size+1]; memcpy( d1a, m.d1a, (size+1)*sizeof(double) ); }
		if (m.d2a != NULL) { d2a = new double[size+1]; memcpy( d2a, m.d2a, (size+1)*sizeof(double) ); }
		if (m.d3a != NULL) { d3a = new double[size+1]; memcpy( d3a, m.d3a, (size+1)*sizeof(double) ); }

		if (m.d1 != NULL) { d1 = new double[size+1]; memcpy( d1, m.d1, (size+1)*sizeof(double) ); }
		if (m.d2 != NULL) { d2 = new double[size+1]; memcpy( d2, m.d2, (size+1)*sizeof(double) ); }
		if (m.d3 != NULL) { d3 = new double[size+1]; memcpy( d3, m.d3, (size+1)*sizeof(double) ); }

		if (m.r1a != NULL) { r1a = new double[size+1]; memcpy( r1a, m.r1a, (size+1)*sizeof(double) ); }
		if (m.r2a != NULL) { r2a = new double[size+1]; memcpy( r2a, m.r2a, (size+1)*sizeof(double) ); }
		if (m.r3a != NULL) { r3a = new double[size+1]; memcpy( r3a, m.r3a, (size+1)*sizeof(double) ); }

		if (m.r1 != NULL) { r1 = new double[size+1]; memcpy( r1, m.r1, (size+1)*sizeof(double) ); }
		if (m.r2 != NULL) { r2 = new double[size+1]; memcpy( r2, m.r2, (size+1)*sizeof(double) ); }
		if (m.r3 != NULL) { r3 = new double[size+1]; memcpy( r3, m.r3, (size+1)*sizeof(double) ); }

		//return *this;
	}

	void clear()
	{
		if (m_mkModel_d1!=NULL) { delete []m_mkModel_d1; m_mkModel_d1 = NULL; }
		if (m_mkModel_d2!=NULL) { delete []m_mkModel_d2; m_mkModel_d2 = NULL; }
		if (m_mkModel_d3!=NULL) { delete []m_mkModel_d3; m_mkModel_d3 = NULL; }
		if (d1a!=NULL) { delete []d1a; d1a = NULL; }
		if (d2a!=NULL) { delete []d2a; d2a = NULL; }
		if (d3a!=NULL) { delete []d3a; d3a = NULL; }
		if (d1!=NULL) { delete []d1; d1 = NULL; }
		if (d2!=NULL) { delete []d2; d2 = NULL; }
		if (d3!=NULL) { delete []d3; d3 = NULL; }
		if (r1a!=NULL) { delete []r1a; r1a = NULL; }
		if (r2a!=NULL) { delete []r2a; r2a = NULL; }
		if (r3a!=NULL) { delete []r3a; r3a = NULL; }
		if (r1!=NULL) { delete []r1; r1 = NULL; }
		if (r2!=NULL) { delete []r2; r2 = NULL; }
		if (r3!=NULL) { delete []r3; r3 = NULL; }
	}

	double* d1a; double* d1;
	double* d2a; double* d2;
	double* d3a; double* d3;

	double* r1a; double* r1;
	double* r2a; double* r2;
	double* r3a; double* r3;

	int m_order;
	double *m_mkModel_d1, *m_mkModel_d2, *m_mkModel_d3;
	double *m_mkModel_r1, *m_mkModel_r2, *m_mkModel_r3;

	bool m_bRelLeft;		// true if relative probabilities go from right to left

	bool Set( int order, double probN );
	bool Reset(void);

	void ResetDir( int ph )
	{
		dph = dirArr[ph];
	}

	void ResetRev( int ph )
	{
		rph = revArr[ph];
	}

	double GetDirA( int adr )
	{
		double x = dph->matA[ adr ];
		dph = dph->next;
		return x;
	}

	double GetDir( int adr )
	{
		double x = dph->mat[ adr ];
		dph = dph->next;
		return x;
	}

	double GetRevA( int adr )
	{
		double x = rph->matA[ adr ];
		rph = rph->next;
		return x;
	}

	double GetRev( int adr )
	{
		double x = rph->mat[ adr ];
		rph = rph->next;
		return x;
	}

	bool MkMarkov(int order, char* buf);
	bool MkMarkovNorm( cMarkovCod& markovCod_1, cMarkovCod& markovCod_0 );
	bool SetMarkov( cMarkovCod& markovCod_1 );
	bool SetMkFromAbs(void);

private:

	struct Phase {
		double* mat;
		double* matA;
		Phase* next;
	} dph0, dph1, dph2, *dph, rph0, rph1, rph2, *rph;

	Phase* dirArr[3];
	Phase* revArr[3];
};
//--------------------------------------------------
class cMarkovNon : public cMatrixMod
{
public:

	cMarkovNon() { m_mkModel = n = na = NULL; m_order = 0; };
	~cMarkovNon() { delete []m_mkModel; delete []n; delete []na; };
	cMarkovNon( const cMarkovNon& m )
	{
		clear();

		m_order = m.m_order;
		int size = 4<<(2*m_order);

		if (m.m_mkModel != NULL) { m_mkModel = new double[size]; memcpy( m_mkModel, m.m_mkModel, size*sizeof(double) ); }

		if (m.na != NULL) { na = new double[size+1]; memcpy( na, m.na, (size+1)*sizeof(double) ); }
		if (m.n != NULL) { n = new double[size+1]; memcpy( n, m.n, (size+1)*sizeof(double) ); }
	}
	void operator=( const cMarkovNon& m )
	{
		clear();

		m_order = m.m_order;
		int size = 4<<(2*m_order);

		if (m.m_mkModel != NULL) { m_mkModel = new double[size]; memcpy( m_mkModel, m.m_mkModel, size*sizeof(double) ); }

		if (m.na != NULL) { na = new double[size+1]; memcpy( na, m.na, (size+1)*sizeof(double) ); }
		if (m.n != NULL) { n = new double[size+1]; memcpy( n, m.n, (size+1)*sizeof(double) ); }

		//return *this;
	}

	void clear()
	{
		if (m_mkModel!=NULL) { delete []m_mkModel; m_mkModel = NULL; }
		if (na!=NULL) { delete []na; na = NULL; }
		if (n!=NULL) { delete []n; n = NULL; }
	}
	double* na;	double* n;
	double *m_mkModel;
	
	int m_order;

	bool Set( int order, double probN );
	bool MkMarkov(int order, char* buf);
	bool MkMarkovNorm( cMarkovNon& markovNon_1, cMarkovNon& markovNon_0 );
	bool SetMarkov( cMarkovNon& markovNon_1 );
	bool SetMkFromAbs(void);

};
//--------------------------------------------------
class cMarkovIntr : private cMatrixMod
{
public:

	cMarkovIntr()
	{
		d=da=r=ra = 0; ddd=rrr = 0;
		m_mkModel_1 = m_mkModel_2 = m_mkModel_3 = NULL;
	};
	~cMarkovIntr()
	{
		delete [] d; delete [] da; delete [] r; delete [] ra; delete [] ddd; delete [] rrr;
		if (m_mkModel_1!=NULL) delete []m_mkModel_1;
		if (m_mkModel_2!=NULL) delete []m_mkModel_2;
		if (m_mkModel_3!=NULL) delete []m_mkModel_3;
	};
	cMarkovIntr( cMarkovIntr& m )
	{
		if (m_mkModel_1!=NULL) { delete []m_mkModel_1; }
		if (m_mkModel_2!=NULL) { delete []m_mkModel_2; }
		if (m_mkModel_3!=NULL) { delete []m_mkModel_3; }

		m_order = m.m_order;
		int size = 4<<(2*m_order);
		m_mkModel_1 = new double[size]; memcpy( m_mkModel_1, m.m_mkModel_1, size*sizeof(double) );
		m_mkModel_2 = new double[size]; memcpy( m_mkModel_2, m.m_mkModel_2, size*sizeof(double) );
		m_mkModel_3 = new double[size]; memcpy( m_mkModel_3, m.m_mkModel_3, size*sizeof(double) );
	}
	cMarkovIntr operator=( cMarkovIntr& m )
	{
		if (m_mkModel_1!=NULL) { delete []m_mkModel_1; }
		if (m_mkModel_2!=NULL) { delete []m_mkModel_2; }
		if (m_mkModel_3!=NULL) { delete []m_mkModel_3; }

		m_order = m.m_order;
		int size = 4<<(2*m_order);
		m_mkModel_1 = new double[size]; memcpy( m_mkModel_1, m.m_mkModel_1, size*sizeof(double) );
		m_mkModel_2 = new double[size]; memcpy( m_mkModel_2, m.m_mkModel_2, size*sizeof(double) );
		m_mkModel_3 = new double[size]; memcpy( m_mkModel_3, m.m_mkModel_3, size*sizeof(double) );

		return *this;
	}
	void clear()
	{
		if (m_mkModel_1!=NULL) { delete []m_mkModel_1; m_mkModel_1 = NULL; }
		if (m_mkModel_2!=NULL) { delete []m_mkModel_2; m_mkModel_2 = NULL; }
		if (m_mkModel_3!=NULL) { delete []m_mkModel_3; m_mkModel_3 = NULL; }

		if (d!=NULL) { delete []d; d = NULL; }
		if (da!=NULL) { delete []da; da = NULL; }
		if (r!=NULL) { delete []r; r = NULL; }
		if (ra!=NULL) { delete []ra; ra = NULL; }
		if (ddd!=NULL) { delete []ddd; ddd = NULL; }
		if (rrr!=NULL) { delete []rrr; rrr = NULL; }
	}

	double* da;	double* d;

	double* ra;	double* r;

	int m_order;
	double *m_mkModel_1, *m_mkModel_2, *m_mkModel_3;

	bool Set( int order, double probN );
	bool MkMarkov(int order, char* buf);
	bool MkMarkovNorm( cMarkovIntr& markovIntr_1, cMarkovIntr& markovIntr_0 );

	double** ddd;
	double** rrr;

};
//--------------------------------------------------

#endif  // __MARKOV_H__

