
#pragma once

#include <string>
#include <vector>
#include "markov.h"
#include "site.h"
#include "filebuf.h"
#include "histogram.h"

#include "defines.h"

//#include "matrix_defs.h"

using std::vector;
using std::string;

// IC = isoform change
#define IC_NOTHING			0x0000
#define IC_CDS_START		0x0001
#define IC_DONOR			0x0002
#define IC_ACCEPTOR			0x0003
#define IC_SKIPPED_EXON		0x0004
#define IC_INSERTED_EXON	0x0005
#define IC_RETAINED_INTRON	0x0006

typedef struct tagIsoformType
{
	int nTypes[8];
} isoformType;

class cModel
{
public:

	char m_alphabet[256];

	cSite					*m_pSite;
	vector<codingRegion>	m_genes;
	vector<codingRegion>	m_filteredGenes;

	int m_nIter;
	string m_gmRunIxDir;
	vector<string> m_intronInfoFiles;

	bool m_bVerbose;		// continually print status to stdout
	bool m_bVeryVerbose;	// continually print even more status to stdout
	bool m_bInteractive;

	bool m_bDebug;
	string m_debugOutDir;

	bool m_bExtr;

	// site data (accumulations of individual cSite data
	vector<string> m_seqStarts, m_seqTerms, m_seqDons, m_seqAccs, m_seqShortAccs, m_seqIntronTerms;
	vector<bpRecord> m_bpData;
	vector<int> m_splicePhase;
	string m_strCoding;
	string m_strIntrons;
	string m_strNonCoding;

	int m_order;

	int m_gmsModelOrder;

	cMarkovCod		m_markovCod_0,	m_markovCod_1, m_markovCod_2, m_markovCod_3, m_markovCod_4, m_markovCod_5;
	cMarkovCod		m_markovCod_1n;
	cMarkovNon		m_markovNon_0,	m_markovNon_1, m_markovNon_2, m_markovNon_3, m_markovNon_4, m_markovNon_5;
	cMarkovNon		m_markovNon_1n;
	cMarkovIntr		m_markovIntr_0,	m_markovIntr_1, m_markovIntr_5;

	cMarkovCod		m_markovCodGms;
	cMarkovNon		m_markovNonGms;

	cMarkovCod		m_markovCodon;

	cMarkovNon		m_markovBpSpacer;

	// model controls
	bool	m_b3phIntron, m_b3phIntronNext;
	bool	m_bLenDistr, m_bLenDistrNext;
	bool	m_bExonTransition, m_bExonTransitionNext;
	bool	m_bBranchPoint, m_bBranchPointNext;
	int		m_orderOneCutoff;
	bool	m_bSeparateIntronModel_5;
	bool	m_bEsPlusIntronLD;
	bool	m_bMarkovOnly;

	int m_intronPeakMargin;
	int m_lenBranchMotif;

	// model parameters
	int m_iorderAtg, m_iorderTaa, m_iorderTag, m_iorderTga;
	int m_iorderDon0, m_iorderDon1, m_iorderDon2;
	int m_iorderAcc0, m_iorderAcc1, m_iorderAcc2;
	// BP models
	int m_iorderShortAcc0, m_iorderShortAcc1, m_iorderShortAcc2;
	int m_iorderBp;

	double *m_pModelMatAtg, *m_pModelMatTaa, *m_pModelMatTag, *m_pModelMatTga;
	double *m_pModelMatDon0, *m_pModelMatDon1, *m_pModelMatDon2;
	double *m_pModelMatAcc0, *m_pModelMatAcc1, *m_pModelMatAcc2;

	// BP models
	double *m_pModelMatShortAcc0, *m_pModelMatShortAcc1, *m_pModelMatShortAcc2;
	double *m_pModelMatBp;

	// for reports
	double *m_pDonMotif, *m_pAccMotif;
	double *m_pATGMotif, *m_pTAAMotif, *m_pTAGMotif, *m_pTGAMotif;
	double *m_pShortAccMotif, *m_pBpMotif;

	int m_startFlankLeft, m_startFlankRight;
	int m_donFlankLeft,	m_donFlankRight;
	int m_accFlankLeft, m_accFlankRight;
	int m_termFlankLeft, m_termFlankRight;
	int m_shortAccFlankLeft, m_shortAccFlankRight;
	int m_intronTermLen;
	int m_intronTermFlankRight;

	// length distr params
	int m_lenMaxSingleExon, m_lenMinSingleExon, m_lenMaxExon, m_lenMinExon;
	int m_lenMaxIniExon, m_lenMinIniExon, m_lenMaxTerExon, m_lenMinTerExon;
	int m_lenMaxIntron, m_lenMinIntron, m_lenMaxInterg, m_lenMinInterg;
	double *m_pSingleExon, *m_pExon, *m_pIniExon, *m_pTerExon, *m_pIntron, *m_pInterg;

	vector<string> taa, tag, tga, other_stops;

	double m_probPhase0, m_probPhase1, m_probPhase2;
	int m_nSingleExon, m_nMultiExon;
	double m_lfToSingle;

	// intron-exon transition probability
	double m_probToTerminalExon;
	double m_probToInternalExon;

	// length distributions
	intDoubleHist m_distrExonSingLen, m_distrExonIniLen, m_distrExonTermLen;
	intDoubleHist m_distrExonIntLen;
	intDoubleHist m_distrIntronLen;
	intDoubleHist m_distrIntergenicLen;
	intDoubleHist m_distrBpToAcc, m_distrDonToBp;

	// histogram
	cHistogram m_hist;

	// arrays containing file names, for post processing
	string m_startSiteFile;
	string m_stopSiteFile;
	string m_donSiteFile;
	string m_accSiteFile;
	string m_codingFile;
	string m_intronFile;
	string m_flankFile;
	string m_nonCodingFile;
	string m_phaseFile;
	string m_shortAccFile;
	string m_bpDataFile;
	string m_intronFastaFile;

	// ctor/dtor
	cModel()
	{
		m_pModelMatAtg = m_pModelMatTaa = m_pModelMatTag = m_pModelMatTga = NULL;
		m_pModelMatDon0 = m_pModelMatDon1 = m_pModelMatDon2 = NULL;
		m_pModelMatAcc0 = m_pModelMatAcc1 = m_pModelMatAcc2 = NULL;
		m_pModelMatBp = m_pBpMotif = NULL;
		m_orderOneCutoff = 2000;

		m_intronPeakMargin = 10;
		m_lenBranchMotif = 9;

		m_bDebug = false;

		m_bEsPlusIntronLD = false;
		m_bMarkovOnly = false;

		cMatrixMod tmpMat;
		tmpMat.SetAlphabet( m_alphabet );

		// defaults
		m_lenMaxSingleExon = 10000;
		m_lenMaxExon = 10000;
		m_lenMaxIniExon = 10000;
		m_lenMaxTerExon = 10000;
		m_lenMaxIntron = 10000;
		m_lenMaxInterg = 30000;
	};
	~cModel() {  };

	cModel(const cModel& m)
	{
		setAlphabet( m_alphabet );

		m_lenMaxSingleExon = m.m_lenMaxSingleExon;
		m_lenMaxExon = m.m_lenMaxExon;
		m_lenMaxIniExon = m.m_lenMaxIniExon;
		m_lenMaxTerExon = m.m_lenMaxTerExon;
		m_lenMaxIntron = m.m_lenMaxIntron;
		m_lenMaxInterg = m.m_lenMaxInterg;

		m_distrIntronLen = m.m_distrIntronLen;

		m_markovCod_0 = m.m_markovCod_0;
		m_markovCod_1 = m.m_markovCod_1;
		m_markovCod_2 = m.m_markovCod_2;
		m_markovCod_3 = m.m_markovCod_3;
		m_markovCod_4 = m.m_markovCod_4;
		m_markovCod_5 = m.m_markovCod_5;

		m_markovNon_0 = m.m_markovNon_0;
		m_markovNon_1 = m.m_markovNon_1;
		m_markovNon_2 = m.m_markovNon_2;
		m_markovNon_3 = m.m_markovNon_3;
		m_markovNon_4 = m.m_markovNon_4;
		m_markovNon_5 = m.m_markovNon_5;

		m_pModelMatDon0 = m_pModelMatDon1 = m_pModelMatDon2 = NULL;
		m_pModelMatAcc0 = m_pModelMatAcc1 = m_pModelMatAcc2 = NULL;

		copyDon( m );
		copyAcc( m );

		m_order = m.m_order;
	}
	void operator=(const cModel& m)
	{
		setAlphabet( m_alphabet );

		m_lenMaxSingleExon = m.m_lenMaxSingleExon;
		m_lenMaxExon = m.m_lenMaxExon;
		m_lenMaxIniExon = m.m_lenMaxIniExon;
		m_lenMaxTerExon = m.m_lenMaxTerExon;
		m_lenMaxIntron = m.m_lenMaxIntron;
		m_lenMaxInterg = m.m_lenMaxInterg;

		m_distrIntronLen = m.m_distrIntronLen;

		m_markovCod_0 = m.m_markovCod_0;
		m_markovCod_1 = m.m_markovCod_1;
		m_markovCod_2 = m.m_markovCod_2;
		m_markovCod_3 = m.m_markovCod_3;
		m_markovCod_4 = m.m_markovCod_4;
		m_markovCod_5 = m.m_markovCod_5;

		m_markovNon_0 = m.m_markovNon_0;
		m_markovNon_1 = m.m_markovNon_1;
		m_markovNon_2 = m.m_markovNon_2;
		m_markovNon_3 = m.m_markovNon_3;
		m_markovNon_4 = m.m_markovNon_4;
		m_markovNon_5 = m.m_markovNon_5;

		copyDon( m );
		copyAcc( m );

		m_order = m.m_order;
	}

	void clear(void)
	{
		m_markovCod_0.clear();
		m_markovCod_1.clear();
		m_markovCod_2.clear();
		m_markovCod_3.clear();
		m_markovCod_4.clear();
		m_markovCod_5.clear();

		m_markovNon_5.clear();
	}

	void setAlphabet( char* ptr )
	{
		memset( ptr, 4, 256 );

		*(ptr + 'a' ) = 0;
		*(ptr + 'A' ) = 0;
		*(ptr + 'c' ) = 1;
		*(ptr + 'C' ) = 1;
		*(ptr + 'g' ) = 2;
		*(ptr + 'G' ) = 2;
		*(ptr + 't' ) = 3;
		*(ptr + 'T' ) = 3;
		*(ptr + 'u' ) = 3;
		*(ptr + 'U' ) = 3;
	}

	void	init(cSite *psite);
	void	getModelParams(void);

	void	setVerbose(bool);
	void	setFileNames(void);

	bool	siteSite(double **pval, int order, int width, double *pData);
	bool	copyDon(const cModel& model);
	bool	copyAcc(const cModel& model);

	// site model manipulation
	bool calcSiteModel( double **pModelMat, vector<string>& site, int& iorder, int nCols, int flankLeft, int flankRight );
	bool calcSiteModel0( double **ppModelMat, vector<string>& site, int nCols, int flankLeft, int flankRight );
	bool normalizeModel( void *func, double *pModelMat, int iorder, int nCols,
						 cMarkovCod *pCod0, cMarkovNon *pNon0, cMarkovCod *pCod1, cMarkovNon *pNon1 );

	void createHeurSiteModels();

	bool normalizeModelDon		( double *pMatOut, double *pModelMatDon, int iphase, int iorderDon, cMarkovCod *pmarkovCod, cMarkovNon *pmarkovNon );
	bool normalizeModelAcc		( double *pMatOut, double *pModelMatAcc, int iphase, int iorderAcc, cMarkovCod *pmarkovCod, cMarkovNon *pmarkovNon );
	bool unNormalizeModelDon( double *pMatOut, double *pModelMatDon, int iphase, int iorderDon, cMarkovCod *pmarkovCod, cMarkovNon *pmarkovNon );
	bool unNormalizeModelAcc( double *pMatOut, double *pModelMatAcc, int iphase, int iorderAcc, cMarkovCod *pmarkovCod, cMarkovNon *pmarkovNon );

	void normalizeDonAcc(void);
	void unNormalizeDonAcc(void);

	bool normalizeModelNon( double *pMatOut, double *pModelMat, int ibegin, int iend, int iorder, cMarkovNon *pmarkovNon );
	bool normalizeModelCod( double *pMatOut, double *pModelMat, int ibegin, int iend, int iphase, int iorder, cMarkovCod *pmarkovCod );

	bool unNormalizeModelCod( double *pMatOut, double *pModelMat, int ibegin, int iend, int iphase, int iorder, cMarkovCod *pmarkovCod );
	bool unNormalizeModelNon( double *pMatOut, double *pModelMat, int ibegin, int iend, int iorder, cMarkovNon *pmarkovNon );

	bool deriveMarkovCod_01(void);
	bool deriveMarkovNon_01(void);

	// HMM-ES file and state mgmt
	bool createMkModel(void);
	bool createMkModel(char *pchCoding);
	void appendGenes( vector<codingRegion>& genes );
	void clearData(void);

	bool calcLenHistograms(bool bCalcHist, bool bFilter);

	bool printStopSites(string& fnameBase);
	void writeStringArray( FILE *pF, vector<string>& strArr );
	string strInt2Str( string, int );

	void appendSiteData(void);

	// debug and diagnostics
	void report(string str);		// verbose status monitoring
	void setDebugDir( string );
	void scoreMarkovModels(void);
	void scoreMarkovModelsTrans(void);
	bool scoreMarkovCod5( double& score, int iframe, const char* buf, char* alphabet );
	bool scoreMarkovCod0( double& score, int iframe, const char* buf, char* alphabet );
	bool scoreMarkov( double& score, double *mat, int order, const char* buf, char* alphabet );
	bool scoreMarkov( double& score, double *mat, double *matn, int order, const char* buf, char* alphabet );
	bool scoreMarkovBack( double& score, int order, const char* buf, char* alphabet );
	//bool scoreMarkov3( double& score, double *mat, int order, int fr, const char* buf, char* alphabet );
	bool scoreMarkov3( double& score, int order, int fr, const char* buf, char* alphabet );
	bool scoreMarkov3_Left( double& score, cMarkovCod& mk, int order, int fr, const char* buf, char* alphabet );
	bool scoreMarkov3_Right( double& score, cMarkovCod& mk, int order, int fr, const char* buf, char* alphabet );
	bool scoreMarkov3( double& score, double *mat1, double *mat2, double *mat3, int order, int fr, const char* buf, char* alphabet );
	bool scoreMarkovNon_Left( double& score, cMarkovNon& mk, int order, const char* buf, char* alphabet );
	bool scoreMarkovNon_Right( double& score, cMarkovNon& mk, int order, const char* buf, char* alphabet );
	bool scoreMarkovNon( double& score, int order, const char* buf, char* alphabet );
	//bool scoreMarkovNon( double& score, cMarkovNon& mk, int order, const char* buf, char* alphabet );
	bool getAddr( vector<int>& adr, const char *buf, int len, int order, char* alphabet );

	void scoreSiteModels(void);
	void scoreSiteModelsTrans(void);
	bool scoreSiteModel( double *pModel, int iorder, vector<string>& sites, vector<double>& scores );
	bool scoreSite( double& score, double *mat, int iorder, int iwidth, const char* buf, char* alphabet );
	bool scoreSiteLin( double& score, double *mat, int order, int iwidth, const char* buf, char* alphabet );
	bool scoreLengths( vector<double>& lenScores, vector<int>& nLengths, string& modelName );
	void getLengthScore( double& score, int nLength, double *pLd, int min, int max );

	void scoreIntronAlignments(string& fname, bool bMark);
	void scoreIntronFlanks(string& fname, bool bMark);
	void checkFrame( int& fr5, int& fr3, double& logScore0_5, double& logScore1_5, double& logScore2_5, double& logScore0_3, double& logScore1_3, double& logScore2_3 );
	void ReorderLeftL( double* mout, double* min, int orderIn, int orderOut );
	void ReorderRightL( double* mout, double* min, int orderIn, int orderOut );
	void ReorderLeft( double* mout, double* min, int orderIn, int orderOut );
	void ReorderRight( double* mout, double* min, int orderIn, int orderOut );

	void scoreLengthModels(vector<sTranscript>& transcripts);
	void scoreLengthModels(vector<codingRegion>& genes);
	void scoreLength( vector<double>& lenScores, vector<int>& nLengths, double *pLd, int min, int max );
	bool scoreDonSites( vector<double>& donScores, vector<string>& donSite, vector<int>& donFrame );
	bool scoreAccSites( vector<double>& accScores, vector<string>& accSite, vector<int>& accFrame );
	// score all three frames
	bool scoreDonSite3( double& logScore0, double& logScore1, double& logScore2, string& donSite, int& iJcnType );
	bool scoreAccSite3( double& logScore0, double& logScore1, double& logScore2, string& accSite, int& iJcnType );
	bool scoreAccSite3( vector<double>& accScores1, vector<double>& accScores2, vector<double>& accScores3, vector<string>& accSite, vector<int>& iJcnTypes );
	bool scoreDonSites3( vector<double>& donScores1, vector<double>& donScores2, vector<double>& donScores3, vector<string>& donSite, vector<int>& iJcnTypes );
	bool scoreAccSites3( vector<double>& accScores1, vector<double>& accScores2, vector<double>& accScores3, vector<string>& accSite, vector<int>& iJcnTypes );
	bool scoreDonSitesLin3( vector<double>& donScores1, vector<double>& donScores2, vector<double>& donScores3, vector<string>& donSite, vector<int>& iJcnTypes );
	bool scoreAccSitesLin3( vector<double>& accScores1, vector<double>& accScores2, vector<double>& accScores3, vector<string>& accSite, vector<int>& iJcnTypes );

	void splitSeq( vector<string>& seqs, vector<string>& mrnaIds, vector<int>& iMark_5s, vector<int>& iMark_3s, vector<int>& starts, vector<int>& stops, vector<char>& strands, char *buf );
	void splitSeq( vector<string>& seqs, vector<string>& idseq, vector<string>& ids, vector<int>& ibegs, vector<int>& iends, vector<int>& iJcnTypes, vector<char>& strands, char *buf );
	void splitSeq( vector<string>& seqs, vector<string>& idseq, vector<string>& ids, vector<int>& ibegs, vector<int>& iends, vector<int>& iJcnTypes,
		vector<char>& strands, vector<int>& iFrames, vector<int>& iMark_5s, vector<int>& iMark_3s, vector<int>& iMark_Is,
		vector<int>& nIntrons, vector<int>& indxIntron, vector<int>& dxAtg, vector<int>& dxStp, char *buf );
	void splitSeq( vector<string>& seqs5, vector<string>& seqs3, char *buf );

	// load/save model data
	bool load( string& fname );
	bool load_2( string& fname );
	bool log2linear( double *pData, int size );
	bool log2linear( double *pData, int order, int width );
	bool rel2abs( double *pData, int order, int width );
	void invNormLog( double* ptr, int size );
	bool ln2hist( intDoubleHist& hist, double *pln, int min, int max );
};

