
#include "site.h"

void cSite::report(string str)
{
	if (m_bVerbose)
		printf("%s\n",str.c_str());

	//fprintf( *m_ppfLog, "%s\n",str.c_str() );
}

void cSite::initParams()
{
	// site definitions
	m_startFlankLeft = 6;
	m_startFlankRight = 3;

	m_donFlankLeft = 3;
	m_donFlankRight = 4;

	m_accFlankLeft = 18;
	m_accFlankRight = 1;

	m_shortAccFlankLeft = 2;
	m_shortAccFlankRight = 1;

	m_intronTermLen = 400;			// default (max), should downselect a smaller window for gibbs sampling
	m_intronTermFlankRight = 4;		// value used in old GM-ES
	m_intronTermThresh = 10;		// ignore all terminal intron sequences shorter than this

	m_termFlankLeft = 0;
	m_termFlankRight = 9;

	m_intronMarginLeft = 20;
	m_intronMarginRight = 27;

	m_minIntronLen = 20;
	m_minIntergLen = 10;

	if (m_bPred)
		m_minCodingLen = 0;
	else
		m_minCodingLen = 0;

	// clear markov model buffers
	m_pModelMatAtg = NULL;
	m_pModelMatTaa = NULL;
	m_pModelMatTag = NULL;
	m_pModelMatTga = NULL;
	m_pModelMatDon0 = NULL;
	m_pModelMatDon1 = NULL;
	m_pModelMatDon2 = NULL;
	m_pModelMatAcc0 = NULL;
	m_pModelMatAcc1 = NULL;
	m_pModelMatAcc2 = NULL;
}

void cSite::init( vector<codingRegion>& genes, SSeq *psseq, int iContig  )
{
	// clear data
	clearData();

	// copy genes to local
	for (int i=0; i<(int)genes.size(); ++i)
	{
		if (genes[i].nContig==iContig)
			m_genes.push_back( genes[i] );
	}

	// get ptr to SSeq
	m_psseq = psseq;

	initParams();
}

//typedef struct tagIntron
//{
//	string strId;
//	int nContig;
//	int nIntronsInGene;		// how many introns in this gene (isoform)?
//	int ixIntron;			// which intron in the gene is this (index)
//	int posAtg, posStp;		// position of intron starting point, relative to CDS start and stop
//	int ibeg;
//	int iend;
//	int iframe;			// {0, 1, 2, 3 (non-coding), -1 (unknown)}
//	int flankLeft;
//	int flankRight;
//	int iStrand;
//	int gc;
//	int iJcnType;			// GT_AG, GC_AG, etc
//	int iIntronType;		// CDS (0,1,2), UTR
//	int nCoverage;
//	unsigned int extr_L;
//	unsigned int extr_R;
//	unsigned int extr_I;
//	int nFpL, nFpR;			// classification of left and right side of a false positive
//} sIntron;

//enum { _MULTI_PART = 0, _MULTI_BEG, _MULTI_ALN_L, _MULTI_GAP, _MULTI_ALN_R, _MULTI_MIS_L, _MULTI_MIS_R, _MULTI_CANO, _MULTI_ORI };

void cSite::addIntrons(int ix, bool bMultiFile)
{
	sIntron intron;
	sIntronCand cand;

	intron.extr_L = intron.extr_R = intron.extr_I = 0;
	intron.nFpL = intron.nFpR = -1;

	for (int i=0; i<(int)m_intronSets[ix].set.size(); ++i)
	{
		cand = m_intronSets[ix].set[i];
		// ixIntron will be a proxy for file position in the list of candidates
		intron.ixIntron = i;
		intron.feat = cand.intron.feat;
		// beg, end
		if (bMultiFile) {
			intron.ibeg = cand.multi_info[_MULTI_BEG] + cand.multi_info[_MULTI_ALN_L];
			intron.iend = intron.ibeg +  cand.multi_info[_MULTI_GAP] - 1;
		}
		else {
			intron.strId = m_intronSets[ix].SAM_line;
			intron.nContig = m_intronSets[ix].nContig;
			intron.ibeg = cand.intron.ibeg;
			intron.iend = cand.intron.iend;
			intron.info = cand.multi_info;
		}

		m_introns.push_back( intron );
	}
}

void cSite::sets2introns(bool bMultiFile)
{
	m_introns.clear();

	for (int i=0; i<(int)m_intronSets.size(); ++i)
	{
		addIntrons(i, bMultiFile);
	}
}

void cSite::initGappedAlignSets( vector<sIntronSet>& sets, SSeq *psseq, int iContig, bool bMultiFile )
{
	// clear data
	clearData();

	// copy intron sets to local
	for (int i=0; i<(int)sets.size(); ++i)
	{
		if (sets[i].nContig==iContig)
			m_intronSets.push_back( sets[i] );
	}

	sets2introns(bMultiFile);

	// get ptr to SSeq
	m_psseq = psseq;

	initParams();
}

void cSite::setGappedAlignSets( vector<sIntronSet>& sets )
{
	int ixSet(-1), ct(0);

	for (int i=0; i<(int)m_introns.size(); ++i)
	{
		if (m_introns[i].ixIntron==0) {
			++ixSet;
			ct = 0;
		}

		sets[ixSet].set[ct].intron = m_introns[i];
		++ct;
	}
}

//void cSite::setAlignments( vector<sAln>& alns )
//{
//	int n = alns.size();
//	int ix;
//
//	for (int i=0; i<m_introns.size(); ++i)
//	{
//		ix = m_introns[i].ixIntron;
//		if (ix>=0 && ix<n) {
//			alns[ix].iStrand = m_introns[i].iStrand;
//			alns[ix].iJcnType = m_introns[i].iJcnType;
//			alns[ix].don = m_introns[i].don;
//			alns[ix].acc = m_introns[i].acc;
//			alns[ix].nFpL = m_introns[i].nFpL;
//			alns[ix].nFpR = m_introns[i].nFpR;
//		}
//	}
//}

void cSite::initTranscripts( vector<sTranscript>& xscripts, SSeq *psseq, int iContig  )
{
	// clear data
	// clear genome partition buffers
	m_strNonCoding.clear();
	m_strOutputFileBase.clear();

	m_seqDons.clear();
	m_seqAccs.clear();

	// clear xscript data
	m_transcripts.clear();

	// copy xscripts to local
	for (int i=0; i<(int)xscripts.size(); ++i)
	{
		if (xscripts[i].nContig==iContig)
			m_transcripts.push_back( xscripts[i] );
	}

	// get ptr to SSeq
	m_psseq = psseq;

	// site definitions
	m_donFlankLeft = 3;
	m_donFlankRight = 4;

	m_accFlankLeft = 18;
	m_accFlankRight = 1;

	m_intronMarginLeft = 20;
	m_intronMarginRight = 27;

	m_minIntronLen = 20;
	m_minIntergLen = 10;

	// clear markov model buffers
	m_pModelMatDon = NULL;
	m_pModelMatAcc = NULL;
}

void cSite::initOrigins( vector<sIntron>& introns, SSeq *psseq, int iContig )
{
	// clear data
	// clear genome partition buffers
	m_strNonCoding.clear();
	m_strOutputFileBase.clear();

	m_seqDons.clear();
	m_seqAccs.clear();

	// clear xscript data
	m_introns.clear();

	string defline;
	psseq->getDefline(defline,iContig);
	string seqName = psseq->getNameTok(defline);

	// copy xscripts to local
	for (int i=0; i<(int)introns.size(); ++i)
	{
		if (introns[i].strId==seqName)
			m_introns.push_back( introns[i] );
	}

	// get ptr to SSeq
	m_psseq = psseq;

	// site definitions
	m_donFlankLeft = 3;
	m_donFlankRight = 4;

	m_accFlankLeft = 18;
	m_accFlankRight = 1;

	m_intronMarginLeft = 20;
	m_intronMarginRight = 27;

	m_minIntronLen = 20;
	m_minIntergLen = 10;

	// clear markov model buffers
	m_pModelMatDon = NULL;
	m_pModelMatAcc = NULL;
}

void cSite::initEstGenes( vector<sEstGene>& genes, SSeq *psseq, int iContig  )
{
	// clear data
	// clear genome partition buffers
	m_strNonCoding.clear();
	m_strOutputFileBase.clear();

	m_seqDons.clear();
	m_seqAccs.clear();

	// clear isoform data
	m_estGenes.clear();

	// copy isoforms to local, CAUTION: NO CONTIG ID CHECKING HERE!!!
	m_estGenes = genes;

	// get ptr to SSeq
	m_psseq = psseq;

	// site definitions
	m_startFlankLeft = 6;
	m_startFlankRight = 3;

	m_termFlankLeft = 0;
	m_termFlankRight = 9;

	m_pModelMatAtg = m_pModelMatTaa = m_pModelMatTag = m_pModelMatTga = NULL;
}

void cSite::clearData(void)
{
	// clear genome partition buffers
	m_strCoding.clear();
	m_strNonCoding.clear();
	m_strOutputFileBase.clear();
	m_strIntronFlank.clear();

	m_seqStarts.clear();
	m_seqTerms.clear();
	m_seqDons.clear();
	m_seqAccs.clear();
	m_splicePhase.clear();
	m_seqShortAccs.clear();

	m_seqBp.clear();

	// clear gene data
	m_genes.clear();

	m_introns.clear();
	m_strIntrons.clear();
	m_intronSets.clear();
}

void cSite::clearModels()
{
	// clear all the old models
	if (m_pModelMatAtg!=NULL) { delete []m_pModelMatAtg; m_pModelMatAtg = NULL; }
	if (m_pModelMatTaa!=NULL) { delete []m_pModelMatTaa; m_pModelMatTaa = NULL; }
	if (m_pModelMatTag!=NULL) { delete []m_pModelMatTag; m_pModelMatTag = NULL; }
	if (m_pModelMatTga!=NULL) { delete []m_pModelMatTga; m_pModelMatTga = NULL; }
	if (m_pModelMatDon0!=NULL) { delete []m_pModelMatDon0; m_pModelMatDon0 = NULL; }
	if (m_pModelMatDon1!=NULL) { delete []m_pModelMatDon1; m_pModelMatDon1 = NULL; }
	if (m_pModelMatDon2!=NULL) { delete []m_pModelMatDon2; m_pModelMatDon2 = NULL; }
	if (m_pModelMatAcc0!=NULL) { delete []m_pModelMatAcc0; m_pModelMatAcc0 = NULL; }
	if (m_pModelMatAcc1!=NULL) { delete []m_pModelMatAcc1; m_pModelMatAcc1 = NULL; }
	if (m_pModelMatAcc2!=NULL) { delete []m_pModelMatAcc2; m_pModelMatAcc2 = NULL; }
}

void cSite::reset()
{
	m_bFirstRun = true;
}

int cSite::getStartSiteLength()
{
	return (m_startFlankLeft + 3 + m_startFlankRight);
}

int cSite::getDonSiteLength()
{
	return (m_donFlankLeft + 2 + m_donFlankRight);
}

int cSite::getAccSiteLength()
{
	return (m_accFlankLeft + 2 + m_accFlankRight);
}

int cSite::getShortAccSiteLength()
{
	return (m_shortAccFlankLeft + 2 + m_shortAccFlankRight);
}

int cSite::getTermSiteLength()
{
	return (m_termFlankLeft + 3 + m_termFlankRight);
}

void cSite::recordSiteData(void)
{
	string strDir;
	string baseOutDir = "sites";

	strDir = baseOutDir + SLASH;

	// files to create
	string strFileStart =		"start_site.txt";
	string strFileStop =		"stop_site.txt";
	string strFileDon =			"don_site.txt";
	string strFileAcc =			"acc_site.txt";
	string strFilePhase =		"splice_phase.txt";
	string strFileCoding =		"coding.txt";
	string strFileIntrons =		"introns.txt";
	string strFileNonCoding =	"non_coding.txt";

	// branch point intron model
	string strFileShortAcc;
	string strBpData;
	if (m_bBranchPointNext)
	{
		strFileShortAcc =		"short_acc_site.txt";
		strBpData =				"bp_data.txt";
	}

	//// append to file lists
	//m_model.m_startSiteFile = strFileStart;
	//m_model.m_stopSiteFile = strFileStop;
	//m_model.m_donSiteFile = strFileDon;
	//m_model.m_accSiteFile = strFileAcc;
	//m_model.m_phaseFile = strFilePhase;
	//m_model.m_codingFile = strFileCoding;
	//m_model.m_intronFile = strFileIntrons;
	//m_model.m_nonCodingFile = strFileNonCoding;
	//if (m_site.m_bBranchPointNext)
	//{
	//	m_model.m_shortAccFile =	strFileShortAcc;
	//	m_model.m_bpDataFile =		strBpData;
	//	m_model.m_intronFastaFile =	strFileBase + "_introns.fas";
	//}

	char pch[2];
	if (m_bFirstRun) {
		m_bFirstRun = false;
		pch[0] = 'w';
	}
	else {
		pch[0] = 'a';
	}
	pch[1] = '\0';

	m_pfSiteStart =			fopen( strFileStart.c_str(), pch );
	m_pfSiteStop =			fopen( strFileStop.c_str(), pch );
	m_pfSiteDon =			fopen( strFileDon.c_str(), pch );
	m_pfSiteAcc =			fopen( strFileAcc.c_str(), pch );
	m_pfSplicePhase =		fopen( strFilePhase.c_str(), pch );
	m_pfSiteCoding =		fopen( strFileCoding.c_str(), pch );
	m_pfSiteIntrons =		fopen( strFileIntrons.c_str(), pch );
	m_pfSiteNonCoding =		fopen( strFileNonCoding.c_str(), pch );
	if (m_bBranchPointNext)
	{
		m_pfSiteShortAcc =		fopen( strFileShortAcc.c_str(), pch );
		m_pfSiteBranchPoint = 	fopen( strBpData.c_str(), pch );
	}

	// record sites
	writeStringArray( m_pfSiteStart, m_seqStarts );
	writeStringArray( m_pfSiteStop, m_seqTerms );
	writeStringArray( m_pfSiteDon, m_seqDons );
	writeStringArray( m_pfSiteAcc, m_seqAccs );
	writeIntArray( m_pfSplicePhase, m_splicePhase );
	if (m_bBranchPointNext)
	{
		writeStringArray( m_pfSiteShortAcc, m_seqShortAccs );
		writeBpData( m_pfSiteBranchPoint, m_seqBp );
	}

	// coding and non-coding regions
	fprintf(m_pfSiteCoding, "%s", m_strCoding.c_str() );
	fprintf(m_pfSiteIntrons, "%s", m_strIntrons.c_str() );
	fprintf(m_pfSiteNonCoding, "%s", m_strNonCoding.c_str() );

	// close files
	fclose( m_pfSiteStart );
	fclose( m_pfSiteStop );
	fclose( m_pfSiteDon );
	fclose( m_pfSiteAcc );
	fclose( m_pfSplicePhase );
	fclose( m_pfSiteCoding );
	fclose( m_pfSiteIntrons );
	fclose( m_pfSiteNonCoding );
	if (m_bBranchPointNext)
	{
		fclose( m_pfSiteShortAcc );
		fclose( m_pfSiteBranchPoint );
	}

}

void cSite::recordCdsData(void)
{
	string strDir;
	string baseOutDir = "sites";

	strDir = baseOutDir + SLASH;

	// files to create
	string strFileCoding =		"coding.txt";

	char pch[2];
	if (m_bFirstRun) {
		m_bFirstRun = false;
		pch[0] = 'w';
	}
	else {
		pch[0] = 'a';
	}
	pch[1] = '\0';

	m_pfSiteCoding =		fopen( strFileCoding.c_str(), pch );

	fprintf(m_pfSiteCoding, "%s", m_strCoding.c_str() );

	// close files
	fclose( m_pfSiteCoding );
}

void cSite::recordSiteDataTrans(void)
{
	string strDir;
	string baseOutDir = "sites";

	strDir = baseOutDir + SLASH;

	// files to create
	string strFileDon =			"don_site.txt";
	string strFileAcc =			"acc_site.txt";
	string strFileIntrons =		"introns.txt";
	string strFileNonCoding =	"non_coding.txt";

	char pch[2];
	if (m_bFirstRun) {
		m_bFirstRun = false;
		pch[0] = 'w';
	}
	else {
		pch[0] = 'a';
	}
	pch[1] = '\0';

	m_pfSiteDon =			fopen( strFileDon.c_str(), pch );
	m_pfSiteAcc =			fopen( strFileAcc.c_str(), pch );
	m_pfSiteIntrons =		fopen( strFileIntrons.c_str(), pch );
	m_pfSiteNonCoding =		fopen( strFileNonCoding.c_str(), pch );

	// record sites
	writeStringArray( m_pfSiteDon, m_seqDons );
	writeStringArray( m_pfSiteAcc, m_seqAccs );

	// coding and non-coding regions
	fprintf(m_pfSiteIntrons, "%s", m_strIntrons.c_str() );
	fprintf(m_pfSiteNonCoding, "%s", m_strNonCoding.c_str() );

	// close files
	fclose( m_pfSiteDon );
	fclose( m_pfSiteAcc );
	fclose( m_pfSiteIntrons );
	fclose( m_pfSiteNonCoding );
}

void cSite::recordGappedAlignData(void)
{
	string strDir;
	string baseOutDir = "sites";

	strDir = baseOutDir + SLASH;

	// files to create
	string strFileDon =			"don_site.txt";
	string strFileAcc =			"acc_site.txt";
	string strFileIntrons =		"introns.txt";
	string strFileFlank =		"flank.txt";

	char pch[2];
	if (m_bFirstRun) {
		m_bFirstRun = false;
		pch[0] = 'w';
	}
	else {
		pch[0] = 'a';
	}
	pch[1] = '\0';

	m_pfSiteDon =			fopen( strFileDon.c_str(), pch );
	m_pfSiteAcc =			fopen( strFileAcc.c_str(), pch );
	m_pfSiteIntrons =		fopen( strFileIntrons.c_str(), pch );
	m_pfSiteFlank =			fopen( strFileFlank.c_str(), pch );

	// record sites
	writeStringArray( m_pfSiteDon, m_seqDons );
	writeStringArray( m_pfSiteAcc, m_seqAccs );

	// coding and non-coding regions
	fprintf(m_pfSiteIntrons, "%s", m_strIntrons.c_str() );
	fprintf(m_pfSiteFlank, "%s", m_strIntronFlank.c_str() );

	// close files
	fclose( m_pfSiteDon );
	fclose( m_pfSiteAcc );
	fclose( m_pfSiteIntrons );
	fclose( m_pfSiteFlank );
}

void cSite::recordEstpData()
{
	// files to create
	string strFileStart =		"start_site.txt";
	string strFileStop =		"stop_site.txt";

	char pch[2];
	if (m_bFirstRun) {
		m_bFirstRun = false;
		pch[0] = 'w';
	}
	else {
		pch[0] = 'a';
	}
	pch[1] = '\0';

	m_pfSiteStart =			fopen( strFileStart.c_str(), pch );
	m_pfSiteStop =			fopen( strFileStop.c_str(), pch );

	// record sites
	writeStringArray( m_pfSiteStart, m_seqStarts );
	writeStringArray( m_pfSiteStop, m_seqTerms );


	// close files
	fclose( m_pfSiteStart );
	fclose( m_pfSiteStop );
}

//bool cSite::setRunState( runMatrix& runMat, int nIter )
//{
//	// Given run matrix (runMat) and iteration (nIter), sets the ON/OFF status
//	//  of BP intron model.
//
//	// bounds check
//	int nRunsTotal = runMat[0].lfParams.size();
//	if (nIter+1>nRunsTotal) {
//		report("Tried to index outside of run matrix.");
//		return true;		// error, out of bounds
//	}
//
//	int iBp = runMat[4].lfParams[nIter];
//
//	m_bBranchPoint = bool( iBp );
//
//	if (nIter<(nRunsTotal-1))
//	{
//		iBp = runMat[4].lfParams[nIter+1];
//
//		m_bBranchPointNext = bool( iBp );
//	}
//	else
//	{
//		// do not change these settings on last run ?
//		m_bBranchPointNext = m_bBranchPoint;
//	}
//
//	return false;
//}


/*
#                 Start Context
#                 -------------
#                 The current models for start context use sequences of the
#                 following format:
#
#                 {6 nucleotides} ATG {3 nucleotides}
#
#                 for a total of 12 nucleotides.
#                 This can be altered by modifying the following variables
#                 in subrotines get_start_context and get_start_context_reverse
#                      $nts_before_start = 6;
#                      $nts_after_start = 3;
#
#                 Stop Context
#                 ------------
#                                 TAA
#                 {0 nucleotides} TAG {9 nucleotides}
#                                 TGA
#
#                 Acceptor Site Context
#                 ---------------------
#                 {18 nucleotides} AG {1 nucleotide}
#
#                 Donor Site Context
#                 ------------------
#                 {3 nucleotides} GT {4 nucleotides}
#
#   REFERENCE:    Simplified representation of Eukaryotic gene structure
#                 ------------------------------------------------------
#
#                 NNNCCCCCCCCCCCCCCCCNNNNNNNNNNNNCCCCCCNNNNNNNNCCCCCCCCCCCCCCNNNNN
#                    <-    exon    -><- intron -><-e -><-- i -><---- e ----->
#                                                                         TAA
#                    ATG+++++++++++++GT........AG++++++GT....AG+++++++++++TAG
#                                                                         TGA
#                    start          donor    acceptor                     stop
#
#                 N = . = noncoding
#                 C = + = coding
#
#                 introns (noncoding) => begin with GT and end with AG
*/

int cSite::getCodingLen( codingRegion& gene )
{
	int nExons = (int)gene.boundaries.size();
	int len = 0;

	for (int i=0; i<nExons; ++i)
	{
		len += (gene.boundaries[i].iend - gene.boundaries[i].ibeg + 1);
	}

	return len;
}

bool cSite::extractStarts( vector<string>& seqStarts, int iSeq, int nl, int nr )
{
	string strStretch;
	int iExonStart;
	int nExons;
	int len = nr + nl + 3;
	int lpos;
	bool bStrand;
	int flag;
	int lenGene;
	unsigned int iMark;

	seqStarts.clear();

	// if no genes, return nothing
	if (m_genes.size()==0) {
		return false;
	}

	// exactly one start region per gene
	for (int ig=0; ig<(int)m_genes.size(); ++ig) {

		if (checkPartial(ig))
			continue;

		// filter out genes shorter than a threshold
		lenGene = getCodingLen( m_genes[ig] );
		if (lenGene<m_minCodingLen) { continue; }

		bStrand = m_genes[ig].bStrand;
		if (bStrand) {
			iExonStart = m_genes[ig].boundaries[0].ibeg;
			lpos = iExonStart - nl;
			iMark = m_genes[ig].boundaries[0].extr_L;
		}
		else {
			nExons = (int)m_genes[ig].boundaries.size();
			iExonStart = m_genes[ig].boundaries[nExons-1].iend;
			lpos = iExonStart + nl;
			iMark = m_genes[ig].boundaries[nExons-1].extr_R;
		}
		if (!m_psseq->boundsCheck(iSeq, lpos, len, bStrand)) { continue; }
		if (!m_psseq->getStretch( strStretch, iSeq, lpos, len, bStrand ))
			{ return true;}	// error

		if (strStretch.at(nl)!='A' || strStretch.at(nl+1)!='T' || strStretch.at(nl+2)!='G')
			flag = 1;

		if (m_bMark)
		{
			addFeatureSuffixSite( strStretch, iMark, m_genes[ig], 0, FEATURE_START );
		}

		seqStarts.push_back( strStretch );
	}	// gene loop

	return false;			// success
}

bool cSite::extractStartsEstp( vector<string>& seqStarts, int iSeq, int nl, int nr )
{
	string strStretch;
	int iExonStart;
	int len = nr + nl + 3;
	int lpos;
	bool bStrand;

	seqStarts.clear();

	// if no genes, return nothing
	if (m_estGenes.size()==0) {
		return false;
	}

	// exactly one start region per gene
	for (int ig=0; ig<(int)m_estGenes.size(); ++ig) {

		bStrand = m_estGenes[ig].isoforms[0].bStrand;
		if (bStrand) {
			iExonStart = m_estGenes[ig].absL - 2;
			lpos = iExonStart - nl;
		}
		else {
			iExonStart = m_estGenes[ig].absR - 1;
			lpos = iExonStart + nl;
		}
		if (!m_psseq->boundsCheck(iSeq, lpos, len, bStrand)) { continue; }
		if (!m_psseq->getStretch( strStretch, iSeq, lpos, len, bStrand ))
			{ return true;}	// error

		if (strStretch.at(nl)!='A' || strStretch.at(nl+1)!='T' || strStretch.at(nl+2)!='G')
			continue;

		seqStarts.push_back( strStretch );
	}	// gene loop

	return false;			// success
}

bool cSite::extractDon( vector<string>& seqDons, vector<int>& phaseDons, int iSeq, int nl, int nr )
{
	string strStretch;
	codingRegion gene;
	int iExonStop;
	int nExons;
	int len = nr + nl + 2;
	int lenCoding = 0;
	int lpos;
	int iph;		// donor phase
	bool bStrand;
	int nOffset = 2;
	nl -= nOffset;
	nr -= nOffset;
	int lenGene, lenExon;
	unsigned int iMark;

	seqDons.clear();

	// if no genes, return nothing
	if (m_genes.size()==0) {
		return false;
	}

	//FILE *pf = fopen("phase_check","w");

	int nPartial = 0;
	int nShort = 0;
	int nBounds = 0;

	// exactly (n-1) donor regions per n-exon gene
	for (int ig=0; ig<(int)m_genes.size(); ++ig) {

		if (checkPartial(ig))
			{ nPartial++; continue; }

		lenCoding = 0;

		// filter out genes shorter than a threshold
		lenGene = getCodingLen( m_genes[ig] );
		if (lenGene<m_minCodingLen)
			{ nShort++; continue; }

		gene = m_genes[ig];
		nExons = (int)gene.boundaries.size();
		bStrand = gene.bStrand;
		if (nExons>1) {
			if (bStrand) {
				// direct strand
				for (int ie=0; ie<(nExons-1); ++ie) {
					iExonStop = m_genes[ig].boundaries[ie].iend;
					lpos = iExonStop - 1 - nl;
					if (!m_psseq->boundsCheck(iSeq, lpos, len, bStrand)) { nBounds++; continue; }
					if (!m_psseq->getStretch( strStretch, iSeq, lpos, len, bStrand ))
						{ printf("error with getStretch #1\n"); return true; }	// error
					// donor phase
					lenExon = gene.boundaries[ie].iend - gene.boundaries[ie].ibeg + 1;
					lenCoding += lenExon;
					iph = floor( fmod( double(lenCoding), 3. ) + 0.5 );
					phaseDons.push_back( iph );
					//fprintf(pf, "%d  %d  %d  %d \n", ie+1, lenExon, lenCoding, iph);
					iMark = m_genes[ig].boundaries[ie].extr_R;
					if (m_bMark)
					{
						addFeatureSuffixSite( strStretch, iMark, m_genes[ig], ie, FEATURE_DON );
					}
					seqDons.push_back( strStretch );
				}	// loop over exons (up to next-to-last one)
			}
			else {
				// reverse strand
				for (int ie=(nExons-1); ie>0; --ie) {
					iExonStop = m_genes[ig].boundaries[ie].ibeg;
					lpos = iExonStop + 1 + nl;
					if (!m_psseq->boundsCheck(iSeq, lpos, len, bStrand))  { nBounds++; continue; }
					if (!m_psseq->getStretch( strStretch, iSeq, lpos, len, bStrand ))
						{ printf("error with getStretch #2\n"); return true; }	// error
					// donor phase
					lenExon = gene.boundaries[ie].iend - gene.boundaries[ie].ibeg + 1;
					lenCoding += lenExon;
					//iph = floor( fmod( double(lenCoding+1), 3. ) + 0.5 );
					iph = floor( fmod( double(lenCoding), 3. ) + 0.5 );
					phaseDons.push_back( iph );
					//fprintf(pf, "%d  %d  %d  %d \n", ie+1, lenExon, lenCoding, iph);
					iMark = m_genes[ig].boundaries[ie].extr_L;
					if (m_bMark)
					{
						addFeatureSuffixSite( strStretch, iMark, m_genes[ig], ie, FEATURE_DON );
					}
					seqDons.push_back( strStretch );
				}	// loop over exons (up to next-to-last one)
			}


		}	// if more than one exon (at least one donor site)

	}	// gene loop

	//fclose(pf);

	return false;			// success
}

bool cSite::extractDonTrans( vector<string>& seqDons, int iSeq, int nl, int nr )
{
	string strStretch;
	sTranscript xscript;
	int iExonStop;
	int nExons;
	int len = nr + nl + 2;
	int lenTranscript = 0;
	int lpos;
	bool bStrand;
	int nOffset = 2;
	nl -= nOffset;
	nr -= nOffset;
	int lenExon;
	unsigned int iMark;

	seqDons.clear();

	// if no transcripts, return nothing
	if (m_transcripts.size()==0) {
		return false;
	}

	//FILE *pf = fopen("phase_check","w");

	int nPartial = 0;
	int nShort = 0;
	int nBounds = 0;

	// exactly (n-1) donor regions per n-exon gene
	for (int ig=0; ig<(int)m_transcripts.size(); ++ig) {

		//if (checkPartial(ig))
		//	{ nPartial++; continue; }

		lenTranscript = 0;

		//// filter out genes shorter than a threshold
		//lenGene = getCodingLen( m_genes[ig] );
		//if (lenGene<m_minCodingLen)
		//	{ nShort++; continue; }

		xscript = m_transcripts[ig];
		nExons = (int)xscript.boundaries.size();
		bStrand = xscript.bStrand;

		if (nExons>1) {
			if (bStrand) {
				// direct strand
				for (int ie=0; ie<(nExons-1); ++ie) {
					iExonStop = m_transcripts[ig].boundaries[ie].iend;
					lpos = iExonStop - 1 - nl;
					if (!m_psseq->boundsCheck(iSeq, lpos, len, bStrand)) { nBounds++; continue; }
					if (!m_psseq->getStretch( strStretch, iSeq, lpos, len, bStrand ))
						{ printf("error with getStretch #1\n"); return true; }	// error
					// donor phase
					lenExon = xscript.boundaries[ie].iend - xscript.boundaries[ie].ibeg + 1;
					lenTranscript += lenExon;

					iMark = m_transcripts[ig].boundaries[ie].extr_R;
					if (m_bMark)
					{
						addFeatureSuffixSite( strStretch, iMark, xscript, ie, FEATURE_DON );
					}
					seqDons.push_back( strStretch );
				}	// loop over exons (up to next-to-last one)
			}
			else {
				// reverse strand
				for (int ie=(nExons-1); ie>0; --ie) {
					iExonStop = m_transcripts[ig].boundaries[ie].ibeg;
					lpos = iExonStop + 1 + nl;
					if (!m_psseq->boundsCheck(iSeq, lpos, len, bStrand))  { nBounds++; continue; }
					if (!m_psseq->getStretch( strStretch, iSeq, lpos, len, bStrand ))
						{ printf("error with getStretch #2\n"); return true; }	// error
					// donor phase
					lenExon = xscript.boundaries[ie].iend - xscript.boundaries[ie].ibeg + 1;
					lenTranscript += lenExon;

					iMark = m_transcripts[ig].boundaries[ie].extr_L;
					if (m_bMark)
					{
						addFeatureSuffixSite( strStretch, iMark, xscript, ie, FEATURE_DON );
					}
					seqDons.push_back( strStretch );
				}	// loop over exons (up to next-to-last one)
			}


		}	// if more than one exon (at least one donor site)

	}	// gene loop

	//fclose(pf);

	return false;			// success
}

bool cSite::extractDonSRA( vector<string>& seqDons, int iSeq, int nl, int nr, bool bCanonical )
{
	string strStretch;
	sIntron intron;
	int iExonStop;
	int len = nr + nl + 2;
	int ml(nl), mr(nr);
	int lenTranscript = 0;
	int lpos;
	bool bStrand;
	int nOffset = 1;
	nl -= nOffset;
	nr -= nOffset;

	seqDons.clear();

	// if no transcripts, return nothing
	if (m_introns.empty())
		return false;

	int nPartial = 0;
	int nShort = 0;
	int nBounds = 0;

	// exactly (n-1) donor regions per n-exon gene
	for (int ii=0; ii<(int)m_introns.size(); ++ii) {

		lenTranscript = 0;

		intron = m_introns[ii];

		//if (intron.iStrand==0)
		//	continue;		// non-canonical splice junctions at alignment gap

		bStrand = intron.iStrand != -1;

		if (bStrand) {
			// direct strand
			iExonStop = m_introns[ii].ibeg - 1;
			lpos = iExonStop - 1 - nl;
			if (!m_psseq->boundsCheck(iSeq, lpos, len, bStrand)) { nBounds++; continue; }
			if (!m_psseq->getStretch( strStretch, iSeq, lpos, len, bStrand )) {		// error, but handle here
				printf("donor site beyond sequence limit! %s : [%d,%d]\n",intron.strId.c_str(),lpos,lpos+len);
				strStretch.assign(len,'N');
			}
			//m_psseq->Upper( strStretch );

			if (bCanonical) {
				if (strStretch.at(ml)=='G' && strStretch.at(ml+1)=='T')
					seqDons.push_back( strStretch );
			}
			else
				seqDons.push_back( strStretch );
		}
		else {
			// reverse strand
			iExonStop = m_introns[ii].iend;
			lpos = iExonStop + nl;
			if (!m_psseq->boundsCheck(iSeq, lpos, len, bStrand))  { nBounds++; continue; }
			if (!m_psseq->getStretch( strStretch, iSeq, lpos, len, bStrand )) {		// error, but handle here
				printf("acceptor site beyond sequence limit! %s : [%d,%d]\n",intron.strId.c_str(),lpos,lpos+len);
				strStretch.assign(len,'N');
			}
			//m_psseq->Upper( strStretch );

			if (bCanonical) {
				if (strStretch.at(ml)=='G' && strStretch.at(ml+1)=='T')
					seqDons.push_back( strStretch );
			}
			else
				seqDons.push_back( strStretch );
		}

	}	// short alignment loop

	return false;
}

bool cSite::getAllSpliceSites( int iSeq )
{
	// extract sites
	// these offsets were hard-coded into the GeneMark-ES 2.0 program (Lomsadze et al)
	if (getIntronDon( iSeq, m_donFlankLeft, m_donFlankRight, false ))
		return true;

	if (getIntronAcc( iSeq, m_accFlankLeft, m_accFlankRight, false ))
		return true;

	return false;
}

bool cSite::getIntronDon( int iSeq, int nl, int nr, bool bCanonical )
{
	string strStretch;
	sIntron intron;
	int iExonStop;
	int len = nr + nl + 2;
	int ml(nl), mr(nr);
	int lenTranscript = 0;
	int lpos;
	bool bStrand;
	int nOffset = 1;
	nl -= nOffset;
	nr -= nOffset;

	// if no transcripts, return nothing
	if (m_introns.empty())
		return false;

	int nPartial = 0;
	int nShort = 0;
	int nBounds = 0;

	// exactly (n-1) donor regions per n-exon gene
	for (int ii=0; ii<(int)m_introns.size(); ++ii) {

		lenTranscript = 0;

		intron = m_introns[ii];

		//if (intron.iStrand==0)
		//	continue;		// non-canonical splice junctions at alignment gap

		bStrand = intron.iStrand != -1;

		if (bStrand) {
			// direct strand
			iExonStop = m_introns[ii].ibeg - 1;
			lpos = iExonStop - 1 - nl;
			if (!m_psseq->boundsCheck(iSeq, lpos, len, bStrand)) { nBounds++; continue; }
			if (!m_psseq->getStretch( strStretch, iSeq, lpos, len, bStrand ))
				{ printf("error with getStretch #1\n"); return true; }	// error

			if (bCanonical) {
				if (strStretch.at(ml)=='G' && strStretch.at(ml+1)=='T')
					m_introns[ii].don = strStretch;
			}
			else
				m_introns[ii].don = strStretch;
		}
		else {
			// reverse strand
			iExonStop = m_introns[ii].iend;
			lpos = iExonStop + nl;
			if (!m_psseq->boundsCheck(iSeq, lpos, len, bStrand))  { nBounds++; continue; }
			if (!m_psseq->getStretch( strStretch, iSeq, lpos, len, bStrand ))
				{ printf("error with getStretch #2\n"); return true; }	// error

			if (bCanonical) {
				if (strStretch.at(ml)=='G' && strStretch.at(ml+1)=='T')
					m_introns[ii].don = strStretch;
			}
			else
				m_introns[ii].don = strStretch;
		}

	}	// short alignment loop

	return false;
}

bool cSite::extractAcc( vector<string>& seqAccs, int iSeq, int nl, int nr )
{
	string strStretch;
	int iExonStart;
	int nExons;
	int len = nr + nl + 2;
	int lpos;
	bool bStrand;
	int nOffset = 2;
	nl += nOffset;
	nr += nOffset;
	int lenGene;
	unsigned int iMark;

	seqAccs.clear();

	// if no genes, return nothing
	if (m_genes.size()==0) {
		return false;
	}

	// exactly (n-1) donor regions per n-exon gene
	for (int ig=0; ig<(int)m_genes.size(); ++ig) {

		if (checkPartial(ig))
			continue;

		// filter out genes shorter than a threshold
		lenGene = getCodingLen( m_genes[ig] );
		if (lenGene<m_minCodingLen) { continue; }

		nExons = (int)m_genes[ig].boundaries.size();
		bStrand = m_genes[ig].bStrand;
		if (nExons>1) {
			if (bStrand) {
				// direct strand
				for (int ie=1; ie<nExons; ++ie) {
					iExonStart = m_genes[ig].boundaries[ie].ibeg;
					lpos = iExonStart - nl;
					if (!m_psseq->boundsCheck(iSeq, lpos, len, bStrand)) continue;
					if (!m_psseq->getStretch( strStretch, iSeq, lpos, len, bStrand ))
						return true;	// error
					iMark = m_genes[ig].boundaries[ie].extr_L;
					if (m_bMark)
					{
						addFeatureSuffixSite( strStretch, iMark, m_genes[ig], ie, FEATURE_ACC );
					}
					seqAccs.push_back( strStretch );
				}	// loop over exons (up to next-to-last one)
			}
			else {
				// reverse strand
				for (int ie=(nExons-2); ie>=0; --ie) {
					iExonStart = m_genes[ig].boundaries[ie].iend;
					lpos = iExonStart + nl;
					if (!m_psseq->boundsCheck(iSeq, lpos, len, bStrand)) continue;
					if (!m_psseq->getStretch( strStretch, iSeq, lpos, len, bStrand ))
						return true;	// error
					iMark = m_genes[ig].boundaries[ie].extr_R;
					if (m_bMark)
					{
						addFeatureSuffixSite( strStretch, iMark, m_genes[ig], ie, FEATURE_ACC );
					}
					seqAccs.push_back( strStretch );
				}	// loop over exons (up to next-to-last one)

			}
		}	// if more than one exon (at least one donor site)
	}	// gene loop

	return false;			// success
}

bool cSite::extractAccTrans( vector<string>& seqAccs, int iSeq, int nl, int nr )
{
	string strStretch;
	int iExonStart;
	int nExons;
	int len = nr + nl + 2;
	int lpos;
	bool bStrand;
	int nOffset = 2;
	nl += nOffset;
	nr += nOffset;
	unsigned int iMark;

	seqAccs.clear();

	// if no genes, return nothing
	if (m_transcripts.size()==0) {
		return false;
	}

	// exactly (n-1) donor regions per n-exon gene
	for (int ig=0; ig<(int)m_transcripts.size(); ++ig) {

		//// filter out genes shorter than a threshold
		//lenGene = getCodingLen( m_transcripts[ig] );
		//if (lenGene<m_minCodingLen) { continue; }

		nExons = (int)m_transcripts[ig].boundaries.size();
		bStrand = m_transcripts[ig].bStrand;

		if (nExons>1) {
			if (bStrand) {
				// direct strand
				for (int ie=1; ie<nExons; ++ie) {
					iExonStart = m_transcripts[ig].boundaries[ie].ibeg;
					lpos = iExonStart - nl;
					if (!m_psseq->boundsCheck(iSeq, lpos, len, bStrand)) continue;
					if (!m_psseq->getStretch( strStretch, iSeq, lpos, len, bStrand ))
						return true;	// error
					iMark = m_transcripts[ig].boundaries[ie].extr_L;
					if (m_bMark)
					{
						addFeatureSuffixSite( strStretch, iMark, m_transcripts[ig], ie, FEATURE_ACC );
					}
					seqAccs.push_back( strStretch );
				}	// loop over exons (up to next-to-last one)
			}
			else {
				// reverse strand
				for (int ie=(nExons-2); ie>=0; --ie) {
					iExonStart = m_transcripts[ig].boundaries[ie].iend;
					lpos = iExonStart + nl;
					if (!m_psseq->boundsCheck(iSeq, lpos, len, bStrand)) continue;
					if (!m_psseq->getStretch( strStretch, iSeq, lpos, len, bStrand ))
						return true;	// error
					iMark = m_transcripts[ig].boundaries[ie].extr_R;
					if (m_bMark)
					{
						addFeatureSuffixSite( strStretch, iMark, m_transcripts[ig], ie, FEATURE_ACC );
					}
					seqAccs.push_back( strStretch );
				}	// loop over exons (up to next-to-last one)

			}
		}	// if more than one exon (at least one donor site)
	}	// gene loop

	return false;			// success
}

bool cSite::extractAccSRA( vector<string>& seqAccs, int iSeq, int nl, int nr, bool bCanonical )
{
	string strStretch;
	int iExonStart;
	int ml(nl), mr(nr);
	int len = nr + nl + 2;
	int lpos;
	bool bStrand;
	int nOffset = 0;
	nl += nOffset;
	nr += nOffset;
	int lenTranscript;

	seqAccs.clear();

	// if no intron alignments, return nothing
	if (m_introns.empty())
		return false;

	for (int ii=0; ii<(int)m_introns.size(); ++ii) {

		lenTranscript = 0;

		//if (m_introns[ii].iStrand==0)
		//	continue;		// non-canonical splice junctions at alignment gap

		bStrand = m_introns[ii].iStrand != -1;

		if (bStrand) {
			// direct strand
			iExonStart = m_introns[ii].iend + 1;
			lpos = iExonStart - len;
			if (!m_psseq->boundsCheck(iSeq, lpos, len, bStrand)) 
				strStretch.assign(len,'N');
			if (!m_psseq->getStretch( strStretch, iSeq, lpos, len, bStrand ))
				strStretch.assign(len,'N');	// error
			//m_psseq->Upper( strStretch );

			if (bCanonical) {
				if (strStretch.at(ml)=='A' && strStretch.at(ml+1)=='G')
					seqAccs.push_back( strStretch );
			}
			else
				seqAccs.push_back( strStretch );
		}
		else {
			// reverse strand
			iExonStart = m_introns[ii].ibeg;
			lpos = iExonStart + nl;
			if (!m_psseq->boundsCheck(iSeq, lpos, len, bStrand))
				strStretch.assign(len,'N');
			if (!m_psseq->getStretch( strStretch, iSeq, lpos, len, bStrand ))
				strStretch.assign(len,'N');	// error
			//m_psseq->Upper( strStretch );

			if (bCanonical) {
				if (strStretch.at(ml)=='A' && strStretch.at(ml+1)=='G')
					seqAccs.push_back( strStretch );
			}
			else
				seqAccs.push_back( strStretch );
		}

	}	// short alignment loop

	return false;			// success
}

bool cSite::getIntronAcc( int iSeq, int nl, int nr, bool bCanonical )
{
	string strStretch;
	int iExonStart;
	int ml(nl), mr(nr);
	int len = nr + nl + 2;
	int lpos;
	bool bStrand;
	int nOffset = 0;
	nl += nOffset;
	nr += nOffset;
	int lenTranscript;

	// if no intron alignments, return nothing
	if (m_introns.empty())
		return false;

	for (int ii=0; ii<(int)m_introns.size(); ++ii) {

		lenTranscript = 0;

		//if (m_introns[ii].iStrand==0)
		//	continue;		// non-canonical splice junctions at alignment gap

		bStrand = m_introns[ii].iStrand != -1;

		if (bStrand) {
			// direct strand
			iExonStart = m_introns[ii].iend + 1;
			lpos = iExonStart - len;
			if (!m_psseq->boundsCheck(iSeq, lpos, len, bStrand)) continue;
			if (!m_psseq->getStretch( strStretch, iSeq, lpos, len, bStrand ))
				return true;	// error

			if (bCanonical) {
				if (strStretch.at(ml)=='A' && strStretch.at(ml+1)=='G')
					m_introns[ii].acc = strStretch;
			}
			else
				m_introns[ii].acc = strStretch;
		}
		else {
			// reverse strand
			iExonStart = m_introns[ii].ibeg;
			lpos = iExonStart + nl;
			if (!m_psseq->boundsCheck(iSeq, lpos, len, bStrand)) continue;
			if (!m_psseq->getStretch( strStretch, iSeq, lpos, len, bStrand ))
				return true;	// error

			if (bCanonical) {
				if (strStretch.at(ml)=='A' && strStretch.at(ml+1)=='G')
					m_introns[ii].acc = strStretch;
			}
			else
				m_introns[ii].acc = strStretch;
		}

	}	// short alignment loop

	return false;			// success
}

bool cSite::extractIntronTerm( vector<bpRecord>& bpData, int iSeq, int lenSeq, int nr_pad )
{
	string strStretch;
	int iStart;
	int nExons;
	bool bStrand;
	int nOffset = 2;
	int lenGene, lenIntron;
	int nGet;
	bpRecord rec;

	bpData.clear();

	// if no genes, return nothing
	if (m_genes.size()==0) {
		return false;
	}

	// exactly (n-1) donor regions per n-exon gene
	for (int ig=0; ig<(int)m_genes.size(); ++ig) {

		if (checkPartial(ig))
			continue;

		// filter out genes shorter than a threshold
		lenGene = getCodingLen( m_genes[ig] );
		if (lenGene<m_minCodingLen) { continue; }

		nExons = (int)m_genes[ig].boundaries.size();
		bStrand = m_genes[ig].bStrand;
		if (nExons>1) {
			if (bStrand) {
				// direct strand
				for (int ie=1; ie<nExons; ++ie) {
					iStart = m_genes[ig].boundaries[ie-1].iend + m_intronMarginLeft + 1;
					lenIntron = m_genes[ig].boundaries[ie].ibeg - nr_pad - iStart;
					if (lenIntron<=m_intronTermThresh)
						continue;
					if (lenSeq>lenIntron)
						nGet = lenIntron;
					else
					{
						nGet = lenSeq;
						iStart += (lenIntron - lenSeq);
					}
					if (!m_psseq->boundsCheck(iSeq, iStart, nGet, bStrand)) continue;
					if (!m_psseq->getStretch( strStretch, iSeq, iStart, nGet, bStrand ))
						return true;	// error

					rec.seqIntronTerm = strStretch;
					rec.istart = iStart - m_genes[ig].boundaries[ie-1].iend;
					rec.lenIntron = m_genes[ig].boundaries[ie].ibeg - m_genes[ig].boundaries[ie-1].iend - 1;
					bpData.push_back( rec );

				}	// loop over exons (up to next-to-last one)
			}
			else {
				// reverse strand
				for (int ie=(nExons-2); ie>=0; --ie) {
					iStart = m_genes[ig].boundaries[ie+1].ibeg - 1 - m_intronMarginLeft;
					lenIntron = iStart - m_genes[ig].boundaries[ie].iend - nr_pad;
					if (lenIntron<=m_intronTermThresh)
						continue;
					if (lenSeq>lenIntron)
						nGet = lenIntron;
					else
					{
						nGet = lenSeq;
						iStart -= (lenIntron - lenSeq);
					}
					if (!m_psseq->boundsCheck(iSeq, iStart, nGet, bStrand)) continue;
					if (!m_psseq->getStretch( strStretch, iSeq, iStart, nGet, bStrand ))
						return true;	// error

					rec.seqIntronTerm = strStretch;
					rec.istart = m_genes[ig].boundaries[ie+1].ibeg - iStart;
					rec.lenIntron = m_genes[ig].boundaries[ie+1].ibeg - m_genes[ig].boundaries[ie].iend - 1;
					bpData.push_back( rec );

				}	// loop over exons (up to next-to-last one)
			}

		}	// if more than one exon (at least one donor site)

	}	// gene loop

	return false;			// success
}

bool cSite::extractTerm( vector<string>& seqTerm, int iSeq, int nl,int  nr )
{
	string strStretch;
	int iExonStop;
	int nExons;
	int len = nr + nl + 3;
	int lpos;
	bool bStrand;
	int lenGene;
	unsigned int iMark;

	seqTerm.clear();

	// if no genes, return nothing
	if (m_genes.size()==0) {
		return false;
	}

	// exactly one stop region per gene
	for (int ig=0; ig<(int)m_genes.size(); ++ig) {

		if (checkPartial(ig))
			continue;

		// filter out genes shorter than a threshold
		lenGene = getCodingLen( m_genes[ig] );
		if (lenGene<m_minCodingLen) { continue; }

		nExons = (int)m_genes[ig].boundaries.size();
		bStrand = m_genes[ig].bStrand;
		if (bStrand) {
			// direct strand
			iExonStop = m_genes[ig].boundaries[nExons-1].iend;
			lpos = iExonStop - nl - 2;
			if (!m_psseq->boundsCheck(iSeq, lpos, len, bStrand)) continue;
			if (!m_psseq->getStretch( strStretch, iSeq, lpos, len, bStrand ))
				return true;	// error
			iMark = m_genes[ig].boundaries[nExons-1].extr_R;
			if (m_bMark)
			{
				addFeatureSuffixSite( strStretch, iMark, m_genes[ig], nExons-1, FEATURE_STOP );
			}
			seqTerm.push_back( strStretch );
		}
		else {
			// reverse strand
			iExonStop = m_genes[ig].boundaries[0].ibeg;
			lpos = iExonStop + nl + 2;
			if (!m_psseq->boundsCheck(iSeq, lpos, len, bStrand)) continue;
			if (!m_psseq->getStretch( strStretch, iSeq, lpos, len, bStrand ))
				return true;	// error
			iMark = m_genes[ig].boundaries[0].extr_L;
			if (m_bMark)
			{
				addFeatureSuffixSite( strStretch, iMark, m_genes[ig], nExons-1, FEATURE_STOP );
			}
			seqTerm.push_back( strStretch );
		}

	}	// gene loop

	return false;			// success
}

bool cSite::extractTermEstp( vector<string>& seqTerm, int iSeq, int nl, int nr )
{
	string strStretch;
	int iExonStop;
	int len = nr + nl + 3;
	int lpos;
	bool bStrand;

	seqTerm.clear();

	// if no isoforms, return nothing
	if (m_estGenes.size()==0) {
		return false;
	}

	// exactly one stop region per gene
	for (int ig=0; ig<(int)m_estGenes.size(); ++ig) {

		bStrand = m_estGenes[ig].isoforms[0].bStrand;
		if (bStrand) {
			iExonStop = m_estGenes[ig].absR;
			lpos = iExonStop - nl - 3;
		}
		else {
			iExonStop = m_estGenes[ig].absL;
			lpos = iExonStop + nl;
		}
		if (!m_psseq->boundsCheck(iSeq, lpos, len, bStrand)) { continue; }
		if (!m_psseq->getStretch( strStretch, iSeq, lpos, len, bStrand ))
			{ return true;}	// error

		seqTerm.push_back( strStretch );

		//bStrand = m_estGenes[ig].isoforms[0].bStrand;
		//if (bStrand) {
		//	// direct strand
		//	iExonStop = m_estGenes[ig].absR;
		//	lpos = iExonStop - nl - 2;
		//	if (!m_psseq->boundsCheck(iSeq, lpos, len, bStrand)) continue;
		//	if (!m_psseq->getStretch( strStretch, iSeq, lpos, len, bStrand ))
		//		return true;	// error
		//	seqTerm.push_back( strStretch );
		//}
		//else {
		//	// reverse strand
		//	iExonStop = m_estGenes[ig].absL;
		//	lpos = iExonStop + nl + 2;
		//	if (!m_psseq->boundsCheck(iSeq, lpos, len, bStrand)) continue;
		//	if (!m_psseq->getStretch( strStretch, iSeq, lpos, len, bStrand ))
		//		return true;	// error
		//	seqTerm.push_back( strStretch );
		//}

	}	// gene loop

	return false;			// success
}

bool cSite::getAllCodingSeq( string& strCod, int iSeq )
{
	string strSeg, strCodReg;
	int nExons;
	int lenExon, lenCod;
	int begOffset, endOffset;
	bool bStrand;
	int iMark_5, iMark_3;	// 5' and 3' annotation match

	strCod.clear();

	for (int ig=0; ig<(int)m_genes.size(); ++ig) {

		if (checkPartial( ig ))
			continue;

		if (m_nGC != -1) {
			if ( m_genes[ig].pctGC < m_nGC || m_genes[ig].pctGC >= (m_nGC+1) )
				continue;
		}

		nExons = (int)m_genes[ig].boundaries.size();
		lenCod = 0;
		strCodReg.clear();

		for (int i=0; i<nExons; ++i) {
			bStrand = m_genes[ig].bStrand;
			// grab from ibeg to iend of each exon (inclusively)
			lenExon = m_genes[ig].boundaries[i].iend - m_genes[ig].boundaries[i].ibeg + 1;
			lenCod += lenExon;

			begOffset = 0;
			endOffset = 0;
			
			if (bStrand) {
				if (getSequence(strSeg, iSeq, m_genes[ig].boundaries[i].ibeg, lenExon, true)) {
					return true;
				}
				strCodReg.append( strSeg );
			}
			else {
				if (getSequence(strSeg, iSeq, m_genes[ig].boundaries[i].iend, lenExon, false)) {
					return true;
				}
				strCodReg.insert( 0, strSeg );
			}

		}	// exon loop

		// copy to output string
		if (lenCod<=m_minCodingLen)
			continue;			// skip short exons

		int imod = (int) fmod( double(lenCod), 3. );
		if (imod != 0)
			continue;

		if (m_genes[ig].bStrand)
		{
			iMark_5 = m_genes[ig].boundaries[0].extr_L;
			iMark_3 = m_genes[ig].boundaries[nExons-1].extr_R;
			if (m_bMark)
			{
				addFeatureSuffix( strCodReg, iMark_5, iMark_3, m_genes[ig], 0, FEATURE_CDS );
			}
		}
		else
		{
			iMark_5 = m_genes[ig].boundaries[nExons-1].extr_R;
			iMark_3 = m_genes[ig].boundaries[0].extr_L;
			if (m_bMark)
			{
				addFeatureSuffix( strCodReg, iMark_5, iMark_3, m_genes[ig], 0, FEATURE_CDS );
			}
		}

		strCod.append( strCodReg );

		// do we need this??
		strCod += "\n";

	}	// gene loop

	return false;			// success
}

bool cSite::getAllCodingSeqEstp( string& strCod, int iSeq )
{
	string strSeg, strCodReg;
	int i;
	int lenExon, cumLen, len, nExons;
	int dCdsL, dCdsR;
	bool bStrand, bCds;

	strCod.clear();

	for (int ig=0; ig<(int)m_estGenes.size(); ++ig) {

		bStrand = m_estGenes[ig].isoforms[0].bStrand;
		nExons = (int)m_estGenes[ig].isoforms[0].exons.size();

		dCdsL = m_estGenes[ig].absL - m_estGenes[ig].isoforms[0].ibeg;
		//dCdsR = m_estGenes[ig].isoforms[0].iend - m_estGenes[ig].absR;
		for (i=nExons-1; i>=0; --i) {
			if ( m_estGenes[ig].isoforms[0].exons[i].iend > m_estGenes[ig].absR )
				dCdsR = m_estGenes[ig].isoforms[0].exons[i].iend - m_estGenes[ig].absR + 1;
			else
				break;
		}
		cumLen = 0;
		bCds = false;

		for (i=0; i<nExons; ++i) {
			bStrand = m_estGenes[ig].isoforms[0].bStrand;
			// grab from ibeg to iend of each exon (inclusively)
			lenExon = m_estGenes[ig].isoforms[0].exons[i].iend - m_estGenes[ig].isoforms[0].exons[i].ibeg + 1;
			cumLen += lenExon;

			if (m_estGenes[ig].isoforms[0].exons[i].iend < m_estGenes[ig].absL) continue;
			if (m_estGenes[ig].isoforms[0].exons[i].ibeg > m_estGenes[ig].absR) break;
		
			if (bStrand) {
				if (!bCds) {	// first CDS exon
					len = m_estGenes[ig].isoforms[0].exons[i].iend - m_estGenes[ig].absL + 1;
					bCds = true;
					if (getSequence(strSeg, iSeq, m_estGenes[ig].absL-2, len, true)) return true;
				}
				else {
					if (m_estGenes[ig].isoforms[0].exons[i].iend <= m_estGenes[ig].absR) {
						if (getSequence(strSeg, iSeq, m_estGenes[ig].isoforms[0].exons[i].ibeg-2, lenExon, true)) return true;
					}
					else {
						if (getSequence(strSeg, iSeq, m_estGenes[ig].isoforms[0].exons[i].ibeg-2, lenExon-dCdsR, true)) return true;
					}
				}
				strCodReg.append( strSeg );
			}
			else {
				if (!bCds) {	// first CDS exon
					len = m_estGenes[ig].isoforms[0].exons[i].iend - m_estGenes[ig].absL + 1;
					bCds = true;
					if (getSequence(strSeg, iSeq, m_estGenes[ig].isoforms[0].exons[i].iend-1, len, false)) return true;
				}
				else {
					if (m_estGenes[ig].isoforms[0].exons[i].iend <= m_estGenes[ig].absR) {
						if (getSequence(strSeg, iSeq, m_estGenes[ig].isoforms[0].exons[i].iend-1, lenExon, false)) return true;
					}
					else {
						if (getSequence(strSeg, iSeq, m_estGenes[ig].absR-3, lenExon-dCdsR, false)) return true;
					}
				}
				strCodReg.insert( 0, strSeg );
			}

		}	// exon loop

		strCod.append( strCodReg );
		strCodReg.clear();
		strSeg.clear();

		// do we need this??
		strCod += "\n";

	}	// gene loop

	return false;			// success
}

int cSite::getContigLength(int iContig)
{
	return m_psseq->getLength(iContig);
}

// feature types
//enum { FEATURE_INTERGENIC, FEATURE_INTRON, FEATURE_CDS,
//       FEATURE_EXON, FEATURE_EXON_INI, FEATURE_EXON_INT, FEATURE_EXON_TER,
//       FEATURE_START, FEATURE_STOP, FEATURE_DON, FEATURE_ACC };

void cSite::addFeatureSuffix( string& nt, int mark5, int mark3, codingRegion& gene, int ix, int ifeature )
{
	char pch[1024], chStrand('-');
	int istart, istop;
	int nExons = gene.boundaries.size();

	switch (ifeature) {
		case FEATURE_INTRON:
			istart = gene.boundaries[ix].iend + 1;
			istop = gene.boundaries[ix+1].ibeg - 1;
			if (gene.bStrand) chStrand = '+';
			break;

		case FEATURE_INTERGENIC:
			istart = istop = 0;
			chStrand = '.';
			break;

		case FEATURE_CDS:
			istart = gene.boundaries[0].ibeg;
			istop = gene.boundaries[nExons-1].iend;
			if (gene.bStrand) chStrand = '+';
			break;

		case FEATURE_EXON: case FEATURE_EXON_INI: case FEATURE_EXON_INT: case FEATURE_EXON_TER:
			istart = gene.boundaries[ix].ibeg;
			istop = gene.boundaries[ix].iend;
			if (gene.bStrand) chStrand = '+';
			break;

		default:
			// error?
			return;
	}

	sprintf( pch, "\t%s\t%d\t%d\t%d\t%d\t%c\t%d", gene.strMrnaId.c_str(), mark5, mark3, istart, istop, chStrand, ifeature );
	string strSuffix = pch;
	nt.append(strSuffix);
}

void cSite::addFeatureSuffix( string& nt, int mark5, int mark3, sTranscript& xscript, int ix, int ifeature )
{
	char pch[1024], chStrand('-');
	int istart, istop;
	int nExons = xscript.boundaries.size();

	switch (ifeature) {
		case FEATURE_INTRON:
			istart = xscript.boundaries[ix].iend + 1;
			istop = xscript.boundaries[ix+1].ibeg - 1;
			if (xscript.bStrand) chStrand = '+';
			break;

		case FEATURE_INTERGENIC:
			istart = istop = 0;
			chStrand = '.';
			break;

		case FEATURE_CDS:
			istart = xscript.boundaries[0].ibeg;
			istop = xscript.boundaries[nExons-1].iend;
			if (xscript.bStrand) chStrand = '+';
			break;

		case FEATURE_EXON: case FEATURE_EXON_INI: case FEATURE_EXON_INT: case FEATURE_EXON_TER:
			istart = xscript.boundaries[ix].ibeg;
			istop = xscript.boundaries[ix].iend;
			if (xscript.bStrand) chStrand = '+';
			break;

		default:
			// error?
			return;
	}

	sprintf( pch, "\t%s\t%d\t%d\t%d\t%d\t%c\t%d", xscript.strMrnaId.c_str(), mark5, mark3, istart, istop, chStrand, ifeature );
	string strSuffix = pch;
	nt.append(strSuffix);
}

void cSite::addFeatureSuffixHet( string& nt, int mark5, int mark3, string& name, sIntron& intron  )
{
	char pch[1024], chStrand('-');
	int istart, istop;

	istart = intron.ibeg;
	istop = intron.iend;

	switch (intron.iStrand)
	{
	case -1:
		chStrand = '-';
		break;

	case 1:
		chStrand = '+';
		break;

	default:
		chStrand = '*';
	}

	int markI = intron.extr_I;
	int iframe = intron.iframe;

	if (m_bMark)
	{
		sprintf( pch, "\t%s\t%s\t%d\t%d\t%d\t%c\t%d\t%d\t%d\t%d\t%d\t%d\t%d\t%d\t%d\t%d\t%d\t%d\t%d\t%d", name.c_str(), intron.strId.c_str(), istart, istop, intron.iJcnType, chStrand, intron.gc, iframe, mark5, mark3, markI,
			intron.nCoverage, intron.nIntronsInGene, intron.ixIntron, intron.posAtg, intron.posStp, intron.nFpL, intron.nFpR, intron.dDon, intron.dAcc );
	}
	else
	{
		sprintf( pch, "\t%s\t%s\t%d\t%d\t%d\t%c\t%d", name.c_str(), intron.strId.c_str(), istart, istop, intron.iJcnType, chStrand, intron.gc );
	}

	string strSuffix = pch;
	nt.append(strSuffix);

	if (!intron.feat.empty())
	{
		nt.append("\t");
		for (int i=0; i<(int)intron.feat.size()-1; ++i) {
			nt.append( intron.feat[i] );
			nt.append( "|" );
		}
		nt.append( intron.feat[intron.feat.size()-1] );
	}
	else
	{
		nt.append("\t");
		nt.append("*");
	}
}

void cSite::addFeatureSuffix( string& nt, int mark5, int mark3, string& name, sIntron& intron  )
{
	char pch[1024], chStrand('-');
	int istart, istop;

	istart = intron.ibeg;
	istop = intron.iend;

	switch (intron.iStrand)
	{
	case -1:
		chStrand = '-';
		break;

	case 1:
		chStrand = '+';
		break;

	default:
		chStrand = '*';
	}

		//strSeg.append("\t");
		//strSeg.append(name);
		//strSeg.append("\t");
		//strSeg.append(m_introns[ii].strId);
		//strSeg.append("\t");
		//strSeg.append(strBeg);
		//strSeg.append("\t");
		//strSeg.append(strEnd);
		//strSeg.append("\t");
		//strSeg.append(strJcnType);
		//strSeg.append("\t");
		//strSeg.append(1,chStrand);

	int markI = intron.extr_I;
	int iframe = intron.iframe;

	if (m_bMark)
	{
		sprintf( pch, "\t%s\t%s\t%d\t%d\t%d\t%c\t%d\t%d\t%d\t%d\t%d\t%d\t%d\t%d\t%d", name.c_str(), intron.strId.c_str(), istart, istop, intron.iJcnType, chStrand, iframe, mark5, mark3, markI,
			intron.nCoverage, intron.nIntronsInGene, intron.ixIntron, intron.posAtg, intron.posStp );
	}
	else
	{
		sprintf( pch, "\t%s\t%s\t%d\t%d\t%d\t%c", name.c_str(), intron.strId.c_str(), istart, istop, intron.iJcnType, chStrand );
	}

	string strSuffix = pch;
	nt.append(strSuffix);
}

void cSite::addFeatureSuffixSite( string& nt, int mark, codingRegion& gene, int ix, int ifeature )
{
	char pch[128], chStrand('-');
	int istart, istop, iframe;
	int nExons = gene.boundaries.size();

	switch (ifeature) {
		case FEATURE_START:
			istart = gene.boundaries[0].ibeg - m_startFlankLeft;
			istop = gene.boundaries[0].ibeg + m_startFlankRight;
			if (gene.bStrand) chStrand = '+';
			iframe = -1;
			break;

		case FEATURE_STOP:
			istart = gene.boundaries[nExons-1].iend - m_termFlankLeft;
			istop = gene.boundaries[nExons-1].iend + m_termFlankRight;
			if (gene.bStrand) chStrand = '+';
			iframe = -1;
			break;

		case FEATURE_DON:
			istart = gene.boundaries[ix].iend - m_donFlankLeft;
			istop = gene.boundaries[ix].iend + m_donFlankRight;
			if (gene.bStrand) chStrand = '+';
			iframe = gene.boundaries[ix].iframe;
			break;

		case FEATURE_ACC:
			istart = gene.boundaries[ix].ibeg - m_accFlankLeft;
			istop = gene.boundaries[ix].ibeg + m_accFlankRight;
			if (gene.bStrand) chStrand = '+';
			iframe = gene.boundaries[ix].iframe;
			break;

		default:
			// error?
			return;
	}

	sprintf( pch, "\t%s\t%d\t%d\t%d\t%c\t%d\t%d", gene.strMrnaId.c_str(), mark, istart, istop, chStrand, iframe, ifeature );
	string strSuffix = pch;
	nt.append(strSuffix);
}

void cSite::addFeatureSuffixSite( string& nt, int mark, sTranscript& xscript, int ix, int ifeature )
{
	char pch[256], chStrand('-');
	int istart, istop, iframe;

	switch (ifeature) {

		case FEATURE_DON:
			istart = xscript.boundaries[ix].iend - m_donFlankLeft;
			istop = xscript.boundaries[ix].iend + m_donFlankRight;
			if (xscript.bStrand) chStrand = '+';
			iframe = xscript.boundaries[ix].iframe;
			break;

		case FEATURE_ACC:
			istart = xscript.boundaries[ix].ibeg - m_accFlankLeft;
			istop = xscript.boundaries[ix].ibeg + m_accFlankRight;
			if (xscript.bStrand) chStrand = '+';
			iframe = xscript.boundaries[ix].iframe;
			break;

		default:
			// error?
			return;
	}

	sprintf( pch, "\t%s\t%d\t%d\t%d\t%c\t%d\t%d", xscript.strMrnaId.c_str(), mark, istart, istop, chStrand, iframe, ifeature );
	string strSuffix = pch;
	nt.append(strSuffix);
}

bool cSite::getAllIntergenicSeq( string& strIntergenic, int iSeq )
{
	string strSeg;
	int nExons;
	int iStart;
	bool bStrand;
	int iMark_5, iMark_3;	// 5' and 3' annotation match

	strIntergenic.clear();

	//FILE *pF = fopen("intg_debug.txt","w");
	//if (pF==NULL)
	//	return true;

	// if no genes, return whole sequence
	if (m_genes.size()==0) {
		if (getSequence(strSeg, iSeq, 0, getContigLength(iSeq), true))
			return true;

		strIntergenic.append( strSeg );
		return false;
	}

	// 1. get all non-coding nt up to beg of first exon
	int nCopy;
	if (false)
	{
		nCopy = m_genes[0].boundaries[0].ibeg - 1;
		if (nCopy>0) {
			bStrand = true;
			if (getSequence(strSeg, iSeq, 0, nCopy, true))
				return true;
			strIntergenic.append( strSeg );
			strIntergenic.append("\n");
		}
	}

	// 2. all intergenic regions between boundaries specified by gene vector
	for (int ig=0; ig<(int)m_genes.size()-1; ++ig) {
		nExons = (int)m_genes[ig].boundaries.size();
		iStart = m_genes[ig].boundaries[nExons-1].iend + 1;
		nCopy = m_genes[ig+1].boundaries[0].ibeg - iStart;
		if (nCopy>m_minIntergLen) {
			bStrand = m_genes[ig].bStrand;
			bStrand = true;
			if (bStrand) {
				if (getSequence(strSeg, iSeq, iStart, nCopy, true)) {
					return true;
				}
				//fprintf(pF, "%d, %d, %d\n", iStart, iStart+nCopy, nCopy);
				iMark_5 = m_genes[ig].boundaries[nExons-1].extr_R;
				iMark_3 = m_genes[ig+1].boundaries[0].extr_L;
				if (m_bMark)
				{
					addFeatureSuffix( strSeg, iMark_5, iMark_3, m_genes[ig], 0, FEATURE_INTERGENIC );
				}
			}
			else {
				if (getSequence(strSeg, iSeq, iStart+nCopy, nCopy, false)) {
					return true;
				}
				//fprintf(pF, "%d, %d, %d\n", iStart, iStart+nCopy, nCopy);
				iMark_5 = m_genes[ig+1].boundaries[0].extr_L;
				iMark_3 = m_genes[ig].boundaries[nExons-1].extr_R;
				if (m_bMark)
				{
					addFeatureSuffix( strSeg, iMark_5, iMark_3, m_genes[ig], 0, FEATURE_INTERGENIC );
				}
			}
			strIntergenic.append( strSeg );
			strIntergenic.append("\n");
		}
	}

	// 3. get final segment, from last exon to end of sequence
	if (false)
	{
		nExons = (int)m_genes[m_genes.size()-1].boundaries.size();
		iStart = m_genes[m_genes.size()-1].boundaries[nExons-1].iend + 1;
		nCopy = getContigLength(iSeq) - iStart - 1;
		if (nCopy>0) {
			if (getSequence(strSeg, iSeq, iStart, nCopy, true))
				return true;
			strIntergenic.append( strSeg );
			strIntergenic.append("\n");
		}
	}

	//fclose(pF);

	return false;		// success
}

bool cSite::getAllIntergenicSeqEstp( string& strIntergenic, int iSeq )
{
	string strSeg;
	int nFlank(300);
	bool bStrand;

	strIntergenic.clear();

	for (int ig=0; ig<(int)m_estGenes.size(); ++ig) {

		bStrand = m_estGenes[ig].isoforms[0].bStrand;
		strSeg.clear();

		// 5' flank
		if (getSequence(strSeg, iSeq, m_estGenes[ig].absL-nFlank-1, nFlank, true)) {
			return true;
		}
		strIntergenic.append( strSeg );
		// do we need this??
		strIntergenic += "\n";

		// 3' flank
		strSeg.clear();
		if (getSequence(strSeg, iSeq, m_estGenes[ig].absR+1, nFlank, true)) {
			return true;
		}
		strIntergenic.append( strSeg );
		// do we need this??
		strIntergenic += "\n";

	}	// gene loop

	return false;			// success
}

bool cSite::getAllIntergenicSeqTrans( string& strIntergenic, int iSeq )
{
	string strSeg;
	int nExons;
	int iStart;
	bool bStrand;
	int iMark_5, iMark_3;	// 5' and 3' annotation match

	strIntergenic.clear();

	//FILE *pF = fopen("intg_debug.txt","w");
	//if (pF==NULL)
	//	return true;

	// if no genes, return whole sequence
	if (m_transcripts.size()==0) {
		if (getSequence(strSeg, iSeq, 0, getContigLength(iSeq), true))
			return true;

		strIntergenic.append( strSeg );
		return false;
	}

	// 1. get all non-coding nt up to beg of first exon
	int nCopy;
	if (false)
	{
		nCopy = m_transcripts[0].boundaries[0].ibeg - 1;
		if (nCopy>0) {
			bStrand = true;
			if (getSequence(strSeg, iSeq, 0, nCopy, true))
				return true;
			strIntergenic.append( strSeg );
			strIntergenic.append("\n");
		}
	}

	// 2. all intergenic regions between boundaries specified by gene vector
	for (int ig=0; ig<(int)m_transcripts.size()-1; ++ig) {
		nExons = (int)m_transcripts[ig].boundaries.size();
		iStart = m_transcripts[ig].boundaries[nExons-1].iend + 1;
		if ( m_transcripts[ig+1].strGeneId.compare( m_transcripts[ig].strGeneId )==0 )
			continue;

		nCopy = m_transcripts[ig+1].boundaries[0].ibeg - iStart;
		if (nCopy>m_minIntergLen) {
			bStrand = m_transcripts[ig].bStrand;
			bStrand = true;
			if (bStrand) {
				if (getSequence(strSeg, iSeq, iStart, nCopy, true)) {
					return true;
				}
				iMark_5 = m_transcripts[ig].boundaries[nExons-1].extr_R;
				iMark_3 = m_transcripts[ig+1].boundaries[0].extr_L;
				if (m_bMark)
				{
					addFeatureSuffix( strSeg, iMark_5, iMark_3, m_transcripts[ig], 0, FEATURE_INTERGENIC );
				}
			}
			else {
				if (getSequence(strSeg, iSeq, iStart+nCopy, nCopy, false)) {
					return true;
				}
				iMark_5 = m_transcripts[ig+1].boundaries[0].extr_L;
				iMark_3 = m_transcripts[ig].boundaries[nExons-1].extr_R;
				if (m_bMark)
				{
					addFeatureSuffix( strSeg, iMark_5, iMark_3, m_transcripts[ig], 0, FEATURE_INTERGENIC );
				}
			}
			strIntergenic.append( strSeg );
			strIntergenic.append("\n");
		}
	}

	// 3. get final segment, from last exon to end of sequence
	if (false)
	{
		nExons = (int)m_transcripts[m_transcripts.size()-1].boundaries.size();
		iStart = m_transcripts[m_transcripts.size()-1].boundaries[nExons-1].iend + 1;
		nCopy = getContigLength(iSeq) - iStart - 1;
		if (nCopy>0) {
			if (getSequence(strSeg, iSeq, iStart, nCopy, true))
				return true;
			strIntergenic.append( strSeg );
			strIntergenic.append("\n");
		}
	}

	return false;		// success
}

bool cSite::getAllIntronSeq( string& strIntrons, int iSeq )
{
	string strSeg;
	int nExons;
	int lenIntron;
	int iStart;
	bool bStrand;
	int iMark_5, iMark_3;

	strIntrons.clear();

	// if no genes, return nothing
	if (m_genes.size()==0) {
		return false;
	}

	// grab all introns
	for (int ig=0; ig<(int)m_genes.size(); ++ig) {

		if (checkPartial( ig ))
			continue;

		if (m_nGC != -1) {
			if ( m_genes[ig].pctGC < m_nGC || m_genes[ig].pctGC >= (m_nGC+1) )
				continue;
		}

		nExons = (int)m_genes[ig].boundaries.size();
		if (nExons>1) {
			for (int i=0; i<nExons-1; ++i) {
				bStrand = m_genes[ig].bStrand;
				// grab from ibeg to iend of each exon (inclusively)
				if (bStrand) {
					iStart = m_genes[ig].boundaries[i].iend + 1 + m_intronMarginLeft;
					lenIntron = m_genes[ig].boundaries[i+1].ibeg - iStart - m_intronMarginRight;
					if (lenIntron<m_minIntronLen) continue;
					if (getSequence(strSeg, iSeq, iStart, lenIntron, true)) {
						return true;
					}
					iMark_5 = m_genes[ig].boundaries[i].extr_R;
					iMark_3 = m_genes[ig].boundaries[i+1].extr_L;
				}
				else {
					iStart = m_genes[ig].boundaries[i+1].ibeg - 1 - m_intronMarginLeft;
					lenIntron = iStart - m_genes[ig].boundaries[i].iend - m_intronMarginRight;
					if (lenIntron<m_minIntronLen) continue;
					if (getSequence(strSeg, iSeq, iStart, lenIntron, false)) {
						return true;
					}
					iMark_5 = m_genes[ig].boundaries[i+1].extr_L;
					iMark_3 = m_genes[ig].boundaries[i].extr_R;
				}

				if (m_bMark)
				{
					addFeatureSuffix( strSeg, iMark_5, iMark_3, m_genes[ig], i, FEATURE_INTRON );
				}
				// copy to output string
				strIntrons.append( strSeg );
				strIntrons.append("\n");

			}	// exon loop
		}
	}	// gene loop

	return false;			// success
}

bool cSite::getAllIntronSeqTrans( string& strIntrons, int iSeq )
{
	string strSeg;
	int nExons;
	int lenIntron;
	int iStart;
	bool bStrand;
	int iMark_5, iMark_3;

	strIntrons.clear();

	// if no genes, return nothing
	if (m_transcripts.size()==0) {
		return false;
	}

	// grab all introns
	for (int ig=0; ig<(int)m_transcripts.size(); ++ig) {

		nExons = (int)m_transcripts[ig].boundaries.size();
		if (nExons>1) {
			for (int i=0; i<nExons-1; ++i) {
				bStrand = m_transcripts[ig].bStrand;
				// grab from ibeg to iend of each exon (inclusively)
				if (bStrand) {
					iStart = m_transcripts[ig].boundaries[i].iend + 1 + m_intronMarginLeft;
					lenIntron = m_transcripts[ig].boundaries[i+1].ibeg - iStart - m_intronMarginRight;
					if (lenIntron<m_minIntronLen) continue;
					if (getSequence(strSeg, iSeq, iStart, lenIntron, true)) {
						return true;
					}
					iMark_5 = m_transcripts[ig].boundaries[i].extr_R;
					iMark_3 = m_transcripts[ig].boundaries[i+1].extr_L;
				}
				else {
					iStart = m_transcripts[ig].boundaries[i+1].ibeg - 1 - m_intronMarginLeft;
					lenIntron = iStart - m_transcripts[ig].boundaries[i].iend - m_intronMarginRight;
					if (lenIntron<m_minIntronLen) continue;
					if (getSequence(strSeg, iSeq, iStart, lenIntron, false)) {
						return true;
					}
					iMark_5 = m_transcripts[ig].boundaries[i+1].extr_L;
					iMark_3 = m_transcripts[ig].boundaries[i].extr_R;
				}

				if (m_bMark)
				{
					addFeatureSuffix( strSeg, iMark_5, iMark_3, m_transcripts[ig], i, FEATURE_INTRON );
				}
				// copy to output string
				strIntrons.append( strSeg );
				strIntrons.append("\n");

			}	// exon loop
		}
	}	// gene loop

	return false;			// success
}

bool cSite::getStrand( SSeq *pseq, int iSeq, int nOffset1, int nOffset2 )
{
	if (getStrand( m_introns, pseq, iSeq, nOffset1, nOffset2 ))
		return true;

		// extract sites
	// these offsets were hard-coded into the GeneMark-ES 2.0 program (Lomsadze et al)
	if (getIntronDon( iSeq, m_donFlankLeft, m_donFlankRight, false ))
		return true;

	if (getIntronAcc( iSeq, m_accFlankLeft, m_accFlankRight, false ))
		return true;

	return false;
}

bool cSite::getStrand( vector<sIntron>& introns, SSeq *pseq, int iSeq, int nOffset1, int nOffset2 )
{
	string str5, str3, strJcnType;
	int ii, iStart, iStrand, iJcnType;
	int len;

	m_psseq = pseq;

	// if no introns, return
	if (introns.empty())
		return false;

	// get appropriate contig name
	string defline;
	m_psseq->getDefline( defline, iSeq );
	string name = m_psseq->getNameTok( defline );

	// grab all introns
	for (ii=0; ii<(int)introns.size(); ++ii) {

		// get intron terminal dinucleotide sequences and check for canonicality (strand-ness)
		// 5' end
		iStart = introns[ii].ibeg - 2;
		len = 4;
		if (getSequence(str5, iSeq, iStart, len, true))
		{		// bad coordinate!
			iStrand = 0;
			iJcnType = INTRON_BAD;
		}

		// 3' end
		iStart = introns[ii].iend - 3;
		len = 4;
		if (getSequence(str3, iSeq, iStart, len, true))
		{		// bad coordinate!
			iStrand = 0;
			iJcnType = INTRON_BAD;
		}

		getJcnType(iStrand, iJcnType, str5, str3, iSeq);

		// set intron[ii] strand here
		introns[ii].iStrand = iStrand;
		introns[ii].iJcnType = iJcnType;

		//if (introns[ii].iStrand == -1) {
		//	introns[ii].ibeg -= nOffset1;
		//	introns[ii].iend -= nOffset2;
		//}
	}

	return false;			// success
}

bool cSite::getAllIntronSeqSRA( string& strIntrons, string& strFlank, int iSeq, int nFlank )
{
	string strSeg, str5, str5i, str3, str3i, strBeg, strEnd, strJcnType;
	int lenIntron;
	int iStart, iStart5, iStart3, len;
	bool bStrand;
	char chStrand;
	int ii, iStrand, iJcnType, iMark_5, iMark_3;

	strIntrons.clear();
	strFlank.clear();
	m_intronTypes.clear();

	// if no gapped SRA, return nothing
	if (m_introns.empty())
		return false;

	// get appropriate contig name
	string defline;
	m_psseq->getDefline( defline, iSeq );
	string name = m_psseq->getNameTok( defline );

	// grab all introns
	for (ii=0; ii<(int)m_introns.size(); ++ii) {

		// get intron terminal dinucleotide sequences and check for canonicality (strand-ness)
		// 5' end
		iStart = m_introns[ii].ibeg - 2;
		len = 4;
		if (getSequence(str5, iSeq, iStart, len, true))
			str5.assign(len,'N');		// bad coordinate!
		// 3' end
		iStart = m_introns[ii].iend - 3;
		len = 4;
		if (getSequence(str3, iSeq, iStart, len, true))
			str3.assign(len,'N');		// bad coordinate!

		getJcnType(iStrand, iJcnType, str5, str3, iSeq);
		m_introns[ii].iJcnType = iJcnType;

		// set m_intron[ii] strand here
		m_introns[ii].iStrand = iStrand;

		strSeg.clear();

		//// beg, end strings
		//sprintf( pch, "%d", m_introns[ii].ibeg );
		//strBeg = pch;
		//sprintf( pch, "%d", m_introns[ii].iend );
		//strEnd = pch;
		// type
		//sprintf( pch, "%d", iJcnType );
		//strJcnType = pch;

		switch (iStrand)
		{
		case -1:
			bStrand = false;
			chStrand = '-';
			break;
		case 1:
			bStrand = true;
			chStrand = '+';
			break;
		default:
			bStrand = true;
			chStrand = '*';
		};

		// grab from ibeg to iend of each exon (inclusively)
		if (bStrand || chStrand == '*') {
			iStart = m_introns[ii].ibeg + 1 + m_intronMarginLeft;
			lenIntron = m_introns[ii].iend - iStart - m_intronMarginRight;

			// intron
			if (lenIntron<m_minIntronLen) {
				strSeg = "*";
			}
			else {
				if (getSequence(strSeg, iSeq, iStart, 10 /*lenIntron*/, true))
					strSeg.assign(10,'N');
			}

			// 5' flank
			iStart5 = m_introns[ii].ibeg - nFlank - 2 + 1;
			if (getSequence(str5, iSeq, iStart5, nFlank, true))
				str5.assign(nFlank,'N');
			if (getSequence(str5i, iSeq, m_introns[ii].ibeg+5, nFlank, true))
				str5i.assign(nFlank,'N');

			// 3' flank
			iStart3 = m_introns[ii].iend;
			if (getSequence(str3, iSeq, iStart3, nFlank, true))
				str3.assign(nFlank,'N');
			if (getSequence(str3i, iSeq, m_introns[ii].iend - 2 - nFlank - 20, nFlank, true))
				str3i.assign(nFlank,'N');

			iMark_5 = m_introns[ii].extr_L;
			iMark_3 = m_introns[ii].extr_R;

			addFeatureSuffixHet( strSeg, iMark_5, iMark_3, name, m_introns[ii] );
		}
		else {
			iStart = m_introns[ii].iend - 1 - m_intronMarginLeft;
			lenIntron = iStart - m_introns[ii].ibeg - m_intronMarginRight;

			if (lenIntron<m_minIntronLen)
				strSeg = "*";
			else {
				if (getSequence(strSeg, iSeq, iStart, 10 /*lenIntron*/, false))
					strSeg.assign(10,'N');
			}

			// 3' flank
			iStart3 = m_introns[ii].ibeg - 2;
			if (getSequence(str3, iSeq, iStart3, nFlank, false))
				str3.assign(nFlank,'N');
			if (getSequence(str3i, iSeq, iStart3 + 5 + nFlank, nFlank, false))
				str3i.assign(nFlank,'N');
			// 5' flank
			iStart5 = m_introns[ii].iend + nFlank - 1;
			if (getSequence(str5, iSeq, iStart5, nFlank, false))
				str5.assign(nFlank,'N');
			if (getSequence(str5i, iSeq, m_introns[ii].iend - 2 - 20, nFlank, false))
				str5i.assign(nFlank,'N');

			iMark_5 = m_introns[ii].extr_R;
			iMark_3 = m_introns[ii].extr_L;

			addFeatureSuffixHet( strSeg, iMark_5, iMark_3, name, m_introns[ii] );
		}

		// copy to output string
		strIntrons.append( strSeg );
		strIntrons.append("\n");

		strFlank.append( str5 );
		strFlank.append("\t");
		strFlank.append( str5i );
		strFlank.append("\t");
		strFlank.append( str3i );
		strFlank.append("\t");
		strFlank.append( str3 );
		strFlank.append("\n");

	}	// short alignment loop

	return false;			// success
}

bool cSite::correctIntronStrand( int iSeq )
{
	string strSeg, str5, str3, strJcnType;
	int iStart, len;
	bool bStrand;
	char chStrand;
	int ii, iStrand, iJcnType;

	int nCorrections(0);

	m_intronTypes.clear();

	// if no introns, return nothing
	if (m_introns.empty())
		return false;

	// get appropriate contig name
	string defline;
	m_psseq->getDefline( defline, iSeq );
	string name = m_psseq->getNameTok( defline );

	// grab all introns
	for (ii=0; ii<(int)m_introns.size(); ++ii) {

		// get intron terminal dinucleotide sequences and check for canonicality (strand-ness)
		// 5' end
		iStart = m_introns[ii].ibeg - 2;
		len = 4;
		if (getSequence(str5, iSeq, iStart, len, true))
			return true;		// bad coordinate!
		// 3' end
		iStart = m_introns[ii].iend - 3;
		len = 4;
		if (getSequence(str3, iSeq, iStart, len, true))
			return true;		// bad coordinate!

		getJcnType(iStrand, iJcnType, str5, str3, iSeq);
		m_introns[ii].iJcnType = iJcnType;

		// set m_intron[ii] strand here
		if (m_introns[ii].iStrand != iStrand)
			++nCorrections;

		m_introns[ii].iStrand = iStrand;

		strSeg.clear();

		switch (iStrand)
		{
		case -1:
			bStrand = false;
			chStrand = '-';
			break;
		case 1:
			bStrand = true;
			chStrand = '+';
			break;
		default:
			bStrand = true;
			chStrand = '*';
		};

	}	// intron loop

	printf("%d corrections found\n", nCorrections);

	return false;			// success
}

bool cSite::getJcnType( int& istrand, int& iJcnType, string& s5, string& s3, int iSeq )
{
	istrand = 0;
	iJcnType = -1;

	// canonical?
	if (s5.at(1)=='G' && s5.at(2)=='T' &&
		s3.at(1)=='A' && s3.at(2)=='G')
	{
		istrand = 1;
		iJcnType = INTRON_GT_AG;
		return false;
	}

	if (s5.at(1)=='C' && s5.at(2)=='T' &&
		s3.at(1)=='A' && s3.at(2)=='C')
	{
		istrand = -1;
		iJcnType = INTRON_GT_AG;
		return false;
	}

	// non-canonical but common (GC-AG)
	if (s5.at(1)=='G' && s5.at(2)=='C' &&
		s3.at(1)=='A' && s3.at(2)=='G')
	{
		istrand = 1;
		iJcnType = INTRON_GC_AG;
		return false;
	}

	if (s5.at(1)=='C' && s5.at(2)=='T' &&
		s3.at(1)=='G' && s3.at(2)=='C')
	{
		istrand = -1;
		iJcnType = INTRON_GC_AG;
		return false;
	}

	// non-canonical but common (AT-AC)
	if (s5.at(1)=='A' && s5.at(2)=='T' &&
		s3.at(1)=='A' && s3.at(2)=='C')
	{
		istrand = 1;
		iJcnType = INTRON_AT_AC;
		return false;
	}

	if (s5.at(1)=='G' && s5.at(2)=='T' &&
		s3.at(1)=='A' && s3.at(2)=='T')
	{
		istrand = -1;
		iJcnType = INTRON_AT_AC;
		return false;
	}

	return false;
}

bool cSite::getCanonicalStrand( int& istrand, string& s5, string& s3, int iSeq )
{
	istrand = 0;

	if (s5.at(1)=='G' && s5.at(2)=='T' &&
		s3.at(1)=='A' && s3.at(2)=='G')
	{
		istrand = 1;
		return false;
	}

	if (s5.at(1)=='C' && s5.at(2)=='T' &&
		s3.at(1)=='A' && s3.at(2)=='C')
	{
		istrand = -1;
		return false;
	}

	return false;
}

bool cSite::getAllNoncodingSeq( string& strNoncoding, int iSeq )
{
	string strIntergenic;
	string strIntrons;

	if ( getAllIntergenicSeq( strIntergenic, iSeq ) )
		return true;

	if ( getAllIntronSeq( strIntrons, iSeq ) )
		return true;

	strNoncoding = strIntergenic + strIntrons;
//	strNoncoding = strIntergenic;

	return false;			// success
}


bool cSite::checkPartial( int ig )
{
	if (m_genes[ig].strComment.find("partial",0) != string::npos)
		return true;

	return false;
}


bool cSite::getBoundarySites(int iSeq, bool bExtr)
{
	m_bExtr = bExtr;
	return getBoundarySites(iSeq);
}

bool cSite::getBoundarySites(int iSeq)
{
	// get introns and intergenic regions - forward and reverse strands
	// first get introns...
	string strIntrons;
	if ( getAllIntronSeq( strIntrons, iSeq ) )
		return true;

	// get intergenic (combine with introns to get all non-coding)
	string strIntergenic;
	if (!m_bExtr && m_nGC==-1)
	{
		if (getAllIntergenicSeq(strIntergenic, iSeq))
			return true;
	}

	// get coding regions
	string strCoding;
	if (getAllCodingSeq(strCoding, iSeq))
		return true;

	// assign member strings
	m_strCoding = strCoding;
	m_strIntrons = strIntrons;

	if (!m_bExtr)
	{
		if (m_bHomNoncoding)
			m_strNonCoding = m_strIntrons + strIntergenic;
		else
			m_strNonCoding = strIntergenic;
	}

	if (m_genes.size()==0)
		return false;			// nothing to do because there are no genes

	if (m_nGC != -1)
		return false;		// only want markov models if we are GC selective

	// extract sites
	// these offsets were hard-coded into the GeneMark-ES 2.0 program (Lomsadze et al)
	if (extractStarts( m_seqStarts, iSeq, m_startFlankLeft, m_startFlankRight ))
		return true;

	if (extractDon( m_seqDons, m_splicePhase, iSeq, m_donFlankLeft, m_donFlankRight ))
		return true;

	if (extractAcc( m_seqAccs, iSeq, m_accFlankLeft, m_accFlankRight ))
		return true;

	// sequences to support more complex intron sub-model
	int flag = 0;
	if (m_bBranchPointNext)
	{
		if (extractAcc( m_seqShortAccs, iSeq, m_shortAccFlankLeft, m_shortAccFlankRight ))
			return true;

		if (m_seqShortAccs.size()!=m_seqAccs.size())
		{
			printf("no. of acceptors = %d, no. of short acceptors = %d\n", (int)m_seqAccs.size(), (int)m_seqShortAccs.size());
			flag = 1;
		}
		assert( m_seqShortAccs.size()==m_seqAccs.size() );

		if (extractIntronTerm( m_seqBp, iSeq, m_intronTermLen, m_intronTermFlankRight ))
			return true;
	}

	if (extractTerm( m_seqTerms, iSeq, m_termFlankLeft, m_termFlankRight ))
		return true;

	// splice site size check
	if ( m_seqDons.size()!=m_splicePhase.size() )
	{
		report("Critical error: number of donors and splice phases are different.");
		printf("No. donors = %d, no. splice phases = %d\n", (int)m_seqDons.size(), (int)m_splicePhase.size() );
		return true;
	}

	return false;			// success
}

bool cSite::getCds(int iSeq)
{
	// get coding regions
	string strCoding;
	if (getAllCodingSeq(strCoding, iSeq))
		return true;

	// assign member strings
	m_strCoding = strCoding;

	return false;			// success
}

bool cSite::getCdsBoundarySites(int iSeq)
{
	// extract sites
	// these offsets were hard-coded into the GeneMark-ES 2.0 program (Lomsadze et al)
	if (extractStartsEstp( m_seqStarts, iSeq, m_startFlankLeft, m_startFlankRight ))
		return true;

	if (extractTermEstp( m_seqTerms, iSeq, m_termFlankLeft, m_termFlankRight ))
		return true;

	return false;
}

bool cSite::getBoundarySitesTrans(int iSeq)
{
	// get introns and intergenic regions - forward and reverse strands
	// first get introns...
	string strIntrons;
	if ( getAllIntronSeqTrans( strIntrons, iSeq ) )
		return true;

	// get intergenic (combine with introns to get all non-coding)
	string strIntergenic;
	if (!m_bExtr)
	{
		if (getAllIntergenicSeqTrans(strIntergenic, iSeq))
			return true;
	}

	// assign member strings
	m_strIntrons = strIntrons;

	if (!m_bExtr)
	{
		if (m_bHomNoncoding)
			m_strNonCoding = m_strIntrons + strIntergenic;
		else
			m_strNonCoding = strIntergenic;
	}

	if (m_transcripts.size()==0)
		return false;			// nothing to do because there are no genes

	// extract sites
	// these offsets were hard-coded into the GeneMark-ES 2.0 program (Lomsadze et al)
	if (extractDonTrans( m_seqDons, iSeq, m_donFlankLeft, m_donFlankRight ))
		return true;

	if (extractAccTrans( m_seqAccs, iSeq, m_accFlankLeft, m_accFlankRight ))
		return true;

	return false;			// success
}

bool cSite::getGappedAlignSitesHet(int iSeq, int nCodFlank, bool bCanonical)
{
	if (m_introns.empty())
		return false;			// nothing to do because there are not gapped alignments in this contig

	// get introns and intergenic regions - forward and reverse strands
	// first get introns...
	string strIntrons, strFlank;
	//if ( getAllIntronSeqSRAHet( strIntrons, strFlank, iSeq, nCodFlank ) )
	if ( getAllIntronSeqSRA( strIntrons, strFlank, iSeq, nCodFlank ) )
		return true;
	
	//m_psseq->Upper( strIntrons );

	// assign member strings
	m_strIntrons = strIntrons;
	m_strIntronFlank = strFlank;

	// extract sites
	// these offsets were hard-coded into the GeneMark-ES 2.0 program (Lomsadze et al)
	if (extractDonSRA( m_seqDons, iSeq, m_donFlankLeft, m_donFlankRight, bCanonical ))
		return true;

	if (extractAccSRA( m_seqAccs, iSeq, m_accFlankLeft, m_accFlankRight, bCanonical ))
		return true;

	return false;			// success
}

bool cSite::getGappedAlignSites(int iSeq, int nCodFlank, bool bCanonical)
{
	if (m_introns.empty())
		return false;			// nothing to do because there are not gapped alignments in this contig

	// get introns and intergenic regions - forward and reverse strands
	// first get introns...
	string strIntrons, strFlank;
	if ( getAllIntronSeqSRA( strIntrons, strFlank, iSeq, nCodFlank ) )
		return true;

	// assign member strings
	m_strIntrons = strIntrons;
	m_strIntronFlank = strFlank;

	// extract sites
	// these offsets were hard-coded into the GeneMark-ES 2.0 program (Lomsadze et al)
	if (extractDonSRA( m_seqDons, iSeq, m_donFlankLeft, m_donFlankRight, bCanonical ))
		return true;

	if (extractAccSRA( m_seqAccs, iSeq, m_accFlankLeft, m_accFlankRight, bCanonical ))
		return true;

	return false;			// success
}

bool cSite::getSequence(string& str, int iSeq, int iStart, int nNt, bool bStrand)
{
	bool bError = !m_psseq->getStretch( str, iSeq, iStart, nNt, bStrand );

	return bError;			// success
}

/*
	I-O fcns
*/
bool cSite::outputSites()
{
//	vector<string> m_seqStarts, m_seqTerms, m_seqDons, m_seqAccs;
	string strOut;
	FILE *pF = NULL;

	// write start sites
	strOut = m_strOutputFileBase + "_start_site.txt";
	pF = fopen( strOut.c_str(), "w" );
	if (pF==NULL)
		return true;		// error
	writeStringArray( pF, m_seqStarts );
	fclose( pF );

	// write start sites
	strOut = m_strOutputFileBase + "_stop_site.txt";
	pF = fopen( strOut.c_str(), "w" );
	if (pF==NULL)
		return true;		// error
	writeStringArray( pF, m_seqTerms );
	fclose( pF );

	// write start sites
	strOut = m_strOutputFileBase + "_don_site.txt";
	pF = fopen( strOut.c_str(), "w" );
	if (pF==NULL)
		return true;		// error
	writeStringArray( pF, m_seqDons );
	fclose( pF );

	// write start sites
	strOut = m_strOutputFileBase + "_acc_site.txt";
	pF = fopen( strOut.c_str(), "w" );
	if (pF==NULL)
		return true;		// error
	writeStringArray( pF, m_seqAccs );
	fclose( pF );

	return false;			// success
}


void cSite::writeBpData( FILE *pF, vector<bpRecord>& bpData )
{
	for (int i=0; i<(int)bpData.size(); ++i) {
		fprintf(pF, "%d  %d\n", bpData[i].istart, bpData[i].lenIntron );
		fprintf(pF, "%s\n", bpData[i].seqIntronTerm.c_str() );
	}
}


void cSite::writeStringArray( FILE *pF, vector<string>& strArr )
{
	for (int i=0; i<(int)strArr.size(); ++i) {
		fprintf(pF, "%s\n", strArr[i].c_str() );
	}
}

void cSite::writeIntArray( FILE *pF, vector<int>& strArr )
{
	for (int i=0; i<(int)strArr.size(); ++i) {
		fprintf(pF, "%d\n", strArr[i] );
	}
}

void cSite::getBpSites( vector<bpRecord>& bpData, char *pchBuf )
{
	char *pch = pchBuf;
	char *pchBegin = pchBuf;
	int n = (int)strlen( pchBuf );
	int nw = 0;
	int ipos = pch - pchBuf;

	bpData.clear();
	bpRecord rec;
	bool bFirst = true;

	while (true) {

		// read intron index and start position
		if (bFirst)
		{
			bFirst = false;
			pch = strtok( pchBuf, "\n" );
		}
		else
			pch = strtok( NULL, "\n" );

		if (pch==NULL)
			break;
#if defined(_WINDOWS_)
		nw = (int)strlen(pch);
		pch[nw-1] = '\0';		// reduce the number of bytes, but do not change iNext
#endif
		sscanf( pch, "%d %d\n", &rec.istart, &rec.lenIntron );

		// read intron term sequence
		pch = strtok( NULL, "\n" );
		if (pch==NULL)
			break;
#if defined(_WINDOWS_)
		nw = (int)strlen(pch);
		pch[nw-1] = '\0';		// reduce the number of bytes, but do not change iNext
#endif
		rec.seqIntronTerm = pch;

		bpData.push_back( rec );

		ipos = pch - pchBegin;
	}
}

void cSite::getSites( vector<string>& site, char *pchBuf )
{
	char *pch = pchBuf;
	char *pchBegin = pchBuf;
	int n = (int)strlen( pchBuf );
	int nw = 0;
	int ipos = pch - pchBuf;

	// clear the input array
	site.clear();

	// get the first line in the buffer
	pch = strtok( pchBuf, "\n" );

	while (ipos < n && pch!=NULL) {

#if defined(_WINDOWS_)
		nw = (int)strlen(pch);
		pch[nw-1] = '\0';		// reduce the number of bytes, but do not change iNext
#endif
		site.push_back( pch );

		pch = strtok( NULL, "\n" );
		ipos = pch - pchBegin;
	}
}

void cSite::getIntVector( vector<int>& int_vect, char *pchBuf )
{
	char *pch = pchBuf;
	char *pchBegin = pchBuf;
	int n = (int)strlen( pchBuf );
	int ipos = pch - pchBuf;

	// clear the input array
	int_vect.clear();

	// get the first line in the buffer
	pch = strtok( pchBuf, "\n" );

	while (ipos < n && pch!=NULL) {

		int_vect.push_back( atoi(pch) );

		pch = strtok( NULL, "\n" );
		ipos = pch - pchBegin;
	}
}

void cSite::splitStopSites( vector<string>& taa, vector<string>& tag, vector<string>& tga, vector<string>& other, char *pchBuf )
{
	char *pch = pchBuf;
	char *pchBegin = pchBuf;
	int n = (int)strlen( pchBuf );
	int nw;
	int ipos = pch - pchBuf;
	int iCodonOffset = 0;		// this may become an input someday

	// clear the input arrays
	taa.clear();
	tag.clear();
	tga.clear();
	other.clear();

	// get the first line in the buffer
	pch = strtok( pchBuf, "\n" );

	while (ipos < n && pch!=NULL) {	// collate each line in appropriate container

#if defined(_WINDOWS_)
		nw = (int)strlen(pch);
		pch[nw-1] = '\0';		// reduce the number of bytes, but do not change iNext
#endif

		if (pch[iCodonOffset]=='T')		// do not count any non-exact matches
		{
			if		(pch[iCodonOffset+1]=='A' && pch[iCodonOffset+2]=='A')	// taa
			{
				taa.push_back( pch );
			}
			else if (pch[iCodonOffset+1]=='A' && pch[iCodonOffset+2]=='G')	// tag
			{
				tag.push_back( pch );
			}
			else if (pch[iCodonOffset+1]=='G' && pch[iCodonOffset+2]=='A')	// tga
			{
				tga.push_back( pch );
			}
			else
				other.push_back( pch );
		}
		else
			other.push_back( pch );

		pch = strtok( NULL, "\n" );
		ipos = pch - pchBegin;

	}
}
