//**************************************************
// file:    filebuf.h
// project: probuild
//          alex lomsadze
//**************************************************
//#ifndef __FILEBUF_H__
//#define __FILEBUF_H__

#pragma once


#include <stdio.h>
#include <sys/types.h>
#include <sys/stat.h>

//--------------------------------------------------
class FileBuf
{
public:

	FileBuf() { begin = end = 0; };
	~FileBuf() { delete []begin; };

	char* begin;
	char* end;

	bool Load( const char* name )
	{
		size_t size;

		if ( name == 0 || *name == '\0' )
		{
			fprintf( stderr, "error: file name is missing\n" );
			return false;
		}

		struct stat st;
		if ( !stat( name, &st ) )
			size = st.st_size;
		else
		{
			fprintf( stderr, "error: file <%s> is not found\n", name );
			return false;
		}

		delete [] begin;

		begin = new char[ size + 1 ];
		if ( !begin )
		{
			fprintf( stderr, "error: not enough memory to load file <%s>\n", name );
			return false;
		}

		FILE* fp = fopen( name, "rb" );
		if ( !fp )
		{
			fprintf( stderr, "error: on opening file <%s>\n", name );
			return false;
		}
		size = fread( begin, 1, size, fp);
		if ( ferror(fp) )
		{
			fprintf( stderr, "error: on reading file <%s>\n", name );
			return false;
		}
		fclose( fp );

		begin[ size ] = '\0';
		end = begin + size;

		return true;
	}

	void Delete()
	{
		if (begin != NULL) {
			delete []begin;
			begin = NULL;
		}
	}

	int size()
	{
		return int(end-begin);
	}
};
//--------------------------------------------------
//#endif  // __FILEBUF_H__
